/*   File Header   ---------------------------------
     File Name:        data_intf.c
     Author and Date:  Paul B. Anderson
     Description:      Client interface routines for the Data Manager
                       and the Data Descriptor structure in particular
     End Header   ---------------------------------- */

/*  DEVELOPMENT SCOREBOARD

    Routine Name                    Coded?   Tested?
    -----------------------------   ------   -------
    create_data_set                   Y         Y
    delete_data_set                   Y         Y
    get_data_set_ids                  N         N
    get_data_id                       Y         Y
    set_data_detail                   Y         Y
    get_data_detail                   Y         Y
    set_fe_data                       Y         Y
    set_cm_data                       Y         Y
    set_ref_data                      Y         Y
    get_data_location                 Y         Y
    get_fe_info                       Y         Y
    get_cm_info                       Y         Y
    get_ref_info                      Y         Y
    set_data_type                     Y         Y
    get_data_type                     Y         Y
    get_mesh_info                     Y         Y
    set_2d_cartesian_coordinates      Y         Y
    set_3d_cartesian_coordinates      Y         Y
    set_angular_coordinates           Y         Y
    set_map_coordinates               Y         Y
    get_data_coord_system             Y         Y
    get_2d_cartesian_info             Y         Y
    get_3d_cartesian_info             Y         Y
    get_angular_info                  Y         Y
    get_map_info                      Y         Y

*/

/*  INCLUDES  */

#include <geodata.h>
#include <geodata_private.h>
#include <types.h>
#include <stdio.h>
#include <ci/data_intf.h>
#include <ci/library.h>

/*  PROTOTYPES  -- Export to geodata_private.h?  */

void GB_message();


/*   ========== Internal routines , definitions, and private data ==========   */

static char etext[50];

/*   ========== Routines past this point are for client use ==========   */

/*   Function Header   -----------------------------
     Description:      Create a data set descriptor
     Parameters:       label:   A character string containing the access
                                label for the data set descriptor.  Must
                                not be the null string.
     Preconditions:    The Geodata software must have been initialized prior
                       to the call.
     Postconditions:   The data descriptor is created if possible.
     Return Value:     The access ID of the created data descriptor
     Logic:
       1.   If label is null issue error message and discontinue
            processing.
       2.   Look up object by label.  If object found, issue
            error message and discontinue processing.
       3.   Create data descriptor and return ID to client.
     End Header   ---------------------------------- */

GB_id GI_create_data_set(label)
  char *label;
  
{
  char *place = "GI_create_data_set";
  GD_data_desc *data_descriptor;
  char tlabel[GB_LABEL_LENGTH];

  strncpy(tlabel,label,GB_LABEL_LENGTH);
  tlabel[GB_LABEL_LENGTH-1] = 0;
  
  if(strlen(tlabel) == 0) {
    GB_message(GB_ERROR,place,"Label string is 0 length"); }
  else {
    if(GB_get_id(tlabel,GB_DATA_DESC) != NULL) {
      GB_message(GB_ERROR,place,"Label is already used"); }
    else {
      data_descriptor = (GD_data_desc *)GB_add_object(tlabel,GB_DATA_DESC);

      if(data_descriptor == NULL) {
        GB_message(GB_ERROR,place,"Data Set not created");
      }
    }
  }
  data_descriptor->load_type=GD_CLIENT;
  return(GB_get_id(tlabel,GB_DATA_DESC));

}

/*   Function Header   -----------------------------
     Description:      Delete a data set descriptor
     Parameters:       dsid:     Data set ID
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
                       The Geodata software must have been initialized prior
                       to the call.
     Postconditions:   The data descriptor is deleted.
     Return Value:     (None)
     Logic:
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Delete object.
     End Header   ---------------------------------- */

void GI_delete_data_set(dsid)
  GB_id dsid;  
{
  char *place = "GI_delete_data_set";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);

  if (data_descriptor != NULL) GB_delete_object(dsid);

}


/*   Function Header   -----------------------------
     Name:             GI_get_data_set_ids
     Description:      Get the ids of all data sets
     Parameters:       array:  array of GB_IDs used to return ids
		       size:   Size of the array.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The array holds all the values in the table.
     Return Value:       0: Cannot fit all values into the array.
		       int: Number of valid ids in the array
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  Copy table values into the given array.
       3.  Return the number of values copied into the array,
           0 if they do not all fit.
     End Header   ---------------------------------- */
/*
int GI_get_data_set_ids (id, size)
     GB_id id[];
     int size;
{
  GS_representation_t rep;
  GS_table_t table;
  int count;

  
  for (i=0; i< GB_HASH_TABLE_SIZE; i++)
    if (hash_table[i].id != 0){
      hat = &hash_table[i];
      while (hat != NULL){
        GB_describe_object(hat->id,s); 
	hat = hat->next;
      }      
    }

    table = rep->table;
    count = 0;
    while ((table != NULL) && (count < size)) {
      values[count] = table->value;
      table = table->next;
      count++;
    }
    if (table == NULL)
      return (count);
    else {
      GB_message (GB_NOTIFY, "GI_get_data_set_ids",
		  "Cannot fit all ids into the given array.");
      return (0);
    }
  }
}
*/

/*   Function Header   -----------------------------
     Description:      Set the data detail string in a data set descriptor
     Parameters:       dsid:    Data set ID
                       detail:  A character string containing additional
                                information for use by client software.  Not
                                used by GeoData software at all.
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   The data set descriptor detail string is copied from 
                       the given value
     Return Value:     (None)
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Copy detail string value into data descriptor, adding null
           terminator.
     End Header   ---------------------------------- */

void GI_set_data_detail(dsid,detail)
  GB_id dsid; char *detail;
{
  char *place = "GI_set_data_detail";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);

  if (data_descriptor != NULL) {
    strncpy(data_descriptor->detail,detail,DETAIL_LENGTH);
    data_descriptor->detail[DETAIL_LENGTH-1] = 0;
  }
}

/*   Function Header   -----------------------------
     Description:      Set the data type code in a data set descriptor
     Parameters:       dsid:     Data set ID
                       code:     Type code, one of GD_POINT, GD_VECTOR, or
                                 GD_MESH
                       origin:   Pointer to an array of three floats
                                 containing the latitude, longitude, altitude
                                 (or x, y, z coordinates) of a mesh (0,0)
                                 or (0, 0, 0) point.  Interpretation of these
                                 values depends on the coordinate system and
                                 units
                       extent:   Pointer to an array of three floats
                                 containing the geographic extent of each
                                 axis of a mesh.  Interpretation of these
                                 values depends on the coordinate system and
                                 units
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   The data set descriptor code is changed to the given value
     Return Value:     (None)
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify supplied code is a legal choice.  If not, issue
           error message and discontinue processing.
       3.  Verify that origin and extent pointers are not NULL for
           mesh types.  If either pointer is null, issue an error
           message and discontinue processing.
       3.  Set new code value in data descriptor.  For meshes, copy
           origin and extent information using supplied pointers.
     End Header   ---------------------------------- */

void GI_set_data_type(dsid, code, origin, extent)
  GB_id dsid; short code;  float *origin, *extent;
{
  char *place = "GI_set_data_type";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);

  if (data_descriptor != NULL) {
    switch (code) {
    case GD_POINT_TYPE:   data_descriptor->type = code;
                     data_descriptor->load_type = GD_CLIENT;
                     break;
    case GD_VECTOR:  data_descriptor->type = code;
                     data_descriptor->load_type = GD_CLIENT_VECTOR;
                     break;
    case GD_MESH:    if (origin == NULL) {
                       sprintf(etext,"Origin field is NULL");
                       GB_message(GB_ERROR,place,etext); }
                     else {
                       if (extent == NULL) {
                         sprintf(etext,"Extent field is NULL");
                         GB_message(GB_ERROR,place,etext); }
                       else {
                         data_descriptor->type = code;
                         data_descriptor->origin[0] = *origin++;
                         data_descriptor->origin[1] = *origin++;
                         data_descriptor->origin[2] = *origin++;
                         data_descriptor->extent[0] = *extent++;
                         data_descriptor->extent[1] = *extent++;
                         data_descriptor->extent[2] = *extent++;
                       }
                     }
                     data_descriptor->load_type = GD_MESH_32;
                     break;
    default:         sprintf(etext,"Supplied code is not valid: %d",code);
                     GB_message(GB_ERROR,place,etext);
                     break;
    }
  }
}

/*   Function Header   -----------------------------
     Description:      Set the coordinate system in a data set descriptor
                       to 3D Cartesian Coordinates
     Parameters:       dsid:     Data set ID
                       units:    An integer value, one of GD_METERS, 
                                 GD_KM, GD_FEET, GD_MILES, GD_EARTH_RADII
                                 denoting the measurement system for all axes
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   The data set descriptor coordinate system is changed
                       to 3D cartesian coordinates.  No data conversion is implied
                       by this operation
     Return Value:     (None)
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify units code is legal.  If not, issue error message and
           discontinue processing.
       3.  Set new coordinate system value in data descriptor.
     End Header   ---------------------------------- */

void GI_set_3d_cartesian_coordinates(dsid,units)
  GB_id dsid;  int units;
{
  char *place = "GI_set_3d_cartesian_coordinates";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if(units != GD_METERS && units != GD_FEET &&
       units != GD_KM && units != GD_MILES && units != GD_EARTH_RADII) {
      sprintf(etext,"Altitude units choice is not valid: %d",units);
      GB_message(GB_ERROR,place,etext); }
    else {
      data_descriptor->coord_syst = GD_CART_3D;
      data_descriptor->units[0]   = units;
      data_descriptor->units[1]   = units;
      data_descriptor->units[2]   = units;
      data_descriptor->origin[0]  = 0.0;
      data_descriptor->origin[1]  = 0.0;
      data_descriptor->origin[2]  = 0.0;
      data_descriptor->surface    = 0.0;
    }
  }
}

/*   Function Header   -----------------------------
     Description:      Set the coordinate system in a data set descriptor
                       to angular coordinates
     Parameters:       dsid:     Data set ID
                       units_1:  An integer value, either GD_RADIANS or
                                 GD_DEGREES denoting the measurement system
                                 for latitudes and longitudes
                       units_2:  An integer value, one of GD_METERS, 
                                 GD_KM, GD_FEET, GD_MILES, GD_EARTH_RADII
                                 denoting the measurement system for altitudes
                       reference_surface:  An integer value, either GD_MSL or GD_CENTER
                                 indicating that altitudes are measured in units_2 units
                                 relative to mean sea level, or relative to the
                                 center of the earth
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   The data set descriptor coordinate system is changed
                       to angular coordinates.  No data conversion is implied
                       by this operation
     Return Value:     (None)
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify units code is legal.  If not, issue error message and
           discontinue processing.
       3.  Set new coordinate system value in data descriptor.
     End Header   ---------------------------------- */

void GI_set_angular_coordinates(dsid,units_1,units_2,reference_surface)
  GB_id dsid;  int units_1,units_2;  int reference_surface;
{
  char *place = "GI_set_angular_coordinates";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if(units_1 != GD_RADIANS && units_1 != GD_DEGREES) {
      sprintf(etext,"Angular units choice is not valid: %d",units_1);
      GB_message(GB_ERROR,place,etext); }
    else {
      if(units_2 != GD_METERS && units_2 != GD_FEET &&
         units_2 != GD_KM && units_2 != GD_MILES && units_2 != GD_EARTH_RADII) {
        sprintf(etext,"Altitude units choice is not valid: %d",units_2);
        GB_message(GB_ERROR,place,etext); }
      else {
        data_descriptor->coord_syst = GD_ANGULAR;
        data_descriptor->units[0] = units_1;
        data_descriptor->units[1] = units_1;
        data_descriptor->units[2] = units_2;
        data_descriptor->origin[0] = 0.0;
        data_descriptor->origin[1] = 0.0;
        data_descriptor->origin[2] = 0.0;
        switch (reference_surface) {
        case GD_MSL:    data_descriptor->surface = 0.0;
                        break;
        case GD_CENTER: data_descriptor->surface = -GS_EARTH_RADIUS;
                        break;
        default:        sprintf(etext,"Reference surface choice is not valid: %d",
                                reference_surface);
                        GB_message(GB_ERROR,place,etext);
        }
      }
    }
  }
}

/*   Function Header   -----------------------------
     Description:      Set the coordinate system in a data set descriptor
                       to 2D Cartesian Coordinates
     Parameters:       dsid:     Data set ID
                       units:    An integer value, one of GD_METERS, 
                                 GD_KM, GD_FEET, GD_MILES, GD_EARTH_RADII
                                 denoting the measurement system for axes
                       r:        Pointer to a radius vector (float array[3]) in 
                                 standard GeoData 3D cartesian coordinates
                                 (Meters, Meters, Meters) giving the position of
                                 the 2D plane origin
                       nv:       Pointer to a vector in 3D cartesian coordinates giving
                                 a direction normal to the desired 2D plane.
                                 This need not be in any particular units since
                                 only the direction is important.
                       y:        Pointer to a radius vector in standard GeoData 3D
                                 cartesian coordinates (Meters, Meters, Meters)
                                 giving a point to used in constructing the y axis on
                                 the 2D plane.  The point must lie on the defined
                                 plane and the vector between r and y is the basis
                                 for the positive y axis.  The x axis is computed from
                                 the y axis and the supplied normal vector via vector
                                 cross product.  This parameter may be supplied as 0 or
                                 NULL in which case the polar axis is used to derive
                                 the 2D planar y axis.  If r is in the northern
                                 hemisphere, the projection of North is used. If 
                                 r is in the southern hemisphere, the projection of South
                                 determines the y axis.  For planes at either pole, the
                                 3D cartesian x and y axes are used.
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   The data set descriptor coordinate system is changed
                       to 3D cartesian coordinates.  No data conversion is implied
                       by this operation
     Return Value:     (None)
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify that the supplied units code is one of the legal choices.
           If an error is an encountered, issue error message and
           discontinue processing.
       3.  Set new coordinate system value in data descriptor.
       4.  Copy r and nv vectors into descriptor.
       5.  If the user supplied y axis is NULL, deallocate any
           prior axis array and set pointer to NULL.
       6.  If the user supplied y axis is not NULL, allocate an
           axis area if needed and copy values to it.
     End Header   ---------------------------------- */

void GI_set_2d_cartesian_coordinates(dsid,units,r,nv,y)
  GB_id dsid;  int units;  float *r, *nv, *y;
{
  char *place = "GI_set_2d_cartesian_coordinates";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if(units != GD_METERS && units != GD_FEET &&
       units != GD_KM && units != GD_MILES && units != GD_EARTH_RADII) {
      sprintf(etext,"Altitude units choice is not valid: %d",units);
      GB_message(GB_ERROR,place,etext); }
    else {
      data_descriptor->coord_syst = GD_CART_2D;
      data_descriptor->units[0] = units;
      data_descriptor->units[1] = units;
      data_descriptor->units[2] = units;
      data_descriptor->r[0]     = *r;
      data_descriptor->r[1]     = *(r+1);
      data_descriptor->r[2]     = *(r+2);
      data_descriptor->nv[0]    = *nv;
      data_descriptor->nv[1]    = *(nv+1);
      data_descriptor->nv[2]    = *(nv+2);
      if (y == NULL) {
        data_descriptor->user_yp = FALSE; }
      else {
        data_descriptor->user_yp = TRUE;
        data_descriptor->yp[0] = *y;
        data_descriptor->yp[1] = *(y+1);
        data_descriptor->yp[2] = *(y+2);
      }
    }
  }
}

/*   Function Header   -----------------------------
     Description:      Set the coordinate system in a data set descriptor
                       to map coordinates
     Parameters:       dsid:     Data set ID
                       proj_info:  Projection information structure returned
                                 from a projection function such as GI_perspective
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   The data set descriptor coordinate system is changed
                       to map coordinates.  No data conversion is implied
                       by this operation
     Return Value:     (None)
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Set new coordinate system value and projection parameters 
           in data descriptor.
       3.  If data type is GD_MESH, verify that the extent parameter
           array has every element between -1 and 1.  If not, issue
           an error message.
     End Header   ---------------------------------- */

void GI_set_map_coordinates(dsid,proj_info)
  GB_id dsid;  GI_proj_desc *proj_info;
{
  char *place = "GI_set_map_coordinates";
  GD_data_desc *data_descriptor;  int i;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (proj_info == NULL) {
      sprintf(etext,"Projection is missing");
      GB_message(GB_ERROR,place,etext); }
    else {
      data_descriptor->coord_syst = GD_MAP;
      data_descriptor->units[0] = 0;
      data_descriptor->units[1] = 0;
      data_descriptor->units[2] = 0;
      data_descriptor->proj_fn = proj_info->code;
      for(i=0; i<6; i++)
        data_descriptor->proj_params[i] = proj_info->param[i];
      if (data_descriptor->type == GD_MESH) {
        for(i=0; i<3; i++)
          if (data_descriptor->extent[i] < -1.0 ||
              data_descriptor->extent[i] >  1.0) {
            sprintf(etext,"Extent is out of range (-1,1) in position %d: %f",
                    i,data_descriptor->extent[i]);
            GB_message(GB_ERROR,place,etext); }
      }
    }
  }
}


/* GI_NP2 returns the next highest power of 2 to n or n if n is a power of 2 */
/* used by set_fe_data when setting appropriate values for the geometry */
/* djones 7/2/91 */
unsigned GI_np2(n)  
     unsigned int n;
{
  unsigned i, c = 0, t;
  for (i = 1 << 31; i > 0; i >>= 1)
    if ((i & n) != 0) {
      c++;
      if (c == 1) t = i;
    }
  if (c == 1) return(t);
  else return (t << 1);
}


/*   Function Header   -----------------------------
     Description:      Set front end data information in a data set descriptor
     Parameters:       dsid:     Data set ID
                       data:     Pointer to an array of data
                       rank:     Integer rank of the data (number of dimensions)
                                 Arrays are assumed to be stored in row major
                                 order
                       shape:    An array of integers of length rank which
                                 gives the number of elements in each dimension
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   The data set descriptor code is changed to the given value
     Return Value:     (None)
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify supplied rank and all elements of shape are positive.  If not,
           issue error message and discontinue processing.
       3.  Set parameters values into data descriptor;
     End Header   ---------------------------------- */

void GI_set_fe_data(dsid, data, rank, shape)
  GB_id dsid; float *data; int rank, shape[];
{
  char *place = "GI_set_fe_data";
  GD_data_desc *data_descriptor;  int i, error;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (data == NULL) {
      sprintf(etext,"Data pointer is NULL");
      GB_message(GB_ERROR,place,etext); }
    else {
      if (rank < 1 || rank > 3) {
        sprintf(etext,"Rank value is negative or greater than 3: %d",rank);
        GB_message(GB_ERROR,place,etext); }
      else {
        for (i=0, error = FALSE; i < rank; i++) {
          if (shape[i] < 1) {
            error = TRUE;
            sprintf(etext,"Shape value is less than 1 at index %d",i);
            GB_message(GB_ERROR,place,etext);
	  }
        }
        if (!error) {
          data_descriptor->rank = rank;
          data_descriptor->fe_data = (int *)data;
          for (i=0; i < rank; i++) data_descriptor->fe_shape[i] = shape[i];
	  /* added by djones to implement front end data */
	  for(i=0;i<4;i++)data_descriptor->geometry[i]=1;
	  for(i=0;i<rank;i++)data_descriptor->geometry[i]=shape[i];
	  for(i=0;i<rank;i++)
	    data_descriptor->geometry[i]=GI_np2(data_descriptor->geometry[i]);
	  while ( data_descriptor->geometry[0]*data_descriptor->geometry[1] < 
		 CM_physical_processors_limit) 
	    data_descriptor->geometry[0]=GI_np2(data_descriptor->geometry[0]+1);
          data_descriptor->values = NULL;
          data_descriptor->context = NULL;
          data_descriptor->cm_fe_flag = 1;
        }
      }
    }
  }
}


/*   Function Header   -----------------------------
     Description:      Set CM data information in a data set descriptor
     Parameters:       dsid:     Data set ID
                       data:     CM field id of field containing data values
                       context:  CM field id of 1 bit field containing
                                 context flags for each data value.  This
                                 argument may be set to NULL if all elements
                                 of the data field are to be used
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   The data set descriptor is modified to record the data
                       and context field ids
     Return Value:     (None)
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify data field id is not null.  If it is null,
           issue error message and discontinue processing.
       3.  Set field ids into data descriptor.
     End Header   ---------------------------------- */

void GI_set_cm_data(dsid, values, context)
  GB_id dsid; CM_field_id_t values, context;
{
  char *place = "GI_set_cm_data";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (values == NULL) {
      sprintf(etext,"Value field is NULL");
      GB_message(GB_ERROR,place,etext); }
    else {
      data_descriptor->values = values;
      *(data_descriptor->vp_set) = CM_field_vp_set(values);
      data_descriptor->context = context;
      data_descriptor->fe_data = NULL;
    /* a slight intrusion by eah....where does this get set?*/
      data_descriptor->rank = CM_geometry_rank
	(CM_vp_set_geometry(CM_field_vp_set(values)));
    }
  }
}


/*   Function Header   -----------------------------
     Description:      Set reference data information in a data set descriptor
     Parameters:       dsid:     Data set ID
                       (rest TBD)
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   The data set descriptor is modified to record the
                       supplied information
     Return Value:     (None)
     Logic:            
       1.  Issue error message for unimplemented function.
     End Header   ---------------------------------- */

void GI_set_ref_data(dsid)
  GB_id dsid;
{
  char *place = "GI_set_ref_data";
  GD_data_desc *data_descriptor;

  sprintf(etext,"This capability is not yet implemented");
  GB_message(GB_ERROR,place,etext);

/*  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (data_descriptor->ref == NULL) {
      sprintf(etext,"Reference descriptor is NULL");
      GB_message(GB_ERROR,place,etext); }
    else {
    }
  }  */
}

/* ==========  Accessor Functions Past This Point  ========== */


/*   Function Header   -----------------------------
     Description:      Get the data set detail field from a data set descriptor
     Parameters:       dsid:     Data set ID
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state
     Return Value:     pointer to a character string containing the detail
                       string.  This is not a copy but is a pointer to the
                       actual string.
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Extract detail pointer from data descriptor and return it.
     End Header   ---------------------------------- */

char *GI_get_data_detail(dsid)
  GB_id dsid;
{
  char *place = "GI_get_data_detail";
  GD_data_desc *data_descriptor; char *result;

  result = NULL;
  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    result = data_descriptor->detail;
  }

  return(result);
}

/*   Function Header   -----------------------------
     Description:      Get the data set id given the label string
     Parameters:       label:  A character string giving the label
                               of an existing data set descriptor
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state
     Return Value:     data set id code
     Logic:            
       1.  Retrieve data set id using label and object type code.
           Return id to client.
     End Header   ---------------------------------- */

GB_id GI_get_data_id(label)
  char *label;
{
  return(GB_get_id(label,GB_DATA_DESC));
}

/*   Function Header   -----------------------------
     Description:      Get the data set type code from a data set descriptor
     Parameters:       dsid:     Data set ID
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state
     Return Value:     data type code
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Extract code from data descriptor and return it.
     End Header   ---------------------------------- */

int GI_get_data_type(dsid)
  GB_id dsid;
{
  char *place = "GI_get_data_type";
  GD_data_desc *data_descriptor; int code;

  code = NULL;
  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    code = data_descriptor->type;
  }

  return(code);
}

/*   Function Header   -----------------------------
     Description:      Get mesh information from a data set descriptor for
                       a mesh data set
     Parameters:       dsid:     Data set ID
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state
     Return Value:     data type code
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify that the data set is of type GD_MESH.  If not,
           issue an error message and discontinue processing.
       3.  Extract origin and extents from data descriptor and return them.
     End Header   ---------------------------------- */

void GI_get_mesh_info(dsid,origin,extent)
  GB_id dsid;  float origin[], extent[];
{
  char *place = "GI_get_mesh_info";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (data_descriptor->type != GD_MESH) {
      sprintf(etext,"Data set is not of type GD_MESH:  %d",
              data_descriptor->type);
      GB_message(GB_ERROR,place,etext); }
    else {
      origin[0] = data_descriptor->origin[0];
      origin[1] = data_descriptor->origin[1];
      origin[2] = data_descriptor->origin[2];
      extent[0] = data_descriptor->extent[0];
      extent[1] = data_descriptor->extent[1];
      extent[2] = data_descriptor->extent[2];
    }
  }
}

/*   Function Header   -----------------------------
     Description:      Get the data set coordinate system code from a
                       data set descriptor
     Parameters:       dsid:     Data set ID
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state
     Return Value:     Coordinate system code
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Extract code from data descriptor and return it.
     End Header   ---------------------------------- */

int GI_get_data_coord_system(dsid)
  GB_id dsid;
{
  char *place = "GI_get_data_coord_system";
  GD_data_desc *data_descriptor; int code;

  code = NULL;
  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    code = data_descriptor->coord_syst;
  }

  return(code);
}

/*   Function Header   -----------------------------
     Description:      Get the data set coordinate system information
                       for angular coordinates from a data set descriptor
     Parameters:       dsid:     Data set ID
                       units_1:  Pointer to an integer value, returned with
                                 either GD_RADIANS or GD_DEGREES denoting
                                 the measurement system for latitudes and longitudes
                       units_2:  Pointer to an integer value, returned as 
                                 one of GD_METERS, GD_KM, GD_FEET, GD_MILES,
                                 GD_EARTH_RADII denoting the measurement system
                                 for altitudes
                       reference_surface:  Pointer to an integer value, returned
                                 with either GD_MSL or GD_CENTER indicating 
                                 that altitudes are measured in units_2 units
                                 relative to mean sea level, or relative to the
                                 center of the earth
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state but client's output variables have
                       been altered
     Return Value:     None -- values returned via calling sequence pointers
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify that coordinate system is angular.  If not,
           issue error message and discontinue processing.
       3.  Extract information from data descriptor and return it.
     End Header   ---------------------------------- */

void GI_get_angular_info(dsid,units_1,units_2,reference_surface)
  GB_id dsid;  int *units_1, *units_2;  int *reference_surface;
{
  char *place = "GI_get_angular_info";
  GD_data_desc *data_descriptor; int code;

  code = NULL;
  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (data_descriptor->coord_syst != GD_ANGULAR) {
      sprintf(etext,"Data set does not have angular coordinates:  %d",
              data_descriptor->coord_syst);
      GB_message(GB_ERROR,place,etext); }
    else {
      *units_1 = data_descriptor->units[0];
      *units_2 = data_descriptor->units[2];
      if (data_descriptor->surface == 0.0) {
        *reference_surface = GD_MSL; }
      else {
        *reference_surface = GD_CENTER;
      }
    }
  }
}


/*   Function Header   -----------------------------
     Description:      Get the data set coordinate system information
                       for 3D cartesian coordinates from a data set descriptor
     Parameters:       dsid:     Data set ID
                       units:    Pointer to an integer value, returned with 
                                 one of GD_METERS, GD_KM, GD_FEET, GD_MILES,
                                 or GD_EARTH_RADII denoting the measurement
                                 system for all axes
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state but client's output variables have
                       been altered
     Return Value:     None -- values returned via calling sequence pointers
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify that coordinate system is 3D cartesian.  If not,
           issue error message and discontinue processing.
       3.  Extract information from data descriptor and return it.
     End Header   ---------------------------------- */

void GI_get_3d_cartesian_info(dsid,units)
  GB_id dsid;  int *units;
{
  char *place = "GI_get_3D_cartesian_info";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (data_descriptor->coord_syst != GD_CART_3D) {
      sprintf(etext,"Data set does not have 3D cartesian coordinates:  %d",
              data_descriptor->coord_syst);
      GB_message(GB_ERROR,place,etext); }
    else {
      *units = data_descriptor->units[0];
    }
  }
}

/*   Function Header   -----------------------------
     Description:      Get coordinate system information from a data set
                       descriptor using map coordinates
     Parameters:       dsid:     Data set ID
                       proj_code: Pointer to an integer returned with the
                                 projection code (one of GS_NO_PROJECTION,
                                 GS_MILLER, GS_MOLLWEIDE, GS_LAMBERT,
                                 GS_POLAR_STEREO, GS_GNOMONIC, or
                                 GS_PERSPECTIVE
                       params    Pointer to an array of projection parameters
                                 returned with processed values generated
                                 from a projection function such as GI_perspective
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state but client's output variables have
                       been altered
     Return Value:     None -- values returned via calling sequence pointers
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify that coordinate system is GD_MAP.  If not,
           issue error message and discontinue processing.
       3.  Extract information from data descriptor and return it.
     End Header   ---------------------------------- */

void GI_get_map_info(dsid,proj_code,param)
  GB_id dsid;  int *proj_code;  float param[];
{
  char *place = "GI_get_map_info";
  GD_data_desc *data_descriptor;  int i;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (data_descriptor->coord_syst != GD_MAP) {
      sprintf(etext,"Data set does not have map coordinates:  %d",
              data_descriptor->coord_syst);
      GB_message(GB_ERROR,place,etext); }
    else {
      *proj_code = data_descriptor->proj_fn;
      for (i=0; i < 10; i++) 
        param[i] = data_descriptor->proj_params[i];
    }
  }
}

/*   Function Header   -----------------------------
     Description:      Get the coordinate system in a data set descriptor
                       to 2D Cartesian Coordinates
     Parameters:       dsid:     Data set ID
                       units:    A pointer to an integer value, returned with
                                 one of GD_METERS, GD_KM, GD_FEET, GD_MILES,
                                 GD_EARTH_RADII denoting the measurement system
                                 for axes
                       r:        Pointer to a radius vector (float array[3]) in 
                                 standard GeoData 3D cartesian coordinates
                                 (Meters, Meters, Meters) returned with the
                                 position of the 2D plane origin
                       nv:       Pointer to a vector in 3D cartesian coordinates
                                 returned with a direction normal to the
                                 desired 2D plane. This will be in the same
                                 units as originally specified
                       y:        Pointer to a radius vector in standard GeoData 3D
                                 cartesian coordinates (Meters, Meters, Meters)
                                 returned with a point determining the y axis on
                                 the 2D plane.  The point lies on the defined
                                 plane and the vector between r and y is the basis
                                 for the positive y axis.  The x axis is computed from
                                 the y axis and the supplied normal vector via vector
                                 cross product.  Unless overridden by the client,
                                 the polar axis is used to derive the 2D planar y
                                 axis.  If r is in the northern hemisphere, the
                                 projection of North is used. If r is in the 
                                 southern hemisphere, the projection of South
                                 determines the y axis.  For planes at either pole,
                                 the 3D cartesian x and y axes are used.
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state but client's output variables have
                       been altered
     Return Value:     None -- values returned via calling sequence pointers
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify that the coordinate system in 2D cartesian.  If not,
           issue error message and discontinue processing.
       3.  Set new coordinate system value in data descriptor.
       4.  Copy r and nv vectors into descriptor.
       5.  If the user supplied y axis is NULL, deallocate any
           prior axis array and set pointer to NULL.
       6.  If the user supplied y axis is not NULL, allocate an
           axis area if needed and copy values to it.
     End Header   ---------------------------------- */

void GI_get_2d_cartesian_info(dsid,units,r,nv,y)
  GB_id dsid;  int *units;  float r[], nv[], y[];
{
  char *place = "GI_get_2d_cartesian_info";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (data_descriptor->coord_syst != GD_CART_2D) {
      sprintf(etext,"Data set does not have 2D cartesian coordinates:  %d",
              data_descriptor->coord_syst);
      GB_message(GB_ERROR,place,etext); }
    else {
      data_descriptor->coord_syst = GD_CART_2D;
      *units = data_descriptor->units[0];
      r[0] = data_descriptor->r[0];
      r[1] = data_descriptor->r[1];
      r[2] = data_descriptor->r[2];
      nv[0] = data_descriptor->nv[0];
      nv[1] = data_descriptor->nv[1];
      nv[2] = data_descriptor->nv[2];
      if (data_descriptor->yp != NULL) {
        y[0] = data_descriptor->yp[0];
        y[1] = data_descriptor->yp[1];
        y[2] = data_descriptor->yp[2];
      }
    }
  }
}

/*   Function Header   -----------------------------
     Description:      Get the data set location code from a
                       data set descriptor
     Parameters:       dsid:     Data set ID
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state
     Return Value:     Data set location code, an integer, taking on one
                       of the following values:  GD_FE, GD_CM, GD_REF
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Determine code from data descriptor and return it.
     End Header   ---------------------------------- */

int GI_get_data_location(dsid)
  GB_id dsid;
{
  char *place = "GI_get_data_location";
  GD_data_desc *data_descriptor;  int code;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  /*  FIX THIS*/
/*  if (data_descriptor != NULL) {
    if (data_descriptor->fe_data != NULL) {
      code = GD_FE; }
    else {
      if (data_descriptor->ref_resolution != NULL) {
        code = GD_REF; }
      else {
        code = GD_CM;
      }
    }
  }*/
  return(code);
}


/*   Function Header   -----------------------------
     Description:      Get front end data information from a data set descriptor
     Parameters:       dsid:     Data set ID
                       data:     Pointer to a variable returned with a pointer
                                 to the array of data
                       rank:     Pointer to an integer returned with the rank
                                 of the data (number of dimensions)
                       shape:    Pointer to an array of integers of length rank
                                 returned with the number of elements in each dimension
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state but client's output variables have
                       been altered
     Return Value:     None -- values returned via calling sequence pointers
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify the data descriptor describes front end data.  If not,
           issue error message and discontinue processing.
       3.  Extract data pointer, rank, and shape, storing into output
           parameters.
     End Header   ---------------------------------- */

void GI_get_fe_info(dsid, data, rank, shape)
  GB_id dsid; float **data; int *rank, shape[];
{
  char *place = "GI_get_fe_info";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (data_descriptor->fe_data == 0) {
      sprintf(etext,"Data descriptor does not describe front end data");
      GB_message(GB_ERROR,place,etext); }
    else {
      *data = (float *)data_descriptor->fe_data;
      *rank = data_descriptor->rank;
      shape[0] = data_descriptor->geometry[0];
      shape[1] = data_descriptor->geometry[1];
      shape[2] = data_descriptor->geometry[2];
    }
  }
}


/*   Function Header   -----------------------------
     Description:      Get CM data information from a data set descriptor
     Parameters:       dsid:     Data set ID
                       value_id:  Pointer to a variable returned with a field id
                                 of the CM data
                       context_id:  Pointer to a variable returned with a field
                                 id of the CM context flag related to the data
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state but client's output variables have
                       been altered
     Return Value:     None -- values returned via calling sequence pointers
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify the data descriptor describes CM data.  If not,
           issue error message and discontinue processing.
       3.  Extract data field id and context field id, storing into output
           parameters.
     End Header   ---------------------------------- */

void GI_get_cm_info(dsid, value_id, context_id)
  GB_id dsid;  int *value_id, *context_id;
{
  char *place = "GI_get_cm_info";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (data_descriptor->values == NULL) {
      sprintf(etext,"Data descriptor does not describe CM data");
      GB_message(GB_ERROR,place,etext); }
    else {
      *value_id   = (int)data_descriptor->values;
      *context_id = (int)data_descriptor->context;
    }
  }
}


/*   Function Header   -----------------------------
     Description:      Get reference data information from a data set descriptor
     Parameters:       dsid:     Data set ID
                       ref_descriptor:  Pointer to a variable returned with a 
                                 pointer to the reference descriptor.  This
                                 descriptor is of type GD_ref_desc and the
                                 pointer points to the actual structure being
                                 used rather than a copy
     Preconditions:    The dsid must be a valid ID of an existing data descriptor
     Postconditions:   No change of state but client's output variables have
                       been altered
     Return Value:     None -- values returned via calling sequence pointers
     Logic:            
       1.  Retrieve data base entry using id and access data descriptor.
           If an error is an encountered, issue error message and
           discontinue processing.
       2.  Verify the data descriptor describes reference data.  If not,
           issue error message and discontinue processing.
       3.  Extract data field id and context field id, storing into output
           parameters.
     End Header   ---------------------------------- */
/*
void GI_get_ref_info(dsid, ref_descriptor)
  GB_id dsid;  GD_ref_resolution_list **ref_descriptor;
{
  char *place = "GI_get_ref_info";
  GD_data_desc *data_descriptor;

  data_descriptor = (GD_data_desc *)GI_get_instance_internal
                                      (place,GB_DATA_DESC,dsid);
  if (data_descriptor != NULL) {
    if (data_descriptor->ref_resolution == NULL) {
      sprintf(etext,"Data descriptor does not describe reference data");
      GB_message(GB_ERROR,place,etext); }
    else {
      *ref_descriptor   = data_descriptor->ref_resolution;
    }
  }
}

*/
