/*   File Header   ---------------------------------
     File Name:        rep_intf.c
     Author and Date:  Glenn Heinle         6/19/90
     Description:      Client interface routines for the
                       data representation data structure
     Functions:
      GS_representation_t GI_get_rep_internal ()
      GB_id       GI_get_symbol_id ()
      GB_id       GI_get_representation_id ()
      GB_id       GI_create_representation ()
      void        GI_delete_representation ()
      int         GI_get_representation_rep_code ()
      void        GI_set_representation_rep_code ()
      int         GI_get_representation_translucency ()
      void        GI_set_representation_translucency ()
      int         GI_get_representation_what_varies ()
      void        GI_set_representation_what_varies ()
      GS_table_t  GI_get_table_t ()
      int         GI_get_table_values ()
      void        GI_add_table_value ()
      void        GI_delete_table_value ()
      void        GI_get_table_color ()
      void        GI_set_table_color ()
      GB_id       GI_get_table_discrete ()
      void        GI_set_table_discrete ()
      float       GI_get_table_scale ()
      void        GI_set_table_scale ()
      float       GI_get_table_rotation ()
      void        GI_set_table_rotation ()
		       
     End Header   ---------------------------------- */


#include <geodata.h>
#include <geodata_private.h>
#include <stdio.h>
#include <types.h>

#define GI_EPSILON 0.000001
#define ASSIGN_TABLE(theTable, val, red, green, blue, scal, rotn) { \
	(theTable)->value = (val); \
	(theTable)->color[0] = (red); \
	(theTable)->color[1] = (green); \
	(theTable)->color[2] = (blue); \
	(theTable)->scale = (scal); \
	(theTable)->rotation = (rotn); \
	}

extern GS_table_t GB_create_table ();


/*   Function Header   -----------------------------
     Name:             GI_get_rep_internal
     Description:      Return a pointer to a GS_representation structure.
     Parameters:       place:   A character string giving the calling
                                routine name, used for error messages
                       id:      An integer object id, used for retrieval
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     Return a pointer to a GS_representation structure.
                       NULL if a GS_representation was not found with the
		       given id.
     Logic:
       1.  Calls GI_get_instance_internal and casts value into
           a GS_representation_t.
     End Header   ---------------------------------- */

GS_representation_t GI_get_rep_internal (place, id)
     char *place;
     GB_id id;
{
  return ((GS_representation_t)
	  GI_get_instance_internal (place, GB_REPRESENTATION, id));
}


/*   Function Header   -----------------------------
     Name:             GI_get_symbol_id
     Description:      Locate a symbol with the given label.
     Parameters:       label: A character string containing the access label
                              for the symbol.  Must not be a null string.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     The GB_id of the symbol with the given label if it
                       exists
     Logic:
       1.  Call GB_get_id with the GB_SYMBOL type.
     End Header   ---------------------------------- */

GB_id GI_get_symbol_id (label)
     char *label;
{
  return (GB_get_id (label, GB_SYMBOL));
}


/*   Function Header   -----------------------------
     Name:             GI_get_representation_id
     Description:      Locate a representation with the given label.
     Parameters:       label: A character string containing the access label
                              for the representation.  Must not be a null
			      string.
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     The GB_id of the representaion with the given label if
                       it exists
     Logic:
       1.  Call GB_get_id with the GB_REPRESENTATION type.
     End Header   ---------------------------------- */

GB_id GI_get_representation_id (label)
     char *label;
{
  return (GB_get_id (label, GB_REPRESENTATION));
}


/*   Function Header   -----------------------------
     Name:             GI_create_representation
     Description:      Create a new representation
     Parameters:       label: A character string containing the label of the
                              new representation.  Must not be a null string.
		       rep_code: One of the representations defined in
		                 geodata.h
     Preconditions:    The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   A new representation is created if possible.
     Return Value:     The GB_id of the new representation.
     Logic:
       1.  Call GB_add_object to create the representation.
       2.  Set the rep_code as supplied.
       3.  Set default values as follows:
              Translucency: 0
	      What_varies:  GS_NOTHING
       4.  Set the first table value with the following attributes:
              Value:    0.0
	      Color:    White
	      Scale:    0.0
	      Rotation: 0.0
     End Header   ---------------------------------- */

GB_id GI_create_representation (label, rep_code)
     char *label;
     unsigned rep_code;
{
  GS_representation_t rep;
  GS_table_t table;
  GB_id id;

  rep = (GS_representation_t) GB_add_object (label, GB_REPRESENTATION);
  id = GB_get_id (label, GB_REPRESENTATION);
  rep->rep_code = rep_code;
  rep->translucency = 0;
  rep->what_varies = GS_NOTHING_VARIES;

  ASSIGN_TABLE(rep->tables, 0.0, 255, 255, 255, 0.0, 0.0);

  return (id);
}


/*   Function Header   -----------------------------
     Name:             GI_delete_representation
     Description:      Delete a representation.
     Parameters:       rep_id: representation ID.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The representation is deleted.
     Return Value:     (None)
     Logic:
       1.  Call GI_get_rep_internal.  This produces an error message
           if the representation does not exist.
       2.  Delete representation object.
     End Header   ---------------------------------- */

void GI_delete_representation (rep_id)
     GB_id rep_id;
{
  if (GI_get_rep_internal ("GI_delete_representation", rep_id))
    GB_delete_object (rep_id);
}


/*   Function Header   -----------------------------
     Name:             GI_get_representation_rep_code
     Description:      Rep_code accessor for representation.
     Parameters:       rep_id: representation ID.
     Preconditions     The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     The representation rep_code.
                       -1 if there is an error.
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  Check for error and return rep_code or -1 accordingly
     End Header   ---------------------------------- */

int GI_get_representation_rep_code (rep_id)
     GB_id rep_id;
{
  GS_representation_t rep;

  rep = GI_get_rep_internal ("GI_get_representation_rep_code", rep_id);
  if (rep == NULL)
    return (-1);
  else
    return (rep->rep_code);
}
     

/*   Function Header   -----------------------------
     Name:             GI_set_representation_rep_code
     Description:      Rep_code modifier for representation.
     Parameters:       rep_id:   representation ID.
                       rep_code: Must be one of the representations defined
		                 in geodata.h.
     Preconditions     The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The rep_code of the representation is set to the new one
                       supplied.
     Return Value:     (None)
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  If no errors, change the rep_code to the one supplied.
     End Header   ---------------------------------- */

void GI_set_representation_rep_code (rep_id, rep_code)
     GB_id rep_id;
     unsigned rep_code;
{
  GS_representation_t rep;

  rep = GI_get_rep_internal ("GI_set_representation_rep_code", rep_id);
  if (rep != NULL)
    rep->rep_code = rep_code;
}


/*   Function Header   -----------------------------
     Name:             GI_get_representation_translucency
     Description:      Translucency accessor for representation.
     Parameters:       rep_id: representation ID.
     Preconditions     The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     The representation translucency.
                       -1 if there is an error.
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  Check for error and return translucency or -1 accordingly
     End Header   ---------------------------------- */

int GI_get_representation_translucency (rep_id)
     GB_id rep_id;
{
  GS_representation_t rep;

  rep = GI_get_rep_internal ("GI_get_representation_translucency", rep_id);
  if (rep == NULL)
    return (-1);
  else
    return (rep->translucency);
}


/*   Function Header   -----------------------------
     Name:             GI_set_representation_translucency
     Description:      Translucency modifier for representation.
     Parameters:       rep_id:   representation ID.
                       translucency: Must be one of those defined
		                     in geodata.h.
     Preconditions     The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The translucency of the representation is set to
                       the new one supplied.
     Return Value:     (None)
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  If no errors, change the translucency to the one supplied.
     End Header   ---------------------------------- */

void GI_set_representation_translucency (rep_id, translucency)
     GB_id rep_id;
     unsigned translucency;
{
  GS_representation_t rep;

  rep = GI_get_rep_internal ("GI_set_representation_translucency", rep_id);
  if (rep != NULL)
    rep->translucency = translucency;
}


/*   Function Header   -----------------------------
     Name:             GI_get_representation_what_varies
     Description:      What_Varies accessor for representation.
     Parameters:       rep_id: representation ID.
     Preconditions     The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     The representation what_varies.
                       -1 if there is an error.
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  Check for error and return what_varies or -1 accordingly
     End Header   ---------------------------------- */

int GI_get_representation_what_varies (rep_id)
     GB_id rep_id;
{
  GS_representation_t rep;

  rep = GI_get_rep_internal ("GI_get_representation_what_varies", rep_id);
  if (rep == NULL)
    return (-1);
  else
    return (rep->what_varies);
}


/*   Function Header   -----------------------------
     Name:             GI_set_representation_what_varies
     Description:      What_Varies modifier for representation.
     Parameters:       rep_id:      representation ID.
                       what_varies: Must be one of those defined
		                     in geodata.h.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The what_varies of the representation is set to
                       the new one supplied.
     Return Value:     (None)
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  If no errors, change the what_varies to the one supplied.
     End Header   ---------------------------------- */

void GI_set_representation_what_varies (rep_id, what_varies)
     GB_id rep_id;
     unsigned what_varies;
{
  GS_representation_t rep;

  rep = GI_get_rep_internal ("GI_set_representation_what_varies", rep_id);
  if (rep != NULL)
    rep->what_varies = what_varies;
}


/*   Function Header   -----------------------------
     Name:             GI_get_table_t
     Description:      Get the table entry corrisponding to value.
     Parameters:       place:       Name of calling function to be passed
                                    through for errors.
                       rep_id:      representation ID.
		       value:       Value to lookup.
		       previous:    Address of a pointer to a GS_table_t.
		       use_epsilon: If this is TRUE, allow for floating point
		                    round off errors by using a tolerance of
				    GI_EPSILON.  If FALSE, do exact testing.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   previous is pointing to the pointer of the table entry
                       holding value.
     Return Value:     NULL if error or value is less than the first table val.
                       Else a pointer to the table entry holding value.
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  If value is less than the first value, set previous to the address
           of the table.  Return NULL.
       3.  Walk down the table entries until value is in the range.  Set
           previous to the address of the pointer pointing to the table entry,
	   Return the table entry.
     End Header   ---------------------------------- */

GS_table_t GI_get_table_t (place, rep_id, value, previous, use_epsilon)
     char *place;
     GB_id rep_id;
     float value;
     GS_table_t **previous;
     int use_epsilon;
{
  GS_representation_t rep;
  GS_table_t table, hat;
  float EPSILON = 0.0;

  if (use_epsilon) EPSILON = GI_EPSILON;
  rep = GI_get_rep_internal (place, rep_id);
  if (rep == NULL)
    return (NULL);
  else {
    table = rep->tables;
    *previous = &(rep->tables);
    if (value < (table->value - EPSILON))
      table = NULL;
    else
      while ((table->next != NULL) &&
	     (value > ((table->next)->value - EPSILON))) {
	*previous = &(table->next);
	table = table->next;
      }
    return (table);
  }
}


/*   Function Header   -----------------------------
     Name:             GI_get_table_values
     Description:      Get the values in a representations table.
     Parameters:       rep_id: representation ID.
                       values: array of floats.  Used for return values.
		       size:   Size of the array.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The array values holds all the values in the table.
     Return Value:      -1: Error getting representation.
                         0: Cannot fit all values into the array.
		       int: Number of valid values in the table.
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  Copy table values into the given array.
       3.  Return the number of values copied into the array,
           0 if they do not all fit.
     End Header   ---------------------------------- */

int GI_get_table_values (rep_id, values, size)
     GB_id rep_id;
     float values[];
     int size;
{
  GS_representation_t rep;
  GS_table_t table;
  int count;

  rep = GI_get_rep_internal ("GI_get_table_values", rep_id);
  if (rep == NULL)
    return (-1);
  else {
    table = rep->tables;
    count = 0;
    while ((table != NULL) && (count < size)) {
      values[count] = table->value;
      table = table->next;
      count++;
    }
    if (table == NULL)
      return (count);
    else {
      GB_message (GB_NOTIFY, "GI_get_table_values",
		  "Cannot fit all values into the given array.");
      return (0);
    }
  }
}


/*   Function Header   -----------------------------
     Name:             GI_add_table_value
     Description:      Add a new value to the table.
     Parameters:       rep_id: representation ID.
                       value:  Float, the new value to add.
		       color:  Array of int.  R, G, B values for the entry.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   A new table entry with the appropriate value and color
                       is created.  Rotation and Scale default to 0.0.
     Return Value:     (None)
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  Get the table entry whose range the value would normally fall in.
       3.  Create a new table entry and insert in the list.
       4.  Set the value and color of the new entry.  Default Scale and
           Rotation to 0.0.
     End Header   ---------------------------------- */

void GI_add_table_value (rep_id, value, color)
     GB_id rep_id;
     float value;
     int color[3];
{
  GS_table_t table;
  GS_table_t *hat;

  table = GB_create_table ();
  if (GI_get_table_t ("GI_add_table_value", rep_id, value, &hat, FALSE)) {
    table->next = (*hat)->next;
    (*hat)->next = table;
  }
  else {
    table->next = (*hat);
    (*hat) = table;
  }    
  ASSIGN_TABLE(table, value, color[0], color[1], color[2], 0.0, 0.0);
}


/*   Function Header   -----------------------------
     Name:             GI_delete_table_value
     Description:      Remove a table entry.
     Parameters:       rep_id: representation ID.
                       value: Float, value identifying range to delete.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The table entry whose range contains value is deleted.
     Return Value:     (None)
     Logic:
       1.  Retrieve the representation reporting errors.
       2.  Report Error if only one value remaining.
       3.  Notify user if value is smaller than any in the list.
       4.  Call GI_get_table_t and delete that table entry.
     End Header   ---------------------------------- */

void GI_delete_table_value (rep_id, value)
     GB_id rep_id;
     float value;
{
  GS_representation_t rep;
  GS_table_t table;
  GS_table_t *hat;

  rep = GI_get_rep_internal ("GI_delete_table_value", rep_id);
  if (rep == NULL)
    return;
  else if (rep->tables->next == NULL) {
    GB_message (GB_ERROR, "GI_delete_table_value",
		"Cannot Delete last remaining table value.");
    return;
  }
  if (!(table = GI_get_table_t ("GI_delete_table_value",
				rep_id, value, &hat, TRUE))) {
    GB_message (GB_NOTIFY, "GI_delete_table_value",
		"Deleting smallest value by default.");
    table = *hat;
  }
  *hat = table->next;
  free (table);
}


/*   Function Header   -----------------------------
     Name:             GI_get_table_color
     Description:      Color accessor for table values.
     Parameters:       rep_id: representation ID.
                       value:  Float, get color for this value.
		       color:  Array of int.  R, G, B values.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The color array will contain R, G, B values for the
                       table entry.
     Return Value:     (None)
     Logic:
       1.  Retrieve the representation and table entry reporting errors.
       2.  If no errors, set the color array to R, G, B values.
     End Header   ---------------------------------- */

void GI_get_table_color (rep_id, value, color)
     GB_id rep_id;
     float value;
     int color[3];
{
  GS_table_t table;
  GS_table_t *hat;

  table = GI_get_table_t ("GI_get_table_color", rep_id, value, &hat, TRUE);
  if (!table)
    table = *hat;
  color[0] = table->color[0];
  color[1] = table->color[1];
  color[2] = table->color[2];
}


/*   Function Header   -----------------------------
     Name:             GI_set_table_color
     Description:      Color modifier for table values.
     Parameters:       rep_id: representation ID.
                       value:  Float.  Get color for this value.
		       color:  Array of int.  Containing R, G, B values.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   The R, G, B values for the table entry are changed to
                       the ones given.
     Return Value:     (None)
     Logic:
       1.  Retrieve the representation and table entry reporting errors.
       2.  If no errors, set the color values of the table entry to the array
           R, G, B values.
     End Header   ---------------------------------- */

void GI_set_table_color (rep_id, value, color)
     GB_id rep_id;
     float value;
     int color[3];
{
  GS_table_t table;
  GS_table_t *hat;

  table = GI_get_table_t ("GI_set_table_color", rep_id, value, &hat, TRUE);
  if (!table)
    table = *hat;
  table->color[0] = color[0];
  table->color[1] = color[1];
  table->color[2] = color[2];
}


/*   Function Header   -----------------------------
     Name:             GI_get_table_discrete
     Description:      Discrete accessor for table values.
     Parameters:       rep_id: representation ID.
                       value:  Float.  Get discrete for this value.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     GB_id, id of discrete symbol, line style, or pattern.
     Logic:
       1.  Retrieve the representation and table entry reporting errors.
       2.  If no errors return discrete field of table entry.
     End Header   ---------------------------------- */

GB_id GI_get_table_discrete (rep_id, value)
     GB_id rep_id;
     float value;
{
  GS_table_t table;
  GS_table_t *hat;

  table = GI_get_table_t ("GI_get_table_discrete", rep_id, value, &hat, TRUE);
  if (!table)
    table = *hat;
  return (table->discrete);
}


/*   Function Header   -----------------------------
     Name:             GI_set_table_discrete
     Description:      Discrete modifier for table values.
     Parameters:       rep_id: representation ID.
                       value:  Float.  Set discrete for this value.
		       discrete_id: ID of a symbol, line style, or pattern.
     Preconditions:    The discrete_id must be a valid id of a symbol, line
                       style, or pattern.
                       The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   Discrete field of the table entry is set to the discrete
                       id given.
     Return Value:     (None)
     Logic:
       1.  Retrieve the representation and table entry reporting errors.
       2.  If no errors set discrete field of table entry to id given.
     End Header   ---------------------------------- */

void GI_set_table_discrete (rep_id, value, discrete_id)
     GB_id rep_id, discrete_id;
     float value;
{
  GS_table_t table;
  GS_table_t *hat;

  table = GI_get_table_t ("GI_set_table_discrete", rep_id, value, &hat, TRUE);
  if (!table)
    table = *hat;
  table->discrete = discrete_id;
}


/*   Function Header   -----------------------------
     Name:             GI_get_table_scale
     Description:      Scale accessor for table values.
     Parameters:       rep_id: representation ID.
                       value:  Float.  Get scale for this value.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     Float.  Scale value for the table entry.
     Logic:
       1.  Retrieve the representation and table entry reporting errors.
       2.  If no errors return scale of table entry.
     End Header   ---------------------------------- */

float GI_get_table_scale (rep_id, value)
     GB_id rep_id;
     float value;
{
  GS_table_t table;
  GS_table_t *hat;

  table = GI_get_table_t ("GI_get_table_scale", rep_id, value, &hat, TRUE);
  if (!table)
    table = *hat;
  return (table->scale);
}


/*   Function Header   -----------------------------
     Name:             GI_set_table_scale
     Description:      Scale modifier for table values.
     Parameters:       rep_id: representation ID.
                       value:  Float.  Set scale for this value.
		       scale:  Float. > 0.0,  New scale factor.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   Scale of table entry is set to scale given.
     Return Value:     (None)
     Logic:
       1.  Retrieve the representation and table entry reporting errors.
       2.  If no errors set scale of table entry.
     End Header   ---------------------------------- */

void GI_set_table_scale (rep_id, value, scale)
     GB_id rep_id;
     float value;
     float scale;
{
  GS_table_t table;
  GS_table_t *hat;
  char message[100];

  if (scale < 0.0) {
    sprintf(message, "New scale %f, cannot be smaller than 0.0", scale);
    GB_message (GB_ERROR, "GI_set_table_scale", message);
  }
  else {
    table = GI_get_table_t ("GI_set_table_scale", rep_id, value, &hat, TRUE);
    if (!table)
      table = *hat;
    table->scale = scale;
  }
}


/*   Function Header   -----------------------------
     Name:             GI_get_table_rotation
     Description:      Rotation accessor for table values.
     Parameters:       rep_id: representation ID.
                       value:  Float.  Get rotation for this value.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   (None)
     Return Value:     Float.  Rotation value for the table entry.
     Logic:
       1.  Retrieve the representation and table entry reporting errors.
       2.  If no errors return rotation of table entry.
     End Header   ---------------------------------- */

float GI_get_table_rotation (rep_id, value)
     GB_id rep_id;
     float value;
{
  GS_table_t table;
  GS_table_t *hat;

  table = GI_get_table_t ("GI_get_table_rotation", rep_id, value, &hat, TRUE);
  if (!table)
    table = *hat;
  return (table->rotation);
}


/*   Function Header   -----------------------------
     Name:             GI_set_table_rotation
     Description:      Rotation modifier for table values.
     Parameters:       rep_id: representation ID.
                       value:  Float.  Set rotation for this value.
		       rotation:  Float.  New rotation factor.
     Preconditions:    The rep_id must be a valid id of a representation.
                       The GeoData software must have been initialized prior
                       to the call.
     Postconditions:   Rotation of table entry is set to rotation given.
     Return Value:     (None)
     Logic:
       1.  Retrieve the representation and table entry reporting errors.
       2.  If no errors set rotation of table entry.
     End Header   ---------------------------------- */

void GI_set_table_rotation (rep_id, value, rotation)
     GB_id rep_id;
     float value;
     float rotation;
{
  GS_table_t table;
  GS_table_t *hat;

  table = GI_get_table_t ("GI_set_table_rotation", rep_id, value, &hat, TRUE);
  if (!table)
    table = *hat;
  table->rotation = rotation;
}
