/*   File Header   ---------------------------------

     File Name:       list.c

     Author and Date: George Lamb--3/19/90
                      reworked--5/1/90

     Description: The list routines needed to maintain a unified
geodata data base.  The data base is organized as an open hash list.
When a hash table slot is full, any value hashing to that slot is
appended as a list element.  This allows simple deletion, yet has the
benefits of a hash table for lookup speed.  The functions available to
manipulate the table are:

         int       *GB_add_object (label, type)
	 int        GB_delete_object (id)
	 int       *GB_get_object (id)
	 GB_entry  *GB_get_entry (id)
	 GB_id      GB_get_id (label, type)
	 void       GB_get_label (id, label)
         void       GB_print_hash_table ()
         void       GB_describe_objects (stream)
	 void       GB_describe_object (id, stream)
	 int        GB_current_time()
	 void       GB_init ()

	 GB_entry  *new_object ()
         GB_id      new_id (label)
         key        hash_label (label)
	 int        hash_id (id)

All functions return NULL for errors, and print messages to stderr.

     End Header   ---------------------------------- */


#include <geodata.h>
#include <geodata_private.h>
#include <stdio.h>

typedef unsigned short key_type;

extern int *GB_create_instance();
extern void GB_delete_instance();


GB_entry GB_hash_table[GB_HASH_TABLE_SIZE];
static key_type UNIQUIFIER = 0;
static int current_time = 0;
static char note[128];   /* for GB_messages */

int GB_current_time() {  return(current_time); }


/*   Function Header   -----------------------------
     NAME:        hash_label
     DESCRIPTION: compute a GB_id from a string using a 
     randomizing hash function
     PARAMETERS:  a string
     PRECON:      none
     POSTCON:     none
     RETURN:      the hashed id
     LOGIC: the function iterates through the string characters,
generating an ID.  The first two characters are stored in the first
and last 8 bits of the ID, respectively.  Sucessive characters get
xor'd at offsets based on their string position, either 0, 4, or 8
bits in.  These new ID values then get rotated 3 bits, to get the
lower bits filled.  Otherwise, unless labels used the lower ascii
values, the full range would be lost.
     End Header   ---------------------------------- */

key_type hash_label (label)
     char label[];
{
  key_type number, new;  
  int i, length = MIN(strlen(label), GB_LABEL_LENGTH);

  number = label[0];
  if (length == 1) return(number);
  number = (number << 8) + label[1];
  for (i=2; i < length; i++) {     
    new = label[i];
    number = number ^ (new << ((i%3)*4));
    number = (number << 3) + (number >> 13);
    if (number == 0) number = 1;  /* should always fail, but better safe! */
  }
  return(number);
}   /* end of hash_label */



/*   Function Header   -----------------------------
     NAME:        hash_id
     DESCRIPTION: generate a hash table bucket index from a data id
     PARAMETERS:  a data id
     PRECOND:     none
     POSTCOND:    none
     RETURN:      the hash table bucket for the id
     LOGIC:       The key for a hash entry is stored in the lower 32 bits of
     the id.  This function selects those bits, and mods by the table size,
     to find the correct bucket for an id.
     End Header   ---------------------------------- */
int hash_id (id)
     GB_id id;
{
  key_type key = id << 15 >> 15;
  return(key % GB_HASH_TABLE_SIZE);
}



/*   Function Header   -----------------------------
     NAME:        GB_get_id
     DESCRIPTION: find the id of the given label/type
     PARAMETERS:  label, type
     PRECOND:     the hash table must be initialized
     POSTCOND:    none
     RETURN:      the id associated with the label/type
     LOGIC: id's are uniquely assigned to data labels, st. a key is
     stored in the lower 16 bits and a uniquifier is stored in the upper 16
     bits.  To retrieve the id for a given label, the label is hashed to
     get the possibly un-unique key.  That key is hashed to find the bucket
     the data lies in.  The string is then compared to every string in the
     bucket, and the unique id for the string is returned.
     End Header ---------------------------------- */
GB_id GB_get_id (label)
     char *label;
{
  key_type key;
  int index;
  GB_entry *hat;

  current_time++;
  key = hash_label (label);
  index = hash_id (key);
  if (GB_hash_table[index]. id == 0) return (NULL);
  if (strncmp(GB_hash_table[index].label, label, GB_LABEL_LENGTH - 1) == 0)
    return (GB_hash_table[index].id);
  hat = GB_hash_table[index].next;
  while (hat != NULL){
    if (strncmp(hat->label, label,GB_LABEL_LENGTH-1) == 0)
      return (hat->id);
    hat = hat->next;
  }
  return(NULL);
}    /* end of GB_get_id */



/*   Function Header   -----------------------------
     NAME:        GB_get_entry
     DESCRIPTION: retrieve the entry with the given id from the hash table
     PARAMETERS:  an id
     PRECOND:     the hash table must be initialized
     POSTCOND:    the retriever has a pointer to a GB entry.
     RETURN:      a GB entry ptr, NULL for error
     LOGIC:       first, the given id is unhashed to find which bucket the
     data lies in.  If the bucket is empty, NULL is returned.  If the id is
     in the first bucket, the address of that bucket is returned.  If
     not, the next entries in the bucket are searched linearly, until the
     desired id is found or the end of the list is reached, at which point
     NULL is returned.  
     End Header ---------------------------------- */

GB_entry *GB_get_entry (id)
     GB_id id;
{
  int index = hash_id (id);
  GB_entry *hat;

  current_time++;
  if (GB_hash_table[index]. id == 0) return (NULL);
  if (GB_hash_table[index]. id == id) return (&GB_hash_table[index]);
  hat = GB_hash_table[index].next;
  while (hat != NULL){
    if (hat->id == id) return(hat);
    hat = hat->next;
  }
  return(NULL);
}    /* end of GB_get_entry */



/*   Function Header   -----------------------------
     NAME:        GB_get_object
     DESCRIPTION: retrieve the instance of the given id from the table
     PARAMETERS:  an id
     PRECOND:     the hash table must be initialized
     POSTCOND:    the retriever has a pointer to a GB entry 
     description. With this pointer he can't get at the actual table 
     which is good.
     RETURN:      a GB entry instance ptr, NULL for error
     LOGIC:       call GB_get_entry and return the entries instance
     NULL is returned.  
     End Header ---------------------------------- */

int *GB_get_object (id)
     GB_id id;
{
  GB_entry *hat = GB_get_entry (id);
  current_time++;
  if (hat == NULL) return (NULL);
  else return(hat->instance);
}    /* end of GB_get_object */



/*   Function Header   -----------------------------
     NAME:        GB_get_label
     DESCRIPTION: retrieve the label of the given id from the table
     PARAMETERS:  an id
     PRECOND:     the hash table must be initialized
     POSTCOND:    the user has a ptr to the given objects label
     RETURN:      void
     LOGIC:       call GB_get_entry and return the entries label
     End Header ---------------------------------- */

char *GB_get_label (id)
     GB_id id;
{
  GB_entry *hat = GB_get_entry (id);
  current_time++;
  if (hat == NULL) return (NULL);
  else return(hat->label);
}    /* end of GB_get_label */


/*   Function Header   -----------------------------
     NAME:        GB_get_object_from_label
     DESCRIPTION: retrieve the given label'd desc from the hash table
     PARAMETERS:  a label and type
     PRECOND:     the hash table must be initialized
     POSTCOND:    none
     RETURN:      a GB_entry instance ptr, NULL for not entered
     LOGIC: id's are uniquely assigned to data labels, st. a key is
     stored in the lower 16 bits and a uniquifier is stored in the upper 16
     bits.  To retrieve the id for a given label, the label is hashed to
     get the possibly un-unique key.  That key is hashed to find the bucket
     the data lies in.  The string is then compared to every string in the
     bucket, and the descriptor assigned to that label is returned.
     End Header ---------------------------------- */

int *GB_get_object_from_label (label )
     char *label;
{
  key_type key;
  int index;
  GB_entry *hat;

  key = hash_label (label);
  index = hash_id (key);
  current_time++;
  if (GB_hash_table[index]. id == 0) return (NULL);
  if ((!strncmp(GB_hash_table[index].label, label, GB_LABEL_LENGTH)))
    return (GB_hash_table[index].instance);
  hat = GB_hash_table[index].next;
  while (hat != NULL){
    if ((!strncmp(hat->label, label, GB_LABEL_LENGTH)))
      return (hat->instance);
    hat = hat->next;
  }
  return(NULL);
}    /* end of GB_get_object_from_label */



/*   Function Header   -----------------------------
     NAME:        GB_print_hash_table
     DESCRIPTION: run through the table printing every entry
     PARAMETERS:  none
     PRECOND:     the hash table must be initialized
     POSTCOND:    a representation of the table is printed on stdout
     RETURN:      none
     LOGIC:       buckets are searched in order.  Anytime a full bucket is
     found the list headed there is printed in order.
     End Header   ---------------------------------- */

void GB_print_hash_table()
{
  int i;
  GB_entry *hat;
  char const_string[GB_LABEL_LENGTH];

  fprintf(stdout, "\n\n;GeoData Hash Table:\n\n");
  
  for (i=0; i< GB_HASH_TABLE_SIZE; i++)
    if (GB_hash_table[i].id != 0){
/*      GB_encode_constant(const_string, GB_hash_table[i].type);*/
      fprintf(stdout, "   bucket %d:\t%-20s\t%-20s\t%d.\n", i, 
	      GB_hash_table[i].label, const_string, GB_hash_table[i].id);
      hat = GB_hash_table[i].next;
      while (hat != NULL){
/*	GB_encode_constant(const_string, hat->type);*/
	fprintf(stdout, "\t\t%-20s\t%-20s\t%d.\n", 
		hat->label, const_string, hat->id);
	hat = hat->next;
      }      
    }
  fprintf(stdout, "\n;End of GeoData Hash Table\n");
}   /* end of GB_print_hash_table */


/*   Function Header   -----------------------------
     NAME:        new_id
     DESCRIPTION: generate a unique id from the given label
     PARAMETERS:  label
     PRECOND:     the hash table must be initialized
     POSTCOND:    a valid unused GB_id is returned
     RETURN:      GB_id, NULL for error
     LOGIC: A GB_id is a 32 bit unique identifier composed of a 16 bit
     key generated from the given string and appended with a 16 bit value,
     which represents the number of the data set being registered.  The key
     can be generated quickly and used to find the hash table bucket where
     the set is stored.  The counter suffices to guarantee uniqeu id's.
     End Header   ---------------------------------- */

GB_id new_id (label)
     char *label;
{
  key_type key = hash_label(label);
  GB_id id = UNIQUIFIER++;
  return((id << 16) + key);
}   /* end of new_id */



/*   Function Header   -----------------------------
     NAME:        GB_describe_object
     DESCRIPTION: print a single object
     PARAMETERS:  id, stream, object flag
     PRECOND:     the hash table must be initialized
     POSTCOND:    a description of the object is printed onto stdout
     RETURN:      void
     LOGIC: find and print the id, label, and type, then call
     write_instance with the instance to a local string and print that onto
     stdout.
     End Header   ---------------------------------- */

void GB_describe_object (id, f, obj_info)
     GB_id id;
     FILE *f;
     int obj_info;
{
  GB_entry *cleo = GB_get_entry(id);
  char const_string[GB_LABEL_LENGTH];

  if (f == NULL) f = stdout;
  current_time++;
  if (cleo != NULL){
    if(obj_info) fprintf (f, "OBJECT #%d, Instance #%d\n", id,cleo->instance);
/*    GB_print_object(&stream, cleo->type, cleo->label); */
/*    fprintf("(%d %s\n",cleo->type,cleo->label); no file pointer eah*/
    GB_write_instance(f, cleo->instance, cleo->type,GB_get_label(id),0);
/*    fprintf(") ; %s\n",cleo->label); ditto*/
  }
  else {sprintf(note, "object %d undefined.", id);
	GB_message(GB_ERROR, "GB_describe_object", note);}
}   /* end of GB_describe_object */


/*   Function Header   -----------------------------
     NAME:        GB_describe_objects
     DESCRIPTION: Iterate through the table, describing every entry
     PARAMETERS:  stream where output is to go, object info flag
     PRECONDITIONS:     the hash table must be initialized
     POSTCONDITIONS:    No change to data structures
     RETURN:      none
     LOGIC:       Buckets are processed in order.  When a nonempty bucket is
                  found, the list headed there is printed in order.
     End Header   ---------------------------------- */

void GB_describe_objects(s, obj_info)
  FILE *s;  int obj_info;
{
  int i;
  GB_entry *hat;
  char const_string[GB_LABEL_LENGTH];

  fprintf(s,"\n\n;GeoData Data Structures:\n\n");
  
  for (i=0; i< GB_HASH_TABLE_SIZE; i++)
    if (GB_hash_table[i].id != 0){
      hat = &GB_hash_table[i];
      while (hat != NULL){
        GB_describe_object(hat->id,s,obj_info); 
	hat = hat->next;
      }      
    }
  fprintf(s,"\n;End of GeoData Data Structures\n");
}   /* end of GB_describe_objects */



/*   Function Header   -----------------------------
     NAME:        GB_recur_describe_object
     DESCRIPTION: print an object and its dependents
     PARAMETERS:  id, stream, object flag
     PRECOND:     the hash table must be initialized
     POSTCOND:    a description of the object and its dependents is printed onto stdout
     RETURN:      void
     LOGIC: find and print the id, label, and type, then call
     recur_write_instance with the instance to a local string and print that onto
     stdout.
     End Header   ---------------------------------- */

void GB_recur_describe_object (id, f, obj_info)
     GB_id id;
     FILE *f;
     int obj_info;
{
  GB_entry *cleo = GB_get_entry(id);
  char const_string[GB_LABEL_LENGTH];

  if (f == NULL) f = stdout;
  current_time++;
  if (cleo != NULL){
    if(obj_info) fprintf (f, "OBJECT #%d, Instance #%d\n", id,cleo->instance);
    GB_recur_write_instance(f, cleo->instance, cleo->type,GB_get_label(id),0);
  }
  else {sprintf(note, "object %d undefined.", id);
	GB_message(GB_ERROR, "GB_recur_describe_object", note);}
}   /* end of GB_recur_describe_object */


/*   Function Header   -----------------------------
     NAME:        GB_add_object
     DESCRIPTION: add an object to the table
     PARAMETERS:  label type
     PRECOND:     the hash table must be initialized
     POSTCOND:    the label has an entry in the hash table, described as given
     RETURN:      a ptr to the label's assigned data desc, no warning for 
     pre-loaded data!
     LOGIC: an id is generated, and the proper bucket for the label is
     searched.  If the bucket is empty, a ddl entry is created, and the
     label inserted.  If it's full, the label is searched for through the
     buckets, and if found, that descriptor is returned.  If not found, a
     new bucket is created at the end of the hash list at that bucket.
     End Header   ---------------------------------- */

int *GB_add_object (label, type)
     char *label;
     unsigned type;
{
  GB_id id;
  int index, i=0;
  GB_entry *hat, *cleo;

  current_time++;
  id = new_id (label);
  index = hash_id (id);
  if (GB_hash_table[index].id == 0) {   /* empty bucket */
    GB_hash_table[index].id = id;
    GB_hash_table[index].type = type;
    strncpy(GB_hash_table[index].label, label, GB_LABEL_LENGTH);
    GB_hash_table[index].label[GB_LABEL_LENGTH - 1] = '\0';
    GB_hash_table[index].next = NULL;    
    GB_hash_table[index].instance = GB_create_instance(type);
    return (GB_hash_table[index].instance);
  }
  hat = &GB_hash_table[index];
  while (hat->next != NULL){
    if ((!strncmp(label, hat->label, GB_LABEL_LENGTH)) && (hat->type == type)){
      /* the object is already used in the table! */
/*      hat->instance = GB_create_instance(type);  ** Don't do this-- PBA  */
      return(hat->instance);
    }
    hat = hat->next;
  }
  if ((!strncmp(label, hat->label, GB_LABEL_LENGTH)) && (hat->type == type)){
    /* the object is already used in the table! */
    return(hat->instance);
  }
  if ((cleo = (GB_entry *) malloc(sizeof(GB_entry))) == NULL){
    sprintf(note, "no space for the new object %s.", label);
    GB_message(GB_ERROR, "GB_add_object", note);
    return(NULL);
  }
  hat->next = cleo;
  cleo->id = id;
  cleo->type = type;
  cleo->instance = GB_create_instance(type);
  strncpy(cleo->label, label, GB_LABEL_LENGTH);
  cleo->label[GB_LABEL_LENGTH - 1] = '\0';
  cleo->next = NULL;
  return(cleo->instance);
}    /* end of GB_add_object */



/*   Function Header   -----------------------------
     NAME:        GB_delete_object
     DESCRIPTION: delete the ddl id from the hash list table
     PARAMETERTS: id
     PRECOND:     the hash table must be initialized
     POSTCOND:    if the id is in the list, it is deleted, and it's space 
     reclaimed
     RETURN:      1 for sucesss, NULL for not in the list
     LOGIC:       if the id is the only entry in the bucket, delete it by
     setting it's id to 0.  If the id is the first entry, copy the last
     entry of the list into that slot and delete the last entry.  If the
     entry is in the middle of the list, simply delete it.
     End Header   ---------------------------------- */

int GB_delete_object (id)
     GB_id id;
{
  int index;
  GB_entry *hat = NULL, *cleo = NULL;
  current_time++;
  index = hash_id (id);
  if (GB_hash_table [index].id == id){   /* first bucket */
    if (GB_hash_table[index].next == NULL){  /* only bucket */
      GB_hash_table[index].id = 0;
      GB_delete_instance(GB_hash_table[index].instance, GB_hash_table[index].type);
      return(1);
    }
    else /* move last in list to head */ {
      hat = &GB_hash_table[index];
      while (hat->next->next != NULL) hat = hat->next;
      cleo = hat->next;
      GB_delete_instance(GB_hash_table[index].instance, GB_hash_table[index].type);
      GB_hash_table[index].id = cleo->id;
      GB_hash_table[index].type = cleo->type;
      strcpy(GB_hash_table[index].label, cleo->label);
      GB_hash_table[index].instance = cleo->instance;
      free(cleo);
      hat->next = NULL;
      return (1);
    }
  }
  if (GB_hash_table[index].next == NULL) return(NULL); /* not first & no list */
  /* else, maybe somewhere in list, find previous bucket */
  hat = &GB_hash_table[index];
  while (hat->next->id != id) {
    if (hat->next->next == NULL) return(NULL);  /* not in list */
    hat = hat->next;
  }
  cleo = hat->next;
  hat->next = cleo->next;
/*  GB_delete_instance(GB_hash_table[index].instance, GB_hash_table[index].type);   PBA   */
  GB_delete_instance(cleo->instance, cleo->type);
  free(cleo);
  return (1);    
}    /* end of GB_delete_object */
  


/*   Function Header   -----------------------------
     NAME:        GB_init
     DESCRIPTION: initialize the table by setting every table id to 0.
     PARAMETERS:  none
     PRECOND:     none
     POSTCOND:    the hash table is initialized.  Any entries in it
     can no longer be retrieved.
     RETURN:      none
     LOGIC:       obvious.
     End Header   ---------------------------------- */

void GB_init ()
{
  int i;
  for (i=0; i< GB_HASH_TABLE_SIZE; i++)
    GB_hash_table[i].id = 0;
}    /* end of GB_init */

