/******************************************************************************
*                                                                             *
* bb.c - Generates broadband correlograms for a set of time series data       *
*          The order of processing is                                         *
*                                                                             *
*           1. Read raw time-series data from DataVault                       *
*           2. Retrieve data to be processed                                  *
*           3. Convert data to floating point                                 *
*           4. Window the data                                                *
*           5. FFT the data                                                   *
*           6. Beamform the data                                              *
*           7. SCOT window the spectrums                                      *
*           8. Bandpass filter the spectrums                                  *
*           9. Center the correlations                                        *
*          10. Cross multiple the spectrums                                   *
*          11. Inverse FFT to calculate correlations                          *
*          12. Normalize the correlations                                     *
*          13. Merge the correlations from a line array                       *
*          14. Discretize the correations                                     *
*          15. Store data to DataVault aand/or display on framebuffer         *
*                                                                             *
*  Running sequence is                                                        *
*                                                                             *
*       bb in out bbparam param cal                                           *
*                                                                             *
*   where                                                                     *
*                                                                             *
*        in     -  DataVault file containing time series data                 *
*        out    -  DataVault file to contain broadband correlation            *
*        bbparam - Front end file containing parameters for bb run            *
*        param   - Front end file containing array description                *
*        cal     - DataVault file containing sensor calibrations              *
*                                                                             *
*                                                                             *
******************************************************************************/

#include <sys/types.h>
#include <cm/cm.h>
#include <cm/cm_stat.h>
#include <cm/cm_file.h>
#include <cm/cmfs.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <cm/paris.h>
#include <cm/cmssl-paris.h>
#define TRUE          1
#define FALSE         0
#define DVFILE int

void main(argc, argv)
     int argc;
     char **argv;
{
/******************************************************************************
*                                                                             *
*   Set-up variables                                                          *
*                                                                             *
******************************************************************************/
  char dv_in_file_name[128], dv_out_file_name[128], dv_gfilt_file_name[128];
  char param_file_name[128],beam_file_name[128],dv_cal_file_name[128];
  DVFILE dv_infile, dv_outfile, dv_gfiltfile;
  u_char *grid;
  int oddsteps,number;
  int aver,save_aver;
  int p,pts, steps, ptr, step, outfile_flag = 0;
  int ARRAY,array,copy=4,field_idno=0;
  int sample_size=32;
  float dstep;
  int time;
  int show_fb = FALSE, take_hann = FALSE,spoke_supp=FALSE,
  take_scot= FALSE, decomp=FALSE,
  overlap =1, save_grid = TRUE, ser_data=FALSE;
  int update,size,zero_fill=FALSE;
  int CLUSTERS,BB_CORR,CHANNELS;
  int channels_per_array_1,channels_per_array_2,channels_per_array;
  int n_beams,sweeps_in,sweeps_out,cut;
  int scot_flag=0, take_gfilt=FALSE,int_data=FALSE;
  int sweep_count=0;
  int debu, dbu, bits;
  unsigned out_total;
  float te[128];
  float alpha=0.;
  float alpha_in;
  float xpts;
  float cnst;
  int nproc;
  int ncut;

  int time_all=0,time_in=1,time_bitrev=2,time_reformat=3,time_window=4,
      time_fft=5,time_unsfft=6,time_beam=7,time_spect=8,time_scr=9,
      time_ifft=10,time_norm=11,time_form=12,time_grey=13,time_trans=14,
      time_copy=15,time_out=16;

  struct cm_stat instat;

  unsigned  pure_data, old_data, gfilt, beam_table[128],cmte, window, fft_aver,
            fft_temp, bb_out, x, y, xamp, yamp, wx, wy, amplitude, 
            signed_amplitude, fft_amp, ali, beam_real, spoke_temp,
            xfl, yfl, xyfl, out_amp, bb_aver, gtemp, temp52;
  
  CM_geometry_id_t new_geometry,send_geometry,cut_geometry;
  unsigned dimension_array[1], new_vp_set;
  CMSSL_fft_setup_t id;
  CMSSL_fft_setup_t idinv;
  CMSSL_op_t ops[2]; 
  CMSSL_bit_order_t sbo[2],dbo[2];
  CMSSL_scale_t sca[2];
  CMSSL_op_t opscor[2]; 
  CMSSL_bit_order_t sbocor[2],dbocor[2];
  CMSSL_scale_t scacor[2];
  struct CM_axis_descriptor axe_desc[32];
  CM_axis_descriptor_t axe_desc_t[32];
  int i;
  int i1,i2,i3,a1,a2,a3;
  int num_filt;
  static char *digit[] = { "0","1","2","3","4","5","6","7","8","9" };
  static char out_file[80]=" this is a test";
  FILE *fopen(), *fp;
  int channel, odd, beam, table_index;
  int mv_offset;
  int offset =0;
  unsigned start[2],stop[2];
  start[0]=4096;stop[0]=4146;start[1]=0;stop[1]=0;


/******************************************************************************
*                                                                             *
*   Initialize CM and file names                                              *
*                                                                             *
******************************************************************************/

  CM_init ();
  CMSSL_init ();
  CM_set_safety_mode (0); 
  CM_reset_timer(0);

  dv_cal_file_name[0] = '\0';
  dv_out_file_name[0] = '\0';
  param_file_name[0] = '\0';
  beam_file_name[0] = '\0';
  dv_cal_file_name[0] = '\0';
  if (argc > 1) strcpy(dv_in_file_name, argv[1]);
  if (argc > 2) strcpy(dv_out_file_name, argv[2]);
  if (argc > 3) strcpy(param_file_name, argv[3]);
  if (argc > 4) strcpy(beam_file_name, argv[4]);
  if (argc > 5) strcpy(dv_cal_file_name, argv[5]);

/******************************************************************************
*                                                                             *
*   Get parameters from front end file                                        *
*                                                                             *
******************************************************************************/

  fp=fopen (param_file_name,"r");
  if (fp == NULL) { fprintf (stderr,"Couldn't open file \n"); exit(0);}
    fscanf (fp,"%d %*S \n",&p);
      pts=1<<p;
      printf ("%d \n",pts);
      update=pts/2;
      size=pts;
    fscanf (fp,"%d %*S \n",&ARRAY);
      printf ("%d \n",ARRAY);
    fscanf (fp,"%d %*S \n",&channels_per_array_1);
      printf ("%d \n",channels_per_array_1);
    fscanf (fp,"%d %*S \n",&channels_per_array_2);
      printf ("%d \n",channels_per_array_2);
      channels_per_array=channels_per_array_1+channels_per_array_2;
    fscanf (fp,"%d %*S \n",&n_beams);
      printf ("%d \n",n_beams);
    fscanf (fp,"%d %*S \n",&sweeps_in);
      printf ("%d \n",sweeps_in);
    fscanf (fp,"%d %*S \n",&sweeps_out);
      printf ("%d \n",sweeps_out);
    fscanf (fp,"%d %*S \n",&cut);
      printf ("%d \n",cut);
      ncut=1<<cut;
    fscanf (fp,"%d %*S \n",&steps);
      printf ("%d \n",steps);
    fscanf (fp,"%d %*S \n",&aver);
      printf ("%d \n",aver);
      save_aver = aver;
    fscanf (fp,"%d %*S \n",&int_data);
      printf ("%d \n",int_data);
    fscanf (fp,"%d %*S \n",&decomp);
      printf ("%d \n",decomp);
    fscanf (fp,"%d %*S \n",&sample_size);
      printf ("%d \n",sample_size);
    fscanf (fp,"%d %*S \n",&ser_data);
      printf ("%d \n",ser_data);
    fscanf (fp,"%d %*S \n",&take_hann);
      printf ("%d \n",take_hann);
    fscanf (fp,"%d %*S \n",&overlap);
      printf ("%d \n",overlap);
      update=update/overlap;
    fscanf (fp,"%d %S %*S %*S \n",&num_filt, &dv_gfilt_file_name[0]);
      printf ("%d %S \n",num_filt,dv_gfilt_file_name);
      take_gfilt=TRUE;
    fscanf (fp,"%d %*S \n",&take_scot);
      printf ("%d \n",take_scot);
    fscanf (fp,"%d %f %*S %*S \n",&spoke_supp, &alpha_in);
      printf ("%d %f \n",spoke_supp,alpha_in);
    fscanf (fp,"%d %*S \n",&time);
      printf ("%d \n",time);
    fscanf (fp,"%d %*S \n",&nproc);
      printf ("%d \n",nproc);

/******************************************************************************
*                                                                             *
*   set up FFT configuration and geometries on CM                             *
*                                                                             *
******************************************************************************/

  ops[0]=CMSSL_f_xform;
  sbo[0]=CMSSL_normal;
  dbo[0]=CMSSL_bit_reversed;
  sca[0]=CMSSL_noscale;
  ops[1]=CMSSL_nop;
  sbo[1]=CMSSL_normal;
  dbo[1]=CMSSL_normal;
  sca[1]=CMSSL_noscale;
  opscor[0]=CMSSL_i_xform;
  sbocor[0]=CMSSL_bit_reversed;
  dbocor[0]=CMSSL_normal;
  scacor[0]=CMSSL_scale_n;
  opscor[1]=CMSSL_nop;
  sbocor[1]=CMSSL_normal;
  dbocor[1]=CMSSL_normal;
  scacor[1]=CMSSL_noscale;
   for (i = 0; i < 2; i++) {
        axe_desc[i].weight = 1-i;
        axe_desc[i].ordering = CM_send_order;
        axe_desc[i].off_chip_bits = 0;
        axe_desc[i].off_chip_pos = 0; 
        axe_desc[i].on_chip_bits = 0;
        axe_desc[i].on_chip_pos = 0;
        axe_desc_t[i] = &(axe_desc[i]);
      }

  array=nproc/(pts/2);
  CLUSTERS=ARRAY/array;
  BB_CORR=(n_beams*CLUSTERS);
  CHANNELS=CLUSTERS*channels_per_array;
  out_total=CLUSTERS*num_filt/ncut;
  axe_desc[0].length = array*pts/2;
  send_geometry = CM_create_detailed_geometry((unsigned *)(&axe_desc_t[0]),1);
  axe_desc[1].length = array;
  axe_desc[0].length = pts/2;
  new_geometry = CM_create_detailed_geometry((unsigned *)(&axe_desc_t[0]),2);
  axe_desc[1].length = ncut*array;
  axe_desc[0].length = pts/2/ncut;
  cut_geometry = CM_create_detailed_geometry((unsigned *)(&axe_desc_t[0]),2);

  new_vp_set = CM_allocate_vp_set(new_geometry);
  CM_set_vp_set(new_vp_set);
  CM_set_context();

/******************************************************************************
*                                                                             *
*  Get bandpass filter from front-end file                                    *
*                                                                             *
******************************************************************************/

  gfilt=CM_allocate_stack_field(32*ncut*num_filt);
  if (take_gfilt) {
  for (i=0;i<num_filt;i++){
     strcpy (out_file,dv_gfilt_file_name);
     strcat (out_file,digit[i+1]);
     READCMFILE ( out_file,gfilt+i*32,32);
   }
    BITREV2D( gfilt,p-1,32*num_filt,0,new_geometry);  
    CM_set_vp_set_geometry(CM_current_vp_set, cut_geometry);
    TRUNCFREQB (gfilt,gfilt,ncut,32*num_filt/ncut,cut);
    CM_set_vp_set_geometry(CM_current_vp_set, new_geometry);
 }
    num_filt=num_filt/ncut;

 
/******************************************************************************
*                                                                             *
*  Allocate FCM memory                                                        *
*                                                                             *
******************************************************************************/

  fft_temp=CM_allocate_stack_field(64*CHANNELS);    
  id= CMSSL_c_c_fft_setup_1L (fft_temp,23,8);
  CM_set_vp_set_geometry(CM_current_vp_set, cut_geometry);
  fft_aver=CM_allocate_stack_field(64*BB_CORR*num_filt);
  bb_aver=CM_allocate_stack_field(64*BB_CORR*num_filt);
  idinv= CMSSL_c_c_fft_setup_1L (fft_aver,23,8);
  CM_set_vp_set_geometry(CM_current_vp_set, new_geometry);
    
  pure_data=CM_allocate_stack_field(sample_size*CHANNELS*2);
  old_data=CM_allocate_stack_field(sample_size*CHANNELS*2*(sweeps_in+1));
  cmte=CM_allocate_stack_field(4*32);
  window=CM_allocate_stack_field(32*2*2);
  amplitude=CM_allocate_stack_field(32);   
  signed_amplitude=CM_allocate_stack_field(8);   
  ali=CM_allocate_stack_field(32*CHANNELS);   
  for (i=0;i<n_beams;i++)
    beam_table[i]=CM_allocate_stack_field(64*channels_per_array);
  beam_real=CM_allocate_stack_field(64*n_beams*CLUSTERS*2);
  temp52=CM_allocate_stack_field(64);
  x=CM_allocate_stack_field(64*BB_CORR*num_filt);
  y=CM_allocate_stack_field(64*BB_CORR*num_filt);
  wx=CM_allocate_stack_field(32*BB_CORR*num_filt);
  wy=CM_allocate_stack_field(32*BB_CORR*num_filt);
  xamp=CM_allocate_stack_field(32*BB_CORR*num_filt);
  yamp=CM_allocate_stack_field(32*BB_CORR*num_filt);
  bb_out=CM_allocate_stack_field(64*BB_CORR*num_filt);
  spoke_temp=CM_allocate_stack_field(32*BB_CORR*num_filt);
  xfl=CM_allocate_stack_field(32*BB_CORR*num_filt);
  yfl=CM_allocate_stack_field(32*BB_CORR*num_filt);
  xyfl=CM_allocate_stack_field(32*BB_CORR*num_filt);
  fft_amp=CM_allocate_stack_field(32*CHANNELS*num_filt);
  out_amp=CM_allocate_stack_field(32*out_total*sweeps_out);
  gtemp=CM_allocate_stack_field(ncut*64);

/******************************************************************************
*                                                                             *
* Open DataVault files for input and output                                   *
*                                                                             *
******************************************************************************/

  if ((dv_infile = CMFS_open(dv_in_file_name, CMFS_O_RDONLY, 0644,
			     CM_physical_processors_limit,
			     CM_user_number_of_processors_limit)) == -1){
    printf("error on dv file open: %s\n", dv_in_file_name);
    CMFS_perror("");
    exit(1);
  }
  if (steps == 0) {
     steps=CMFS_fstat(dv_infile,&instat);
     steps=(instat.cmst_size/(sample_size/8)/pts/ARRAY/
              CHANNELS*(pts/(2*update)))-1;
     printf (" Number of output sweeps = %i \n",steps/save_aver);
   }
  printf (" file opened \n");

  if ((dv_outfile = CMFS_open(dv_out_file_name, CMFS_O_RDWR | CM_CREAT, 0644,
			      CM_physical_processors_limit,
			      CM_user_number_of_processors_limit)) == -1){
    printf("error on dv file open: %s\n", dv_out_file_name);
    
    exit(1);
  }
  printf (" file opened \n");
  

/******************************************************************************
*                                                                             *
*  Initialize tables, CM memory, and input buffers                            *
*                                                                             *
******************************************************************************/

    for (debu = 0;  debu < 128; debu++) te[debu] = 0;
    COMPUTE_DECOMP_TABLE(te);
    if (decomp) LOAD_DECOMP_TABLE(te, cmte);
    CM_u_move_zero_1L(amplitude,32);
    CM_u_move_zero_1L(fft_aver,32*CHANNELS);
    CM_u_move_zero_1L(ali,32*CHANNELS);
    CM_u_move_zero_1L(wx,32*BB_CORR*num_filt);
    CM_u_move_zero_1L(wy,32*BB_CORR*num_filt);
    HANN2(window, size);
    HAMM2(window+64,size);
    for (i=0;i<n_beams;i++)
      BEAMTABLE(beam_file_name,beam_table[i],channels_per_array_1,
              channels_per_array_2,1,i,n_beams,&cnst,dv_cal_file_name);
    cnst=cnst/ncut;
    aver=0;
    printf (" cnst = %f \n",cnst);
    for (i=0;i<n_beams;i++)
       BITREV2D( beam_table[i],p-1,64*channels_per_array,0,new_geometry);  
    CM_set_vp_set_geometry(CM_current_vp_set, cut_geometry);
    for (i=0;i<n_beams;i++) {
     if ( cut !=0) COPYFREQB(beam_table[i],beam_table[i],
                      64*channels_per_array,cut);
    }
    CM_set_vp_set_geometry(CM_current_vp_set, new_geometry);

    if (time) {for (i=0;i<64;i++) CM_timer_clear(i);}
    if (time) CM_timer_start(time_all);
    if (time) CM_timer_start(time_in);
    INITDATA (old_data,CHANNELS*sample_size*2*(sweeps_in+1),
              sample_size/4,ser_data,dv_infile);
    if (time) CM_timer_stop(time_in);
    for(step = 0; step < steps; step++){
      
/******************************************************************************
*                                                                             *
*  Retrieve next sweep of input data                                          *
*                                                                             *
******************************************************************************/

    if (time) CM_timer_start(time_in);
    LOADDATA (pure_data,old_data,CHANNELS*sample_size*2,dv_infile,
               offset,size/2,sweeps_in,FALSE,sample_size/4,ser_data);
    if (time) CM_timer_stop(time_in);
    if ( offset >= sweeps_in*pts/2) offset += -sweeps_in*pts/2;
    offset += update;

/******************************************************************************
*                                                                             *
*  Float input data and window if required                                    *
*                                                                             *
******************************************************************************/
    if (time) CM_timer_start(time_reformat);
    if (decomp) {
      for (channel =0; channel < 2*CHANNELS; channel++){
        DECOMPAND(fft_temp+32*channel, pure_data+sample_size*channel, cmte);
       }
    }
    else if (int_data){
      for (channel = 0; channel <2*CHANNELS; channel++){
         CM_f_s_float_2_2L (fft_temp+32*channel,
                     pure_data+sample_size*channel, sample_size,23,8);
       }
     }

      else {
      for (channel = 0; channel < 2*CHANNELS; channel++){
        CM_u_move_1L (fft_temp+32*channel, pure_data+sample_size*channel, 
                         sample_size);
       }
      }
    if (time) CM_timer_stop(time_reformat);

    if (time) CM_timer_start(time_window);
    if (take_hann != 0) {
      for (channel = 0; channel < CHANNELS; channel++){
	  CM_f_multiply_2_1L(fft_temp+64*channel,window,23,8);
	  CM_f_multiply_2_1L(fft_temp+64*channel+32,window+32,23,8);
        }
      }
    if (time) CM_timer_stop(time_window);

/******************************************************************************
*                                                                             *
*  Peform real-to-complex FFT on time-series data                             *
*                                                                             *
******************************************************************************/
    if (time) CM_timer_start(time_fft);
    for (channel = 0; channel < CHANNELS; channel ++){
        CMSSL_c_c_fft_detailed_a_1L (fft_temp+64*channel,id,ops,sbo,dbo,
               sca,23,8);
     }
    if (time) CM_timer_stop(time_fft);
    if (time) CM_timer_start(time_unsfft);
    UNS2REAL2D(fft_temp,  p, CHANNELS,1,TRUE);
    if (time) CM_timer_stop(time_unsfft);

/******************************************************************************
*                                                                             *
*  Truncate spectrum to desired low pass band                                 *
*                                                                             *
******************************************************************************/
    if (time) CM_timer_start(time_copy);
    CM_set_vp_set_geometry(CM_current_vp_set, cut_geometry);
    if (cut !=0 ) COPYFREQB (fft_temp,fft_temp,64*CHANNELS,cut);
    if (time) CM_timer_stop(time_copy);

/******************************************************************************
*                                                                             *
*  Beamform the subarrays                                                     *
*                                                                             *
******************************************************************************/

    if (time) CM_timer_start(time_beam);
   for ( number =0; number < CLUSTERS ; number ++) {
      for(beam = 0, table_index = 0;beam < n_beams;beam++){
        BEAM (beam_real+64*(beam+number*n_beams),
              fft_temp+64*(number*(channels_per_array)),
              beam_table[beam]+64*(table_index),
	      channels_per_array_1);
       }
    }
   for ( number =0; number < CLUSTERS ; number ++) {
      for(beam = 0, table_index = channels_per_array_1;beam < n_beams;beam++){
        BEAM (beam_real+64*(beam+number*n_beams+CLUSTERS*n_beams),
              fft_temp+64*
               (channels_per_array_1+number*(channels_per_array)),
              beam_table[beam]+64*(table_index),
	      channels_per_array_2);
       }
    }
    if (time) CM_timer_stop(time_beam);


/******************************************************************************
*                                                                             *
*  SCOT window, bandpass filter, and center subarray spectrums                *
*                                                                             *
******************************************************************************/

    if (time) CM_timer_start(time_spect);
    for ( channel=0; channel< n_beams*CLUSTERS; channel ++) {
    CM_u_move_1L (x+64*channel,beam_real+64*channel,64);
    CM_u_move_1L (y+64*channel,beam_real+64*channel+64*n_beams*CLUSTERS,64);
       }

    if (take_scot ) {
       SCOT (wx,x,BB_CORR,scot_flag);
       SCOT (wy,y,BB_CORR,scot_flag);
       scot_flag=1;
     }

     if (take_gfilt) {
      for (number=num_filt-1;number>=0; number -= 1) {
       for ( channel=0; channel < 2*BB_CORR; channel++ ) {
        CM_f_multiply_3_1L (x+32*channel+32*number*2*BB_CORR,x+32*channel,
               gfilt+32*number,23,8);
        CM_f_multiply_3_1L (y+32*channel+32*number*2*BB_CORR,y+32*channel,
               gfilt+32*number,23,8);
       }
      }
    }

    NEGODDC (y,BB_CORR*num_filt,1);

/******************************************************************************
*                                                                             *
*  Form cross-spectrum of subarrays                                           *
*                                                                             *
******************************************************************************/

    CONJMULT (fft_aver,x,y,BB_CORR*num_filt);
    if (time) CM_timer_stop(time_spect);

/******************************************************************************
*                                                                             *
*  Calculate broadband correlation using inverse FFT algorithm and normalize  *
*  output to between -1. and 1.                                               *
*                                                                             *
******************************************************************************/

   if (time) CM_timer_start(time_scr);
   SCR2REAL2D(fft_aver,  p-cut, BB_CORR*num_filt,1,TRUE);
   if (time) CM_timer_stop(time_scr);

   if (time) CM_timer_start(time_ifft);
   for (channel = 0; channel < BB_CORR*num_filt; channel ++){
        CMSSL_c_c_fft_detailed_a_1L (fft_aver+64*channel,idinv,opscor,sbocor
                    ,dbocor,scacor,23,8);
     }
   if (time) CM_timer_stop(time_ifft);

   if (time) CM_timer_start(time_norm);
   for (channel = 0; channel < BB_CORR*num_filt; channel ++){
          TAKE_AMPLITUDE_SQ (xamp+32*channel,x+64*channel);
          TAKE_AMPLITUDE_SQ (yamp+32*channel,y+64*channel); 
     }
   for (channel = 0; channel < BB_CORR*num_filt; channel ++){
         CM_spread_with_f_add_1L ( xfl,xamp+32*channel,0,23,8);
         CM_spread_with_f_add_1L ( yfl,yamp+32*channel,0,23,8);
         CM_f_multiply_3_1L (xyfl+32*channel,xfl,yfl,23,8);
         CM_f_sqrt_1_1L(xyfl+32*channel,23,8);
         xpts=pts/2./ncut;
         CM_f_divide_constant_2_1L (xyfl+32*channel,xpts,23,8);
     }
   for (channel = 0; channel < BB_CORR*num_filt; channel ++){
       CM_f_divide_2_1L ( fft_aver+64*channel,xyfl+32*channel,23,8);
       CM_f_divide_2_1L ( fft_aver+32+64*channel,xyfl+32*channel,23,8);
     }
    if (time) CM_timer_stop(time_norm);

/******************************************************************************
*                                                                             *
*  Merge correlations from beams into single broadband correlogram            *
*                                                                             *
******************************************************************************/

  if (time) CM_timer_start(time_form);
  if (n_beams > 0) 
   for (number=0;number<num_filt; number ++) {
    BMMACHELL (bb_out+64*(number),
            fft_aver+64*number*n_beams,n_beams,64,cnst);
       CM_set_context();
 }

/*    wrnf (bb_out,start,stop,23,8);*/
     FORMBB(bb_out,out_total,p-1-cut);
/******************************************************************************
*                                                                             *
*  Suppress spokes in broadband correlation                                   *
*                                                                             *
******************************************************************************/

    for (i=0;i< out_total;i++) CM_f_add_2_1L (bb_aver+i*32,bb_out+i*64,23,8);
    aver ++ ;
    if (aver == save_aver) {
    aver=0;
    for (i=0;i< out_total;i++)
        CM_f_divide_constant_2_1L (bb_aver+i*32,1.*save_aver,23,8);
    if (spoke_supp) {
      for (channel=0; channel< out_total; channel ++) {
         SUPPRESS (bb_aver+32*channel, spoke_temp+32*channel,alpha);
       }
      alpha=alpha_in;
   }
    if (time) CM_timer_stop(time_form);

/******************************************************************************
*                                                                             *
*  Prepare data for out to DataVAult and when output buffer full dump to      *
*  DataVault                                                                  *
*                                                                             *
******************************************************************************/

    if (time) CM_timer_start(time_trans);
    for (i=0;i<out_total;i++) { 
      CM_swap_2_1L ( bb_aver+32*i+24, bb_aver+32*i,8);
      CM_swap_2_1L ( bb_aver+32*i+16, bb_aver+32*i+8,8);
      CMFS_transpose_always(bb_aver+32*i,32,CMFS_IO_WRITE_TO_SERIAL_DATA,NULL);
    }
    if (time) CM_timer_stop(time_trans);
    if (time) CM_timer_start(time_out);
    CM_u_move_1L(out_amp+32*sweep_count*out_total,bb_aver,32*out_total);
    for (i=0;i< out_total;i++) CM_f_move_zero_1L (bb_aver+i*32,23,8);
    sweep_count += 1;
    if (sweep_count == sweeps_out) {
     SAVE_AMPLITUDE(out_amp, 0, dv_outfile,4*sweeps_out*out_total);
     sweep_count=0;
   }
    if (time) CM_timer_stop(time_out);
  }
    if (time && step==(aver-1)) {
       CM_timer_stop(time_all);
       for (i=0;i<64;i++) CM_timer_clear(i);
       CM_timer_start(time_all);
      }
    CM_set_vp_set_geometry(CM_current_vp_set, new_geometry);
    }
    
  if (time) CM_timer_stop(time_all);
  if (time)  TIMEBB();
  fflush(stdout);	      

/******************************************************************************
*                                                                             *
*  Dump residue in output buffer and cleanup                                  *
*                                                                             *
******************************************************************************/

  SAVE_AMPLITUDE(out_amp, 0, dv_outfile,4*sweep_count*out_total);
  CMFS_close(dv_infile);
  CMFS_close(dv_outfile);
}











