/*

  vis.c - Generates standard Lofargram data for a set of time series data 
             The order of processing is

            1. Read 8-bit companded data from data vault
            2. Bitreverse order of the data
            3. Decompand if necessary and move the data
            4. Window the data
            5. FFT the data
            6. Beamform the data for channels from a line array.
            7. Compute amplitude
            8. Noise Spectrun Equalize outputs
            9. Clip and descritize output to 8 bits
           10. Store spectrum in file on DataVault and/or display first
               channel on framebuffer

       Signal processing parameters are established in file visnew.h . visnew
       should be remade using make if visnew.h changed.  The following switches
       may be used on calling sequence:
   
         -t n    time part n of program
         -s n    Process data for a total of n steps ( # of FFT's)
         -a n    Average n amplitudes before NSE and output
         -f      Display on framebuffer
         -h      Hanning window
         -v      Overlap data
         -d      Decompand
         -c n    field_idno 
         -o name Save data to file name on data vault

*/

#include <sys/types.h>
#include <cm/cm.h>
#include <cm/cm_file.h>
#include <cm/cmfs.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
#include <cm/paris.h>
#include <cm/cmssl-paris.h>
#include <cm/cmfb.h>
#include "/s4/wanderson/APT/DAVE/vis.h"

/* Routine to print error message when run line incorrect */

void usage()
{
  fprintf(stderr,"useage: vis \"/optical/6.2-ONT-data/\"filename options...\n");
  fprintf(stderr, "\t-s n ==>take n fft steps, defaults to 8.\n");
  fprintf(stderr, "\t-a n ==>average every a steps, defaults to 1.\n");
  fprintf(stderr, "\t-t n ==>time part n of program.\n");
  fprintf(stderr, "\t-f   ==>automatically dump to the fb.\n");
  fprintf(stderr, "\t-h   ==>use the hanning window.\n");
  fprintf(stderr, "\t-v   ==>overlap 50%.\n");
  fprintf(stderr, "\t-o   ==>dv output file.\n");
  exit (1);
}



void main(argc, argv)
     int argc;
     char **argv;
{
  char dv_in_file_name[128], dv_out_file_name[128],beam_file_name[128];
  char dv_cal_file_name[128];
  DVFILE dv_infile, dv_outfile;
  u_char *grid;
  int oddsteps,number;
  int aver = AVERAGE,save_aver =  AVERAGE;
  int pts = PTS, steps = STEPS, ptr, step, outfile_flag = 0;
  int copy,np,nproc;
  float cnst,dstep;
  int show_fb = FALSE, take_hann = FALSE,
  decomp=FALSE, int_data=FALSE,
  do_overlap = FALSE, save_grid = TRUE, ser_data=FALSE;
  int time=FALSE;
  int time_all=0,time_in=1,time_bitrev=2,time_reformat=3,time_window=4,
      time_fft=5,time_unsfft=6,time_beam=7,time_amp=8,time_nse=9,
      time_clip=10,time_frame=11,time_trans=12,time_out=13,time_axis=16;
  int offset=0,update=PTS/2;
  int noff,ndisp=8;
  int sample=32;
  int field_idno=0;
  int sweep_count=0;
  int debu, dbu;
  int ncut=1<<CUT,out_channels;
  float te[128];   /* the decomansion table */
  unsigned start[32],stop[32];
  
  unsigned  pure_data,
  old_data,
  new_data,
  beam_table,
  cmte,                    /* decompansion table */
  window,                  /* Hanning & Hamming windows */
  fft_aver,
  fft_temp,
  amplitude,  
  signed_amplitude,
  fft_amp,
  fft_angle_old,
  fft_angle_new,
  fft_angle,
  table,h,s,v,
  amp_out,    
  ali,        
  beam_real,
  temp;
  
 CM_geometry_id_t send_geometry,news_geometry;
  unsigned dimension_array[2], new_vp_set;
  CMSSL_fft_setup_t id;
  CMSSL_op_t ops[2];
  CMSSL_bit_order_t sbo[2],dbo[2];
  CMSSL_scale_t sca[2];
  CMSSL_op_t opscor[2];
  CMSSL_bit_order_t sbocor[2],dbocor[2];
  CMSSL_scale_t scacor[2];
  struct CM_axis_descriptor axe_desc[32];
  CM_axis_descriptor_t axe_desc_t[32];
  int i;
  start[0]=510;stop[0]=514;start[1]=0;stop[1]=0;
  CM_init ();

  CMSSL_init ();
  CM_set_safety_mode (1); 
  CM_reset_timer(0);
  INIT_FRAME_BUFFER(1024,1024);

  dv_cal_file_name[0] = '\0';
/* Process filename & arguments: */
  if (argc == 1) usage();
  strcpy(dv_in_file_name, argv[1]);
  for (ptr=2; ptr<argc; ptr++) {
    char *arg = argv[ptr];
    if (arg[0] == '-') {
      switch (arg[1]) {
      case 's':                           /* -s steps */
	if (++ptr >= argc) usage ();
	steps = atoi(argv[ptr]);
	continue;
      case 't':                           /* -t time */
        time = TRUE;
        continue;
      case 'a':                           /* -a steps */
	if (++ptr >= argc) usage ();
	aver = atoi(argv[ptr]);
        save_aver = aver;
	continue;
      case 'r':
	ser_data = TRUE;
  	continue;
      case 'h':                           /* -h ==> hann window  */
	take_hann = TRUE;  
	continue;
      case 'd':                           /* -d ==> decompand data  */
	decomp = TRUE;
        sample=8;  
	continue;
      case 'i':                           /* -i sample_size */
      if (++ptr >= argc) usage ();
        int_data=TRUE;
      sample = atoi(argv[ptr]);
      continue;
      case 'v':                           /* -v ==> overlap  */
        do_overlap = TRUE;  
        update=update/2;
	continue;
       case 'o':                           /* -display host:dpy */
        if (++ptr >= argc) usage ();
        outfile_flag = 1;
        strcpy(dv_out_file_name, argv[ptr]);
        continue;
     case 'b':                           /* -c filename */
	if (++ptr >= argc) usage ();
	strcpy(beam_file_name, argv[ptr]);
	continue;
      case 'y':                           /* -c filename */
	if (++ptr >= argc) usage ();
	strcpy(dv_cal_file_name, argv[ptr]);
	continue;
      default: usage ();
      }
    }
    else usage();
  }

/* set up FFT configuration */
  ops[0]=CMSSL_f_xform;
  sbo[0]=CMSSL_bit_reversed;
  dbo[0]=CMSSL_normal;
  sca[0]=CMSSL_noscale;
  ops[1]=CMSSL_nop;
  sbo[1]=CMSSL_normal;
  dbo[1]=CMSSL_normal;
  sca[1]=CMSSL_noscale;
   for (i = 0; i < 2; i++) {
        axe_desc[i].weight = (1-i)*10;
        axe_desc[i].ordering = CM_send_order;
        axe_desc[i].off_chip_bits = 0;
        axe_desc[i].off_chip_pos = 0; 
        axe_desc[i].on_chip_bits = 0;
        axe_desc[i].on_chip_pos = 0;
        axe_desc_t[i] = &(axe_desc[i]);
      }
  nproc=CM_geometry_total_processors(CM_vp_set_geometry(CM_current_vp_set));
  copy=nproc/(PTS/2);
   axe_desc[1].length = copy;
   axe_desc[0].length = PTS/2;
   send_geometry = CM_create_detailed_geometry((unsigned *)(&axe_desc_t[0]),2);
   axe_desc[0].ordering = CM_news_order;
   axe_desc[1].ordering = CM_news_order;
   axe_desc[1].length = copy*ncut;
   axe_desc[0].length = PTS/2/ncut;
   news_geometry = CM_create_detailed_geometry((unsigned *)(&axe_desc_t[0]),2);

  new_vp_set = CM_allocate_vp_set(send_geometry);
  CM_set_vp_set(new_vp_set);
  CM_set_context();
  out_channels=(SAVED_CHANNELS/copy+ncut-1)/ncut;

/* Allocate FCM memory */    
  pure_data=CM_allocate_stack_field(2*sample*CHANNELS/copy);
  old_data=CM_allocate_stack_field(2*sample*CHANNELS*(SWEEPS_IN+1)/copy);
  cmte=CM_allocate_stack_field(4*32);
  window=CM_allocate_stack_field(32*2*2);
  fft_aver=CM_allocate_stack_field(32*SAVED_CHANNELS/copy);    
  fft_temp=CM_allocate_stack_field(64*CHANNELS/copy);    
  table=CM_allocate_stack_field(24);   
  h=CM_allocate_stack_field(32);   
  s=CM_allocate_stack_field(32);   
  v=CM_allocate_stack_field(32);   
  amplitude=CM_allocate_stack_field(32);   
  signed_amplitude=CM_allocate_stack_field(8);   
  fft_amp=CM_allocate_stack_field(8*SAVED_CHANNELS/copy);     
  fft_angle=CM_allocate_stack_field(8*SAVED_CHANNELS/copy);     
  fft_angle_old=CM_allocate_stack_field(32*SAVED_CHANNELS/copy);     
  fft_angle_new=CM_allocate_stack_field(32*SAVED_CHANNELS/copy);     
  amp_out=CM_allocate_stack_field(8*out_channels*SWEEPS_OUT);     
  ali=CM_allocate_stack_field(32*SAVED_CHANNELS/copy);         
  beam_table=CM_allocate_stack_field(64*CHANNELS_PER_CLUSTER 
                * N_BEAMS* CLUSTERS/copy);
  beam_real=CM_allocate_stack_field(64*N_BEAMS* CLUSTERS/copy);
  temp=CM_allocate_stack_field(8);

  id= CMSSL_c_c_fft_setup_1L (fft_temp,23,8);

/* Open DataVault files for input and output */  
  if ((dv_infile = CMFS_open(dv_in_file_name, CMFS_O_RDONLY, 0644,
			     CM_physical_processors_limit,
			     CM_user_number_of_processors_limit)) == -1){
    printf("error on dv file open: %s\n", dv_in_file_name);
    exit(1);
  }
  if (!outfile_flag) strcpy(dv_out_file_name, "vis.out");
  if ((dv_outfile = CMFS_open(dv_out_file_name, CMFS_O_RDWR | CM_CREAT, 0644,
			      CM_physical_processors_limit,
			      CM_user_number_of_processors_limit)) == -1){
    printf("error on dv file open: %s\n", dv_out_file_name);
    exit(1);
  }
  
/* Setup decompansion table */
  if (decomp) {
    for (debu = 0;  debu < 128; debu++) te[debu] = 0;
    COMPUTE_DECOMP_TABLE(te);
  }

  {
    int channel, bits, odd, beam, table_index, i;
    int mv_offset;
    
/* Initialize CM memory */
    CM_u_move_zero_1L(amplitude,32);
    if (decomp) LOAD_DECOMP_TABLE(te, cmte); 
    CM_u_move_zero_1L(fft_angle_old,32*SAVED_CHANNELS/copy);
    CM_u_move_zero_1L(fft_aver,32*SAVED_CHANNELS/copy);
    CM_u_move_zero_1L(ali,32*SAVED_CHANNELS/copy);
    HANN2(window, PTS);
    HAMM2(window+64, PTS);
    BITREV2D (window, P-1, 4*32, 0);
    BEAMTABLE(beam_file_name,beam_table,CHANNELS_PER_CLUSTER_1,
         CHANNELS_PER_CLUSTER_2,CLUSTERS/copy,N_BEAMS,&cnst,dv_cal_file_name);
    if (time) {for (i=0;i<64;i++) CM_timer_clear(i);}
    if (time) CM_timer_start(time_all);
/* Input time series data */
  INITDATA (old_data,2*CHANNELS*sample*(SWEEPS_IN+1)/copy,
              sample/4,ser_data,dv_infile);
  CM_u_move_1L (pure_data, old_data,2*CHANNELS * sample/copy );
    oddsteps=0;
    
    for(step = 0; step < steps; step++){

/* Retrieve next set of data */
    LOADDATA (pure_data,old_data,2*CHANNELS*sample/copy,dv_infile,
               offset,PTS/2,SWEEPS_IN,TRUE,sample/4,ser_data);
    if ( offset >= SWEEPS_IN*PTS/2) offset += -SWEEPS_IN*PTS/2;
    offset += update;

/* Decompand data                                                         */
    if (time) CM_timer_start(time_reformat);
    if (decomp) {
      for (channel =0; channel < 2*CHANNELS/copy; channel++){
	DECOMPAND(fft_temp+32*channel, pure_data+sample*channel, cmte);
       }
    }
    else if (int_data){
      for (channel = 0; channel <2*CHANNELS/copy; channel++){
	 CM_f_s_float_2_2L (fft_temp+32*channel,
                     pure_data+sample*channel, sample,23,8);
       }
     }

      else {
      for (channel = 0; channel < 2*CHANNELS/copy; channel++){
	CM_u_move_1L (fft_temp+32*channel, pure_data+sample*channel, sample);
       }
      }
  if (time) CM_timer_stop(time_reformat);

/* Hanning window the data*/
    if (time) CM_timer_start(time_window);
    if (take_hann != 0) {
      for (channel = 0; channel < 2*CHANNELS/copy; channel++){
	  CM_f_multiply_2_1L(fft_temp+32*channel,window,23,8);
          channel ++;
	  CM_f_multiply_2_1L(fft_temp+32*channel,window+32,23,8);
        }
      }
  if (time) CM_timer_stop(time_window);
/* FFT the data */
    if (time) CM_timer_start(time_fft);
    for (number=0; number< CLUSTERS/copy; number++) {
     for (channel = CHANNELS_PER_CLUSTER_1; channel < CHANNELS_PER_CLUSTER; 
                     channel += 1){
        CMSSL_c_c_fft_detailed_a_1L (
                    fft_temp+64*(channel+number*CHANNELS_PER_CLUSTER),
                    id,ops,sbo,dbo,sca,23,8);
     }
   }
  if (time) CM_timer_stop(time_fft);

/* Unscramble real and imaginery parts  */
    if (time) CM_timer_start(time_unsfft);
    for (number=0;number< CLUSTERS/copy; number++) {
      UNS2REAL2D (
         fft_temp+64*(CHANNELS_PER_CLUSTER_1+number*CHANNELS_PER_CLUSTER), 
           P, CHANNELS_PER_CLUSTER_2,0,FALSE); 
    }
    if (time) CM_timer_stop(time_unsfft);

/* form N_BEAMS for clusters */
    if (time) CM_timer_start(time_beam);
   channel = 0;
    for (number=0; number<CLUSTERS/copy;number++) {
      table_index=CHANNELS_PER_CLUSTER_1+number*CHANNELS_PER_CLUSTER;      
      for(beam = 0;
	  beam < N_BEAMS;
	  beam++, table_index += CHANNELS_PER_CLUSTER*CLUSTERS/copy){
           BEAM (beam_real+64*(beam+number*N_BEAMS),
            fft_temp+64*(CHANNELS_PER_CLUSTER_1+number*(CHANNELS_PER_CLUSTER)),
            beam_table+64*table_index,
            CHANNELS_PER_CLUSTER_2);
       }
    }
  if (time) CM_timer_stop(time_beam);

/*  Calculate amplitudes */
    if (time) CM_timer_start(time_amp);
     for( channel = 0 ; channel < SAVED_CHANNELS/copy; channel++){
	TAKE_AMPLITUDE(fft_temp+64*channel, beam_real+64*channel);
	CM_f_add_2_1L (fft_aver+32*channel,fft_temp+64*channel,23,8);
      }
  if (time) CM_timer_stop(time_amp);
/* NSE the average amplitudes */
      aver=aver - 1 ;
      if ( aver == 0) {
    if (time) CM_timer_start(time_axis);
    TRUNCFREQ (beam_real,beam_real,SAVED_CHANNELS/copy,64,CUT,
                   send_geometry,news_geometry);
    TRUNCFREQ (fft_aver,fft_aver,SAVED_CHANNELS/copy,32,CUT,
                   send_geometry,news_geometry);
    if (time) CM_timer_stop(time_axis);

   CM_set_vp_set_geometry(CM_current_vp_set,news_geometry);

    if (time) CM_timer_start(time_nse);
          NSE_SAXA(fft_aver, P-1-CUT, out_channels, 3, 67);
    if (time) CM_timer_stop(time_nse);

/* Clip and discretize spectrum */
    if (time) CM_timer_start(time_clip);
    for( channel = 0 ; channel < out_channels; channel++){
        CLIP(fft_aver+32*channel, 5., 51.);
        CM_f_divide_constant_3_1L(v,fft_aver,255.,23,8);
        FLOOR(signed_amplitude, fft_aver+32*channel);
        CM_u_move_1L (fft_amp+8*channel,signed_amplitude,8);
        CM_f_atan2_3_1L(fft_angle_new+32*channel,beam_real+64*channel+32,
          beam_real+64*channel,23,8);
        wrnf(fft_angle_new,start,stop,23,8);
        CM_f_subtract_3_1L(fft_temp+32*channel,fft_angle_old+32*channel,
          fft_angle_new+32*channel,23,8);
        CM_f_move_1L(fft_angle_old+32*channel,fft_angle_new+32*channel,23,8);
        CM_f_lt_zero_1L(fft_temp+32*channel,23,8);
        CM_logand_context_with_test();
        CM_f_add_constant_2_1L(fft_temp+32*channel,2.*3.1415926,23,8);
        CM_set_context();
        CM_f_ge_constant_1L(fft_temp+32*channel,2.*3.1415926,23,8);
        CM_logand_context_with_test();
        CM_f_subtract_constant_2_1L(fft_temp+32*channel,2.*3.1415926,23,8);
        CM_set_context();
        CM_f_move_1L(h,fft_temp,23,8);
        CM_f_multiply_constant_2_1L(h,360./6.2831852,23,8);
        wrnf(h,start,stop,23,8);
       CM_f_move_constant_1L(s,1.,23,8);
        CM_f_multiply_constant_2_1L(fft_temp+32*channel,255./6.2831852,23,8);
        FLOOR(fft_angle, fft_temp+32*channel);
        CM_u_add_3_3L (ali+32*channel,ali+32*channel,fft_amp+8*channel,32,32,8);
       }
  if (time) CM_timer_stop(time_clip);
/* CM_f_move_constant_1L (v,1.,23,8);*/
  HSVTORGB(table,h,s,v);

  if (time) CM_timer_start(time_trans);
  NEWSTOSEND (fft_amp,8*out_channels);
  NEWSTOSEND (fft_angle,8*out_channels);
  NEWSTOSEND (table,24);
  FRAMEBUFFER(table,sweep_count,1,0);
  CM_set_vp_set_geometry(CM_current_vp_set,send_geometry);
  if (time) CM_timer_stop(time_trans);

     for(channel = 0; channel < out_channels; channel++){
     CM_u_move_1L (amp_out+8*channel+8*sweep_count*out_channels,
                       fft_amp+8*channel,8);
     CMFS_transpose_always(amp_out+8*channel+8*sweep_count*out_channels,
             8,CMFS_IO_WRITE_TO_SERIAL_DATA,CMFS_write_to_column_major);
      }
  if (time) CM_timer_stop(time_trans);
  sweep_count += 1;
  if (time) CM_timer_start(time_out);
  if (sweep_count == SWEEPS_OUT) {
    SAVE_AMPLITUDE(amp_out,step/save_aver,dv_outfile,
           SWEEPS_OUT*out_channels);
    sweep_count=0;
  }
  if (time) CM_timer_stop(time_out);
    aver = save_aver;
    CM_u_move_zero_1L(fft_aver,32*SAVED_CHANNELS/copy);
      }
    if (time && step==(aver-1)) {
    if (time) CM_timer_stop(time_all);
    for (i=0;i<64;i++) CM_timer_clear(i);
    if (time) CM_timer_start(time_all);
    }
      
    }   /* end of for step */
    if (time) CM_timer_stop(time_all);
    if (time)  TIMEVIS();
    
    fflush(stdout);	      
  } 
    SAVE_AMPLITUDE(amp_out,step/save_aver,dv_outfile,
           sweep_count*out_channels);
  CMFS_close(dv_infile);
  CMFS_close(dv_outfile);
}


