/******************************************************************************
*                                                                             *
*  NSE_SAXA - Performs a Noise Spectrum Equalization on the amplitude of      *
*             the spectrum using the split average exclude average (SAXA)     *
*             algorithm described by Struzinski in " A New Normalization      *
*             Algorithm for Detection Systems." JASA Suppl 1 75 S43 , 1984    *
*                                                                             *
*     real_part -  Slice of CM memory containing amplitude of spectrum        *
*     p         -  Log to tne base 2 of the number of points in the           *
*                  spectrum                                                   *
*     slices    -  Number of spectrums                                        *
*     W         -  Outer window width                                         *
*     G         -  Inner window width                                         *
*                                                                             *
*  NOTE : (W-G) must be a power of 2 !                                        *
*                                                                             *
******************************************************************************/

#include <cm/paris.h>
#include <math.h>
#include <stdio.h>

void NSE_SAXA (real_part, p, slices, G, W)
unsigned real_part;
int G, W, p, slices;
  {
    int count, i, j, k, N, csize;
    double K;
    unsigned narrower, noise, noise_lag, temp, save_context, shifted, avgcount;
    unsigned acshift;

    K = 2./(W-G);
    N = 1 << p;

    count = 0;
    for (i = 1; i < (W - G) / 2; i *= 2) count++;
    csize = count + 2;

    noise = CM_allocate_stack_field(slices * 32);
    noise_lag = CM_allocate_stack_field(slices * 32);
    temp = CM_allocate_stack_field(slices * 32);
    avgcount = CM_allocate_stack_field(slices * csize);
    shifted = CM_allocate_stack_field(slices * 32);
    acshift = CM_allocate_stack_field(slices * csize);
    narrower = CM_allocate_stack_field(1);
    save_context = CM_allocate_stack_field(1);
    CM_u_move_zero_1L(noise,32*slices);
    CM_u_move_zero_1L(noise_lag,32*slices);
    CM_u_move_zero_1L(temp,32*slices);
    CM_u_move_zero_1L(shifted,32*slices);
    CM_u_move_zero_1L(acshift,slices*csize);
    CM_u_move_zero_1L(avgcount,slices*csize);

    CM_store_context(save_context);
    CM_set_context();
    CM_u_move_zero_1L(temp,32);
    CM_my_news_coordinate_1L(temp, 0, 32);
    CM_u_ge_constant_1L(temp, (W - 1) / 2, 32);
    CM_logand_context_with_test();
    CM_u_le_constant_1L(temp, N - (W + 1) / 2, 32);
    CM_logand_context_with_test();
    CM_store_context(narrower);
    CM_set_context();

    for (k = 0; k < slices; k++)
      CM_u_move_constant_1L(avgcount + k * csize, 1, csize);
    CM_u_move_1L(temp, real_part, 32 * slices);

/******************************************************************************
*                                                                             *
*  First pass of algorithm : Locate spikes                                    *
*                                                                             *
******************************************************************************/
    CM_u_move_1L(noise, temp, 32 * slices);
    for (j = 0; j < count; j++) {
      CM_get_from_power_two_always_1L(shifted,noise,0,j,CM_upward,32*slices);
      for (k = 0; k < slices; k++) {
	  CM_f_add_2_1L (noise + k * 32, shifted + k * 32, 23, 8);
       }
     }
    SHIFT(noise_lag, noise, 0, -(W-1)/2, 32*slices);
    SHIFT(noise, noise, 0, (G+1)/2, 32*slices);
    for (k = 0; k < slices; k++) {
      CM_f_add_2_1L (noise + k * 32, noise_lag + k * 32, 23, 8);
     }
    for (k = 0; k < slices; k++) {
	CM_load_context(narrower);
	CM_f_multiply_constant_2_1L (noise + k * 32, K, 23, 8);
	CM_f_lt_1L(noise + k * 32, temp + k * 32, 23, 8);
	CM_logand_context_with_test();
	CM_f_move_zero_1L(temp + k * 32, 23, 8);
	CM_u_move_zero_1L(avgcount + k * csize, csize);
      }
     CM_set_context();

/******************************************************************************
*                                                                             *
* Second pass of algorithm : Calculate noise background and equalize          *
*                                                                             *
******************************************************************************/
     for (j = 0; j < count; j++) {
       CM_get_from_power_two_always_1L(shifted,temp,0,j,CM_upward, 32*slices);
       CM_get_from_power_two_always_1L(acshift,avgcount,0,j,
            CM_upward,csize*slices);
       for (k = 0; k < slices; k++) {
	 CM_f_add_2_1L (temp + k * 32, shifted + k * 32, 23, 8);
	 CM_u_add_2_1L (avgcount + k * csize, acshift + k * csize, csize);
        }
      }
     SHIFT(shifted, temp, 0, - (W-G) / 2, 32*slices);
     SHIFT(acshift, avgcount,0,-(W-G)/2,csize*slices);
     for (k = 0; k < slices; k++) {
	CM_f_add_2_1L (temp + k * 32, shifted + k * 32, 23, 8);
	CM_u_add_2_1L (avgcount + k * csize, acshift + k * csize, csize);
	CM_f_divide_2_1L(real_part + k * 32, temp + k * 32, 23, 8);
	CM_f_u_float_2_2L(temp + k * 32, avgcount + k * csize, csize, 23, 8);
	CM_f_multiply_2_1L(real_part + k * 32, temp + k * 32, 23, 8);
      }
    CM_deallocate_stack_through(noise);
  }
