/*   File Header   ---------------------------------

     File Name:       convert.c

     Author and Date: George Lamb--4/11/90

     Description: functions to convert between different data types,
coordinate systems, and units.  Defined functions:

	void 	GD_map_2d_plane_3d(dest, px, py, x_axis, y_axis, origin) 
	void 	GD_map_2d_volume_3d(dest, px, x_axis, y_axis, nv, origin) 
	void 	GD_map_3d_angular(dest, src, origin) 
	void 	GD_map_angular_3d(dest, src, origin) 

     End Header   ---------------------------------- */


#include <geodata.h>
#include <geodata_private.h>
#include <cm/paris.h>
#include <stdio.h>
#include <math.h>
#include "vector.h"


/*   Function Header   -----------------------------
     NAME:        GD_map_2d_plane_3d, GD_map_2d_volume_3d
     DESCRIPTION: map a 2d plane or volume to a 3d coordinate system
     PARAMETERS:  dest--cm dest, should contain 3 floats, (x, y, z)
                  px, py--cm floats, coordinates on the plane
		  x-axis, y-axis--axii which define the plane
		  origin--the local origin on the plane
		  (volumes) pz--cm volume coordinates
		  (volumes) nv--the normal vector
     PRECOND:     dest is 3 32bit consecutive CM floats
     POSTCOND:    dest is modified
     RETURN:      void
     LOGIC:       The mapping of planar points into 3 space is
     accomplished using:
        3d_plane = px * x-axis + py * y-axis + origin
        3d_volume = px * x-axis + py * y-axis + pz * nv + origin
	where x & y axis, nv, & origin are 3d vectors.
     End Header   ---------------------------------- */

void GD_map_2d_plane_3d(dest, px, py, x_axis, y_axis, origin)
     CM_field_id_t dest, px, py;
     GD_point *x_axis, *y_axis;
     float *origin;
{
  CM_f_mult_const_add_const_1L(dest, py, y_axis->lat, origin[0], 23, 8);
  CM_f_mult_const_add_1L(dest, px, x_axis->lat, dest, 23, 8);
  CM_f_mult_const_add_const_1L(dest+32, py, y_axis->lon, origin[1], 23, 8);
  CM_f_mult_const_add_1L(dest+32, px, x_axis->lon, dest+32, 23, 8);
  CM_f_mult_const_add_const_1L(dest+64, py, y_axis->alt, origin[2], 23, 8);
  CM_f_mult_const_add_1L(dest+64, px, x_axis->alt, dest+64, 23, 8);
}       /* End of GD_map_2d_plane_3d*/


void GD_map_2d_volume_3d(dest, px, x_axis, y_axis, nv, origin)
     CM_field_id_t dest, px;
     GD_point *x_axis, *y_axis, *nv;
     float *origin;
{
  CM_field_id_t py, pz;
  GD_point v;
  unify(nv, &v);
  py = px + 32; pz = px + 64;
  CM_f_mult_const_add_const_1L(dest, py, y_axis->lat, origin[0], 23, 8);
  CM_f_mult_const_add_1L(dest, px, x_axis->lat, dest, 23, 8);
  CM_f_mult_const_add_1L(dest, pz, v.lat, dest, 23, 8);
  CM_f_mult_const_add_const_1L(dest+32, py, y_axis->lon, origin[1], 23, 8);
  CM_f_mult_const_add_1L(dest+32, px, x_axis->lon, dest+32, 23, 8);
  CM_f_mult_const_add_1L(dest+32, pz, v.lon, dest+32, 23, 8);
  CM_f_mult_const_add_const_1L(dest+64, py, y_axis->alt, origin[2], 23, 8);
  CM_f_mult_const_add_1L(dest+64, px, x_axis->alt, dest+64, 23, 8);
  CM_f_mult_const_add_1L(dest+64, pz, v.alt, dest+64, 23, 8);
}       /* end of GD_map_2d_volume_3d */




/*   Function Header   -----------------------------
     NAME:        GD_map_3d_angular
     DESCRIPTION: map 3d cartesian coordinates to 3d angular coordinates
     PARAMETERS:  dest--cm dest, should contain 3 floats, (lat, lon, alt)
                  src--cm src, should contain 3 floats, (x, y, z)
		  origin--the local origin on cartesian system
     PRECOND:     dest and src are 3 32bit consecutive CM floats
     POSTCOND:    dest is modified
     RETURN:      void
     LOGIC:       use these equations:
                  alt = sqrt(x^2 + y^2 + z^2)
		  lon = arctan (y/x)
		  lat = arctan (z/x)
		  alt -= GD_earth_radius
     End Header   ---------------------------------- */
void GD_map_3d_angular(dest, src, origin)
     CM_field_id_t dest, src;
     float  *origin;
{
  CM_field_id_t temp, temp_z, ctx;

  temp = CM_allocate_stack_field(3 * 32);
  temp_z = CM_allocate_stack_field(32);
  ctx = CM_allocate_stack_field(1);

/*  if ((origin[2] != 0) || (origin[0] != 0) || (origin[1] != 0)){
    printf("using local origin for mapping 3d to angular.\n");
    CM_f_add_constant_3_1L(temp, src, origin[0], 23, 8);
    CM_f_add_constant_3_1L(temp+32, src+32, origin[1], 23, 8);
    CM_f_add_constant_3_1L(temp+64, src+64, origin[2], 23, 8);
  }
  else */
  CM_u_move_1L(temp, src, 3*32);        /* for overlap! */

  CM_f_multiply_constant_3_1L(temp_z,src+64, GS_EARTH_RADIUS, 23, 8);  

  CM_f_multiply_3_1L(dest+64, temp, temp, 23, 8);             /* alt = x^2 */
  CM_f_mult_add_1L(dest+64, temp+32, temp+32, dest+64, 23,8); /* alt += y^2 */
  CM_f_mult_add_1L(dest+64, temp+64, temp+64, dest+64, 23,8); /* alt += z^2 */
  CM_f_sqrt_1_1L(dest+64, 23, 8);                             /* alt sqrt=  */

  CM_store_context(ctx);
  CM_f_atan2_3_1L(dest+32, temp+32, temp, 23, 8);    /* lon = atan(y/x) */
  CM_load_context(CM_test_flag);                     /* test => bad atan */
  CM_f_move_constant_1L(dest+32,atan(1.0), 23, 8);
  CM_load_context(ctx);

  CM_f_divide_3_1L(dest, temp_z, dest+64, 23, 8);    /* lat = z/alt */
  CM_load_context(CM_test_flag);                     /* test => bad atan */
  CM_f_move_constant_1L(dest,0.0, 23, 8);
  CM_invert_context();
  CM_f_asin_1_1L(dest, 23, 8);                       /* lat = asin */
  CM_load_context(ctx);

  CM_f_subtract_constant_2_1L(dest+64, GS_EARTH_RADIUS, 23, 8);
  CM_deallocate_stack_through(temp);
}       /* end of GD_map_3d_angular */




/*   Function Header   -----------------------------
     NAME:        GD_map_angular_3d
     DESCRIPTION: map angular coordinates to 3d cart coordinates
     PARAMETERS:  dest--cm src, should contain 3 floats, (x, y, z)
                  src--cm dest, should contain 3 floats, (lat, lon, alt)
		  origin--the local origin on angular system
     PRECOND:     dest and src are 3 32bit consecutive CM floats
     POSTCOND:    dest contains src in 3d cart coordinates
     RETURN:      void
     LOGIC:       use these equations:
                  r = alt + earth_radius
		  rp = r cos(lat)
                  x = -rp * cos(lon)
		  y = rp * sin(lon)
		  z = r * sin(lat)
     End Header   ---------------------------------- */
void GD_map_angular_3d(dest, src, origin)
     CM_field_id_t dest, src;
     float *origin;
{
  CM_field_id_t cart, r_prime;
  cart = CM_allocate_stack_field(3*32);
  r_prime = CM_allocate_stack_field(32);
  CM_u_move_1L(cart, src, 3*32);                /* for overlap! */
                                                /* alt += RADIUS */
/*  GD_transfer_statistics (1,cart+64);*/
  CM_f_add_constant_2_1L(cart+64, GS_EARTH_RADIUS, 23, 8);  
  CM_f_divide_constant_2_1L(cart+64, GS_EARTH_RADIUS, 23, 8);  

  CM_f_cos_2_1L(r_prime, cart, 23, 8);          /* r_prime = cos(lat) */
  CM_f_multiply_2_1L(r_prime, cart+64, 23, 8);  /* r_prime *= r */
  CM_f_cos_2_1L(dest, cart+32, 23, 8);          /* x = cos(lon) */
  CM_f_multiply_2_1L(dest, r_prime, 23, 8);     /* x *= r_prime */
  /* hack*/
  CM_f_negate_1_1L (dest,23,8);                 /* x = -x */

  CM_f_sin_2_1L(dest+32, cart+32, 23, 8);       /* y = sin(lon) */
  CM_f_multiply_2_1L(dest+32, r_prime, 23, 8);  /* y *= r_prime */
  CM_f_sin_2_1L(dest+64, cart, 23, 8);          /* z = sin(lat) */
  CM_f_multiply_2_1L(dest+64, cart+64, 23, 8);  /* z *= r_prime */
  CM_deallocate_stack_through(cart);
}       /* end of GD_map_angular_3d */



void pr_vector(i, buf)
     int i;
     CM_field_id_t buf;
{  printf("before %d: (%-7.2f %-7.2f %-7.2f) to (%-7.2f %-7.2f %-7.2f).\n",i,
	  CM_f_read_from_processor(i,buf,23,8), 
	  CM_f_read_from_processor(i,buf+32,23,8),
	  CM_f_read_from_processor(i, buf+2*32, 23, 8),
	  CM_f_read_from_processor(i, buf+4*32, 23, 8),
	  CM_f_read_from_processor(i, buf+5*32, 23, 8),
	  CM_f_read_from_processor(i, buf+6*32, 23, 8));}



/*   Function Header   -----------------------------
     NAME:        GD_match_desc
     DESCRIPTION: make a given data source match a desired format and
     set the src data to point to the correct data.
     PARAMETERS:  src
     PRECOND:     src->desired != null
     POSTCOND:    src->data is allocated & filled
     RETURN:      void
     LOGIC:       hopefully
     End Header   ---------------------------------- */
/*

void GD_match_desc(src)
     GD_data_source *src;
{
  src->data = GD_cmalloc_std(src->desired->type, 1);


}*/   /* end of GD_match_desc */


