/*   Function Header   -----------------------------
     NAME:        GD_mv_fe_to_cm
     DESCRIPTION: Move data from front end array to CM memory
     PARAMETERS:  from,to:  two data descriptors 
                            from->fe_data is the front end array.
			    to->values will be the CM memory
     PRECOND:     The front end data must be an array of rank no more 
                  than 2. The elements of the array are floats or structures
		  containing only an array of floats. The number of floats 
		  per element depends on the data type (see below). 

		  Type:           # of floats in structure
                  GD_MESH                1     (val)
                  GD_POINT_TYPE          4     (x,y,z,val)
                  GD_VECTOR              8     (x,y,z,val,x,y,z,val)

		  The following elements of the from data descriptor 
		  must be set:
		  
		  rank       (either 1 or 2)
		  geometry   (dimensions of the cm vp_set )
		  type       (see above)

     POSTCOND:    to->values will be set to a heap field containing 
                  the values in from->fe_data in geodata standard format.
		  A vp_set will be allocated and to->vp_set will set to it.
		  This vp_set will also become the current vp_set.
     RETURN:      None
     LOGIC:  1.   Allocate space according to data type.
             2.   Read array pointed to by from->fe_data into temp array.
	     3.   Write temp array onto CM. 
	     4.   Repeat 2 & 3 for each field in the array structure.
     End Header ---------------------------------- */

#include <geodata.h>
#include <geodata_private.h>
#include <cm/paris.h>
#include <stdio.h>
#include <math.h>

int array_offset[]={0,0};
int start_vector[]={0,0};
int axis_vector[]={0,1};
char place[]="GD_mv_fe_to_cm";

void GD_mv_fe_to_cm(desc)
     GD_data_desc *desc;
{
  float *point,*temp,*float_t;
  int size,i,j,elements,offset,shape[2];
  char text[40];
  CM_field_id_t context_temp;

  /*   set up temp array */
  if(desc->rank==1)  elements = desc->geometry[0]; 
  else if(desc->rank==2) elements = desc->geometry[0]*desc->geometry[1];
  else 
    { 
      /** error **/
      sprintf(text,"Invalid rank: %d",desc->rank);
      GB_message(GB_ERROR,place,text); 
      return;
    }

  temp = (float *)malloc(elements*sizeof(float));

  if(temp==NULL) 
    {
      /** error **/
      sprintf(text,"Couldn't allocate space for %d floats",elements);
      GB_message(GB_ERROR,place,text); 
      return;
    }

  /* size = number of floats per array element */
  if(desc->type==GD_MESH)size=1;
  else if(desc->type==GD_POINT_TYPE)size=4;
  else if(desc->type==GD_VECTOR)size=7;
  else 
    {
      /** error **/
      sprintf(text,"Invalid data type: %d",desc->type);
      GB_message(GB_ERROR,place,text); 
      return;
    }

  desc->values = CM_allocate_heap_field(size*32);
  desc->context = CM_allocate_heap_field(1);

  /* set up proper shape for one or two dimension arrays */
  for(i=0;i<2;i++)shape[i]=1;
  for(i=0;i<desc->rank;i++)shape[i]=desc->fe_shape[i];

  /* copy each element in turn of fe_data array into temp array */
  /* and then onto CM */
  for(offset=0;offset<size;offset++){
    point=(float *)desc->fe_data;
    point+= offset;
    for(j=0;j<shape[1];j++) {
	for(i=0;i<shape[0];i++) 
	  *(temp+j*desc->geometry[0]+i) = 
	    *(point+size*(j*desc->fe_shape[0]+i)); 
	for(i=shape[0];i<desc->geometry[0];i++) 
	  *(temp+j*desc->geometry[0]+i) = 0.0;
      }
    for(j=shape[1];j<desc->geometry[1];j++) 
      for(i=0;i<desc->geometry[0];i++) 
	*(temp+j*desc->geometry[0]+i) = 0.0;

    CM_f_write_to_news_array_1L
      (temp,array_offset,start_vector,desc->geometry,axis_vector,
       CM_add_offset_to_field_id(desc->values,32*offset),
       23,8,desc->rank,desc->geometry,CM_float_single);
  }
  /* create a front end mask and use it to set the context field 
     associated with desc->values to only those positions corresponding
     to elements in the front end array */
  
  for(j=0;j<shape[1];j++) 
    {
      for(i=0;i<shape[0];i++) 
	*(temp+i+(j*desc->geometry[0])) = 1.0;
      for(i=shape[0];i<desc->geometry[0];i++) 
	*(temp+j*desc->geometry[0]+i) = 0.0;
    }
  for(j=shape[1];j<desc->geometry[1];j++) 
    for(i=0;i<desc->geometry[0];i++) 
      *(temp+j*desc->geometry[0]+i) = 0.0;

  context_temp = CM_allocate_stack_field(32);
  
  CM_f_write_to_news_array_1L
    (temp,array_offset,start_vector,desc->geometry,axis_vector,
     context_temp,23,8,desc->rank,desc->geometry,CM_float_single);

  CM_f_eq_constant_1L(context_temp,1.0,23,8);
  CM_store_test(desc->context);
  CM_deallocate_stack_through(context_temp);

  if(desc->print_stats){
    context_temp = CM_allocate_stack_field(1);
    CM_store_context(context_temp);
    CM_load_context(desc->context);
    GD_transfer_statistics(size,desc->values);
    printf("context is %d\n",CM_global_count_context());
    CM_load_context(context_temp); 
  }

}


