

/*  File Header  --------------------------------
    File Name:   globe_view.c
    Author and Date:  Gary Meeker  6/18/90
    Description:  Performs oblique perspective map projection. Contains
                  both serial and parallel code descriptions for both
		  the forward and reverse transformations.
    End Header   --------------------------------  */


#include <cm/paris.h>
#include <math.h>
#include <geodata.h>
#include <geodata_private.h>

#define PI 3.14159265358979323846
#define TFOV 0.7

/*  Function Header  ----------------------------
    Description:  Performs oblique perspective projection on front end
                  machine.
    Parameters:
                data: Array of float containing (lat, lon, alt)
		parameters: Array of float containing projection parameters.
		            The parameters are the center point latitude phi0;
			    the center point longitude lambda0, the distance
			    of the perspective point from the center of the
			    earth P in earth radius units, The angle gamma
			    east of north in a plane perpendicular to the line
			    from the perspective point to the center, and the
			    angle omega which is the downward tilt of the 
			    plane of projection with respect to perpendicular
			    plane. 
    Preconditions:  Data is in latitude, longitude, altitude form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly.
    Logic:
    End Header       ----------------------------  */

void GD_fe_globe_view(data, success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  float kp, clim, altmin, cosc, x, y, c, A, yoff, H, theta, s;
  int able_to_plot;
  double max, rmax, R;
   
  *success = 1;
  R = 1.0 + data[2]/GS_EARTH_RADIUS;
  clim = acos(1.0/parameters[2]);
  cosc = sin(parameters[0])*sin(data[0]) + cos(parameters[0])*cos(data[0])*
         cos(data[1] - parameters[1]);
  c = acos(cosc);
  kp = (parameters[2]-1)/(parameters[2]-R*cosc);
  if (c < clim) able_to_plot = 1;
  else 
    {
      if ((c - clim) >= PI/2) able_to_plot = 0.0;
      else
	{
	  altmin = 1.0/cos(c - clim) - 1.0;
	  if (data[2]/GS_EARTH_RADIUS >= altmin) able_to_plot = 1;
	  else able_to_plot = 0;
	}
    }
  
  x = R*kp*cos(data[0])*sin(data[1] - parameters[1]);
  y = R*kp*(cos(parameters[0])*sin(data[0]) - sin(parameters[0])*
	    cos(data[0])*cos(data[1] - parameters[1]));

  if (!able_to_plot)
    {
/*Map out of range points to circle corresponding to the horizon*/
      rmax = R*sqrt((parameters[2] - 1)/(parameters[2] + 1));
     /* rmax = (parameters[2] - 1)/tan(clim);*/
      theta = atan2(y, x);
      y = rmax*sin(theta);
      x = rmax*cos(theta);
      *success = 0;
    }


/*Translate coordinates to tilted plane*/
  max =  (parameters[2] - 1.0)*cos(parameters[4])*TFOV;
  /*yoff puts y = 0 at point on tilted plane where line through viewing
    position is perpendicular to the plane*/
  yoff = (parameters[2] - 1.0)*sin(parameters[4]);
  H = R*(parameters[2] - 1);
  A = (y*cos(parameters[3]) + x*sin(parameters[3]))*sin(parameters[4])/H +
      cos(parameters[4]);
  data[0] = (x*cos(parameters[3]) - y*sin(parameters[3]))*cos(parameters[4])/A;
  data[1] = (y*cos(parameters[3]) + x*sin(parameters[3]))/A - yoff;
  able_to_plot = 1;

/*Field of view clipping*/
  if (data[0] > max) *success= 0;
  if (data[0] < -max) *success = 0;
  if (data[1] > max) *success = 0;
  if (data[1] < -max) *success = 0;

  data[0] = data[0]/max;
  data[1] = data[1]/max;
}


/*  Function Header  ----------------------------
    Description:  Performs inverse perspective projection on front end
                  machine.
    Parameters:
                data: Array of float containing (lat, lon, alt)
		parameters: Array of float containing projection parameters.
		            The parameters are the center point latitude phi0;
			    the center point longitude lambda0, the distance
			    of the perspective point from the center of the
			    earth P in earth radius units, The angle gamma
			    east of north in a plane perpendicular to the line
			    from the perspective point to the center, and the
			    angle omega which is the downward tilt of the 
			    plane of projection with respect to perpendicular
			    plane. 
    Preconditions:  Data is in latitude, longitude, altitude form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly.
    Logic:
    End Header       ----------------------------  */


void GD_fe_inverse_globe_view(data, success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  float  c, s4, sc, p1, max;
  float  rho, x, y, R, H, M, Q;
  

  *success =1;
  p1 = parameters[2] - 1.0;
  s4 = sin(parameters[4]);
  max = (p1)*cos(parameters[4])*TFOV;
  data[0] = data[0]*max;
  data[1] = (data[1]*max) + ((p1)*s4);
  R = 1.0 + data[2]/GS_EARTH_RADIUS;
  H = R*(p1);
  M = H*data[0]/(H - data[1]*s4);
  Q = H*data[1]*cos(parameters[4])/(H - data[1]*s4);

/*First translate back to vertical plane*/
  x = M*cos(parameters[3]) + Q*sin(parameters[3]);
  y = Q*cos(parameters[3]) - M*sin(parameters[3]);

/*Transform back to lat, lon*/
  rho = sqrt(x*x + y*y);
  if (rho > (R*p1*sqrt(parameters[2]*parameters[2] - 1.0)/
      (parameters[2]*parameters[2] - R) + 0.000001))
    {
/*      printf("%3.20f\n" ,R*p1*sqrt(parameters[2]*parameters[2] - 1.0)/
      (parameters[2]*parameters[2] - R));*/
/*      printf("-> %3.20f\n", rho);*/
      rho = 0.0;
      *success = 0;
    }
  if (rho == 0.0)
    {
      data[0] = parameters[0];
      data[1] = parameters[1];
    }
  else
    {
      M = parameters[2]/R - sqrt(1.0 - rho*rho*((parameters[2]*parameters[2])/
						(R*R) - 1.0)/(p1*p1));
      Q = (p1)/rho + rho/(p1);
      sc = M/Q;
      c = asin(sc);
      if (parameters[2] < 0.0)
	if (rho > R*(p1)/parameters[2])
	  c = PI/2 - c;
      data [0] = asin(cos(c)*sin(parameters[0]) + y*sc*cos(parameters[0])/
		      rho);
      data[1] = parameters[1] + atan2(x*sc,(rho*cos(parameters[0])*cos(c)- 
					       y*sin(parameters[0])*sc));
    }
}



/* Function Header ---------------------------- 
	Description: 
   		Performs oblique perspective projection on connection
		machine.  
	Parameters: 
		data: 	CM_field_id_t pointing to an array of
			float containing (lat, lon, alt)
	parameters: 	Array of float containing projection
			parameters.
		The parameters are: 
			the center point latitude phi0; 
			the center point longitude lambda0;
			the distance of the perspective point from the
			  center of the earth P in earth radius units,
			The angle gamma east of north in a plane
			  perpendicular to the line from the
			  perspective point to the center, 
			the angle omega which is the downward tilt of
			  the plane of projection with respect to
			  perpendicular plane.  
	Preconditions: 	Data is in latitude, longitude, altitude form.
	Postconditions: None.  
	Return value: 	Performs operation on memory
			elements so no value returned explicitly.  
	Logic: 

End Header---------------------------- */

void GD_globe_view (data, success, parameters)
     CM_field_id_t  data;
     CM_field_id_t  success;
     float  *parameters;
{
  CM_field_id_t  x = CM_allocate_stack_field(64*3);
  CM_field_id_t  y = CM_add_offset_to_field_id (x,64);
  CM_field_id_t  alt = CM_add_offset_to_field_id(y, 64);
  CM_field_id_t  R = CM_allocate_stack_field(64);
  CM_field_id_t  c = CM_allocate_stack_field(64);
  CM_field_id_t  plotable = CM_allocate_stack_field(1);
  CM_field_id_t  kp = CM_allocate_stack_field(64);
  CM_field_id_t  theta = CM_allocate_stack_field(64);
  CM_field_id_t  xp = CM_allocate_stack_field(64);
  CM_field_id_t  yp = CM_allocate_stack_field(64);
  CM_field_id_t  saved_context = CM_allocate_stack_field(1);
  float clim, yoff;
  double max, rmax;

  CM_f_move_2L (x,data,52,11,23,8);
  CM_f_move_2L (y,data+32,52,11,23,8);
  CM_f_move_2L (alt,data+64,52,11,23,8);

  CM_u_move_constant_1L(success, 1, 1);
  /*compute all constants first*/

  clim = acos(1.0/parameters[2]);
  rmax = sqrt((parameters[2] - 1)/(parameters[2] + 1));
  max =  (parameters[2] - 1.0)*cos(parameters[4])*TFOV;
  yoff = (parameters[2] - 1.0)*sin(parameters[4]);
  CM_f_divide_constant_3_1L(R, alt, GS_EARTH_RADIUS, 52, 11);
  CM_f_add_constant_2_1L(R, 1.0, 52, 11);
/*Use kp as temp field for cos(c) calculation*/
  CM_f_subtract_constant_3_1L(kp, y, parameters[1], 52, 11);
  CM_f_cos_1_1L(kp, 52, 11);
  CM_f_cos_2_1L(c, x, 52, 11);
  CM_f_multiply_2_1L(kp, c, 52, 11);
  CM_f_multiply_constant_2_1L(kp, cos(parameters[0]), 52, 11);
  CM_f_sin_2_1L(c, x, 52, 11);
  CM_f_multiply_constant_2_1L(c, sin(parameters[0]), 52, 11);
  CM_f_add_2_1L(c, kp, 52, 11);

/*Compute kp*/
  CM_f_multiply_3_1L(kp, R, c, 52, 11);
  CM_f_subfrom_constant_2_1L(kp, parameters[2], 52, 11);
  CM_f_divinto_constant_2_1L(kp, (parameters[2] - 1.0), 52, 11);
  CM_f_acos_1_1L(c, 52, 11);

/*determine if point if visible*/
  CM_u_move_zero_1L(plotable, 1);
  CM_store_context(saved_context);
  CM_f_lt_constant_1L(c, clim, 52, 11);
  CM_logand_context_with_test();
  CM_u_move_constant_1L(plotable, 1, 1);
  CM_invert_context();                       /*else condition*/
  CM_logand_context(saved_context);          /*c>clim*/
  CM_f_lt_constant_1L(c, (PI/2 + clim), 52, 11); /*c-clim<90*/
  CM_logand_context_with_test();
/*Use xp to temporarily hold min altitude*/
  CM_f_subtract_constant_3_1L(xp, c, clim, 52, 11);
  CM_f_cos_1_1L(xp, 52, 11);
  CM_f_divinto_constant_2_1L(xp, 1.0, 52, 11);
  CM_f_subtract_constant_2_1L(xp, 1.0, 52, 11);
  CM_f_multiply_constant_2_1L(xp, GS_EARTH_RADIUS, 52, 11);
  CM_f_ge_1L(alt, xp, 52, 11);               /*alt > min alt*/
  CM_logand_context_with_test();
  CM_u_move_constant_1L(plotable, 1, 1);
  CM_load_context(saved_context);

/*compute x and y projected onto the vertical plane*/
  CM_f_subtract_constant_3_1L(xp, y, parameters[1], 52, 11);
  CM_f_sin_1_1L(xp, 52, 11);
  /*Use theta field temporarily in computation*/
  CM_f_cos_2_1L(theta, x, 52, 11);
  CM_f_multiply_2_1L(xp, theta, 52, 11);
  CM_f_multiply_2_1L(xp, kp, 52, 11);
  CM_f_multiply_2_1L(xp, R, 52, 11);
  
  CM_f_subtract_constant_3_1L(yp, y, parameters[1], 52, 11);
  CM_f_cos_1_1L(yp, 52, 11);
  CM_f_multiply_2_1L(yp, theta, 52, 11);
  CM_f_multiply_constant_2_1L(yp, sin(parameters[0]), 52, 11);
  CM_f_sin_2_1L(theta, x, 52, 11);
  CM_f_multiply_constant_2_1L(theta, cos(parameters[0]), 52, 11);
  CM_f_subfrom_2_1L(yp, theta, 52, 11);
  CM_f_multiply_2_1L(yp, kp, 52, 11);
  CM_f_multiply_2_1L(yp, R, 52, 11);


/*Map points that cannot be plotted to a circle corresponding to the horizon*/
  CM_store_context(saved_context);
  CM_u_eq_constant_1L(plotable, 0, 1);
  CM_logand_context_with_test();
  CM_f_atan2_3_1L(theta, yp, xp, 52, 11);
  CM_f_sin_2_1L(yp, theta, 52, 11);
  CM_f_cos_2_1L(xp, theta, 52, 11);
  CM_f_multiply_constant_2_1L(yp, rmax, 52, 11);
  CM_f_multiply_constant_2_1L(xp, rmax, 52, 11);
  CM_u_move_constant_1L(success, 0, 1);
  CM_load_context(saved_context);
  

/*Translate coordinates to tilted plane coordinates*/
  /*Use theta as a temporary field here also*/
  /*kp = H, c = A*/
  CM_f_multiply_constant_3_1L(kp, R, (parameters[2] - 1.0), 52, 11);
  CM_f_multiply_constant_3_1L(c, yp, cos(parameters[3]), 52, 11);
  CM_f_multiply_constant_3_1L(theta, xp, sin(parameters[3]), 52, 11);
  CM_f_add_2_1L(c, theta, 52, 11);
  CM_f_multiply_constant_2_1L(c, sin(parameters[4]), 52, 11);
  CM_f_divide_2_1L(c, kp, 52, 11);
  CM_f_add_constant_2_1L(c, cos(parameters[4]), 52, 11);

  CM_f_multiply_constant_3_1L(x, xp, cos(parameters[3]), 52, 11);
  CM_f_multiply_constant_3_1L(theta, yp, sin(parameters[3]), 52, 11);
  CM_f_subtract_2_1L(x, theta, 52, 11);   
  CM_f_multiply_constant_2_1L(x, cos(parameters[4]), 52, 11);
  CM_f_divide_2_1L(x, c, 52, 11);
  
  CM_f_multiply_constant_3_1L(y, yp, cos(parameters[3]), 52, 11);
  CM_f_multiply_constant_3_1L(theta, xp, sin(parameters[3]), 52, 11);
  CM_f_add_2_1L(y, theta, 52, 11);
  CM_f_divide_2_1L(y, c, 52, 11);
  CM_f_subtract_constant_2_1L(y, yoff, 52, 11);
  

/*Check for field of view limits*/
  CM_f_gt_constant_1L(x, max, 52, 11);
  CM_logand_context_with_test();
  CM_u_move_zero_1L(success, 1);
  CM_load_context(saved_context);
  CM_f_lt_constant_1L(x, -max, 52, 11);
  CM_logand_context_with_test();
  CM_u_move_zero_1L(success, 1);
  CM_load_context(saved_context);
  CM_f_gt_constant_1L(y, max, 52, 11);
  CM_logand_context_with_test();
  CM_u_move_zero_1L(success, 1);
  CM_load_context(saved_context);
  CM_f_lt_constant_1L(y, -max, 52, 11);
  CM_logand_context_with_test();
  CM_u_move_zero_1L(success, 1);
  CM_load_context(saved_context);

/*scale output*/
  CM_f_multiply_constant_2_1L(x, 1.0/max, 52, 11);
  CM_f_multiply_constant_2_1L(y, 1.0/max, 52, 11);

  CM_f_move_2L (data,x,23,8,52,11);
  CM_f_move_2L (data+32,y,23,8,52,11);
  CM_f_move_2L (data+64,alt,23,8,52,11);
  CM_deallocate_stack_through(x);
}



/*  Function Header  ----------------------------
    Description:  Performs inverse perspective projection on front end
                  machine.
    Parameters:
                data: CM_field_id_t pointing to an array of float 
		       containing (lat, lon, alt)
		parameters: Array of float containing projection parameters.
		            The parameters are the center point latitude phi0;
			    the center point longitude lambda0, the distance
			    of the perspective point from the center of the
			    earth P in earth radius units, The angle gamma
			    east of north in a plane perpendicular to the line
			    from the perspective point to the center, and the
			    angle omega which is the downward tilt of the 
			    plane of projection with respect to perpendicular
			    plane. 
    Preconditions:  Data is in latitude, longitude, altitude form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly.
    Logic:
    End Header       ----------------------------  */

void GD_inverse_globe_view(data, success, parameters)
     CM_field_id_t  data;
     CM_field_id_t  success;
     float  *parameters;
{
  float  yoff;
  double max, rmax;
  CM_field_id_t  lat = data;
  CM_field_id_t  lon = CM_add_offset_to_field_id(data, 32);
  CM_field_id_t  alt = CM_add_offset_to_field_id(lon, 32);
  CM_field_id_t  R = CM_allocate_stack_field(32);
  CM_field_id_t  H = CM_allocate_stack_field(32);
  CM_field_id_t  M = CM_allocate_stack_field(32);
  CM_field_id_t  Q = CM_allocate_stack_field(32);
  CM_field_id_t  x = CM_allocate_stack_field(32);
  CM_field_id_t  y = CM_allocate_stack_field(32);
  CM_field_id_t  rho = CM_allocate_stack_field(32);
  CM_field_id_t  saved_context = CM_allocate_stack_field(1);
  CM_field_id_t  temp_context = CM_allocate_stack_field(1);

  CM_u_move_constant_1L(success, 1, 1);
/*compute necessary constants first*/
  max = (parameters[2] - 1.0)*cos(parameters[4])*TFOV;
  yoff = (parameters[2] - 1.0)*sin(parameters[4]);
/*Rescale x, y data*/
  CM_f_multiply_constant_2_1L(lat, max, 23, 8);
  CM_f_mult_const_add_const_1L(lon, lon, max, yoff, 23, 8);
  CM_f_divide_constant_3_1L(R, alt, GS_EARTH_RADIUS, 23, 8);
  CM_f_add_constant_2_1L(R, 1.0, 23, 8);
  CM_f_multiply_constant_3_1L(H, R, (parameters[2] - 1.0), 23, 8);
/*use x for temporary storage here*/
  CM_f_multiply_constant_3_1L(x, lon, sin(parameters[4]), 23, 8);
  CM_f_subfrom_2_1L(x, H, 23, 8);
  CM_f_divide_3_1L(M, lat, x, 23, 8);
  CM_f_multiply_2_1L(M, H, 23, 8);
  
  CM_f_divide_3_1L(Q, lon, x, 23, 8);
  CM_f_multiply_2_1L(Q, H, 23, 8);
  CM_f_multiply_constant_2_1L(Q, cos(parameters[4]), 23, 8);
  
/*Translate coordinates back to the vertical plane*/
/*Use rho temporarily in computations*/
  CM_f_multiply_constant_3_1L(rho, Q, sin(parameters[3]), 23, 8);
  CM_f_multiply_constant_3_1L(x, M, cos(parameters[3]), 23, 8);
  CM_f_add_2_1L(x, rho, 23, 8);
  
  CM_f_multiply_constant_3_1L(rho, M, sin(parameters[3]), 23, 8);
  CM_f_multiply_constant_3_1L(y, Q, cos(parameters[3]), 23, 8);
  CM_f_subtract_2_1L(y, rho, 23, 8);
  
/*Transform back to lat, lon*/
/*Store rmax in M*/
  rmax = (parameters[2] - 1.0)*sqrt(parameters[2]*parameters[2] - 1.0);
  CM_f_subfrom_constant_3_1L(M, R, parameters[2]*parameters[2], 23, 8);
  CM_f_divinto_constant_2_1L(M, rmax, 23, 8);
  CM_f_multiply_2_1L(M, R, 23, 8);

  CM_f_multiply_3_1L(H, y, y, 23, 8);
  CM_f_multiply_3_1L(rho, x, x, 23, 8);
  CM_f_add_2_1L(rho, H, 23, 8);
  CM_f_sqrt_1_1L(rho, 23, 8);
  CM_store_context(saved_context);
/*Check for rho too large */
  CM_f_gt_1L(rho, M, 23, 8);
  CM_logand_context_with_test();
  CM_f_move_constant_1L(rho, 0.0, 23, 8);
  CM_u_move_constant_1L(success, 0, 1);
  CM_load_context(saved_context);
  
/*check for center (0,0) point*/  
  CM_f_eq_zero_1L(rho, 23, 8);
  CM_logand_context_with_test();
  CM_f_move_constant_1L(lat, parameters[0], 23, 8);
  CM_f_move_constant_1L(lon, parameters[1], 23, 8);
/*invert context for else*/
  CM_invert_context();    
  CM_logand_context(saved_context);
  
/*Compute the lat, lon for all other cases*/
  CM_f_multiply_3_1L(H, R, R, 23, 8);
  CM_f_divinto_constant_2_1L(H, parameters[2]*parameters[2], 23, 8);
  CM_f_subtract_constant_2_1L(H, 1.0, 23, 8);
  CM_f_divide_constant_2_1L(H, (parameters[2] - 1.0)*(parameters[2] - 1.0),
			    23, 8);
  CM_f_multiply_3_1L(M, rho, rho, 23, 8);
  CM_f_multiply_2_1L(M, H, 23, 8);
  CM_f_subfrom_constant_2_1L(M, 1.0, 23, 8);
  CM_f_sqrt_1_1L(M, 23, 8);

  CM_f_divinto_constant_3_1L(H, R, parameters[2], 23, 8);
  CM_f_subfrom_2_1L(M, H, 23, 8);
  

  CM_f_divide_constant_3_1L(H, rho, (parameters[2] - 1.0), 23, 8);
  CM_f_divinto_constant_3_1L(Q, rho, (parameters[2] - 1.0), 23, 8);
  CM_f_add_2_1L(Q, H, 23, 8);
  
  CM_f_divide_3_1L(H, M, Q, 23, 8);
  CM_f_asin_1_1L(H, 23, 8);

/*Check to insure that angle H is correct*/
  if (parameters[2] < 0.0)
    {
      CM_store_context(temp_context);
      CM_f_multiply_constant_2_1L(R, (parameters[2] - 1.0)/parameters[2], 23,
				  8);
      CM_f_gt_1L(rho, R, 23, 8);
      CM_logand_context_with_test();
      CM_f_subfrom_constant_2_1L(H, PI/2, 23, 8);
      CM_load_context(temp_context);
    }
 /*Compute resulting values*/
  /*Use M as a temporary field for the calculation*/
  CM_f_sin_2_1L(M, H, 23, 8);
  CM_f_multiply_2_1L(M, y, 23, 8);
  CM_f_divide_2_1L(M, rho, 23, 8);
  CM_f_multiply_constant_2_1L(M, cos(parameters[0]), 23,8);
  CM_f_cos_2_1L(lat, H, 23, 8);
  CM_f_multiply_constant_2_1L(lat, sin(parameters[0]), 23, 8);
  CM_f_add_2_1L(lat, M, 23, 8);
  CM_f_asin_1_1L(lat, 23, 8);
  
  CM_f_sin_2_1L(M, H, 23, 8);
  CM_f_multiply_2_1L(M, y, 23, 8);
  CM_f_multiply_constant_2_1L(M, sin(parameters[0]), 23, 8);
  CM_f_cos_2_1L(lon, H, 23, 8);
  CM_f_multiply_2_1L(lon, rho, 23, 8);
  CM_f_multiply_constant_2_1L(lon, cos(parameters[0]), 23, 8);
  CM_f_subtract_2_1L(lon, M, 23, 8);
  CM_f_sin_2_1L(M, H, 23, 8);
  CM_f_multiply_2_1L(M, x, 23, 8);
  CM_f_atan2_3_1L(lon, M, lon, 23, 8);
  CM_f_add_constant_2_1L(lon, parameters[1], 23, 8);
  
  CM_load_context(saved_context);
  CM_deallocate_stack_through(R);
}


void GD_globe_view_border(dest,parameters)
     CM_field_id_t dest;
     float *parameters;
{
  float radius =  sqrt((parameters[2] - 1)/(parameters[2] + 1))/
    ((parameters[2] - 1.0)*cos(parameters[4])*TFOV);
  
  CM_my_news_coordinate_1L (dest+96,0,32);
  CM_f_u_float_2_2L (dest+64,dest+96,32,23,8);
  CM_f_mult_const_add_const_1L (dest+64,dest+64,
				2.0*PI/65536.0,0.0,23,8);
  CM_f_cos_2_1L (dest,dest+64,23,8);
  CM_f_sin_2_1L (dest+32,dest+64,23,8);
  CM_f_multiply_constant_2_1L (dest+32,radius,23,8);
  CM_f_multiply_constant_2_1L (dest,radius,23,8);
  CM_get_from_news_always_1L (dest+128,dest,0,CM_upward,64);
  CM_f_move_constant_1L (dest+64,0.0,23,8);
  CM_f_move_constant_1L (dest+96,0.0,23,8);
  CM_f_move_constant_1L (dest+64+128,0.0,23,8);
  CM_f_move_constant_1L (dest+96+128,0.0,23,8);
}
