
/*  File Header  --------------------------------
    File Name:   gnomonic.c
    Author and Date:  Gary Meeker  6/26/90
    Description:  Performs gnomonic map projection. Contains
                  both serial and parallel code descriptions for both
		  the forward and reverse transformations.
    End Header   --------------------------------  */


#include <cm/paris.h>
#include <math.h>

#define PI 3.14159265358979323846
#define CMAX PI/3
#define RMAX 1.7320507764816284

/*  Function Header  ----------------------------
    Description:  Performs gnomonic projection on front end
                  machine. 

    Parameters:
                data: Array of float containing (lat, lon)
		success: 
		parameters: Array of float containing projection parameters.
                            The parameters are the center point latitude phi0,
			    and the center point longitude lambda0.
    Preconditions:  Data is in latitude, longitude form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly.
    Logic: 
    End Header       ----------------------------  */

void GD_fe_gnomonic(data, success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  float c, cosc, kp, theta, lat, lon;

 *success = 1;
/* Compute some constants first */
  cosc = sin(parameters[0])*sin(data[0]) + cos(parameters[0])*cos(data[0])*
    cos(data[1] - parameters[1]);
  kp = 1.0/cosc;
  c = acos(cosc);

/* Check for points which will be projected beyond the limits*/
  if (c > PI/2)
    {
      data[0] = RMAX;
      data[1] = 0.0;
      *success = 0;
    }
 else
   {
     lat = data[0];
     lon = data[1];
     data[0] = kp*cos(lat)*sin(lon - parameters[1]);
     data[1] = kp*(cos(parameters[0])*sin(lat) - sin(parameters[0])*
		   cos(lat)*cos(lon - parameters[1]));
   }

/*map out of range points to circle of max radius*/
  if (c > CMAX)
    {
      theta = atan2(data[1], data[0]);
      data[0] = RMAX*cos(theta);
      data[1] = RMAX*sin(theta);
      *success = 0;
    }

/*Scale data for output*/
  data[0] = data[0]/RMAX;
  data[1] = data[1]/RMAX;
}


/*  Function Header  ----------------------------
    Description:  Performs inverse gnomonic projection on front end
                  machine.
    Parameters:
                data: Array of float containing (x, y)
		parameters: Array of float containing projection parameters.
		            The parameters are the center point latitude phi0, 
			    and the center point longitude lambda0. 
    Preconditions:  Data is in 2-D rectangular form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in (lat,lon) form.
    Logic:
    End Header       ----------------------------  */

void GD_fe_inverse_gnomonic(data,success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  float rho, c, theta, x, y, sc, cc, s1, c1;
  
  *success = 1;
  data[0] *=RMAX;
  data[1] *=RMAX;
  rho = sqrt(data[0]*data[0] + data[1]*data[1]);
  if (rho > RMAX)
    {
      rho = RMAX;
      theta = atan2(data[1], data[0]);
      data[0] = rho*cos(theta);
      data[1] = rho*sin(theta);
      *success = 0;
    }
  else if (rho == 0)
    {
      data[0] = parameters[0];
      data[1] = parameters[1];
    }
  else
    {
      c = atan(rho);
      sc = sin(c);
      cc = cos(c);
      s1 = sin(parameters[0]);
      c1 = cos(parameters[0]);
      x = data[0];
      y = data[1];
      data[0] = asin(cc*s1 + (y*sc*c1/rho));
      if (fabs(parameters[0]) == PI/2)
	{
	  if (parameters[0] == PI/2) y = -y;
	  data[1] = parameters[1] + atan2(x, y);
	}
      else
	data[1] = parameters[1] + atan2(x*sc,(rho*c1*
				        cc - y*s1*sc));
    }
	  
}


/*  Function Header  ----------------------------
    Description:  Performs gnomonic projection on connection
                  machine. Maps all data on the hemisphere opposite
		  the hemisphere tangent to the observation plane 
		  to the equator.
    Parameters:
                data: CM_field_id_t pointing to array containing (lat, lon)
		parameters: Array of float containing projection parameters.
		            The parameters are the center point latitude phi0,
			    and the center point longitude lambda0;
    Preconditions:  Data is in angular (lat, lon) form, and data field is in
                    current vp_set.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in rectangular form.
    Logic:
    End Header       ----------------------------  */


void GD_gnomonic (data, success, parameters)
     CM_field_id_t  data;
     CM_field_id_t  success;
     float  *parameters;
{
  CM_field_id_t x = data;
  CM_field_id_t y =  CM_add_offset_to_field_id(data, 32);
  CM_field_id_t c = CM_allocate_stack_field(32);
  CM_field_id_t kp = CM_allocate_stack_field(32);
  CM_field_id_t theta = CM_allocate_stack_field(32);
  CM_field_id_t lat = CM_allocate_stack_field(32);
  CM_field_id_t lon = CM_allocate_stack_field(32);
  CM_field_id_t saved_context = CM_allocate_stack_field(1);
  
  CM_u_move_constant_1L(success, 1, 1);
/*compute constants first*/
  CM_f_subtract_constant_3_1L(kp, y, parameters[1], 23, 8);
  CM_f_cos_1_1L(kp, 23, 8);
  CM_f_cos_2_1L(c, x, 23, 8);
  CM_f_multiply_2_1L(kp, c, 23, 8);
  CM_f_multiply_constant_2_1L(kp, cos(parameters[0]), 23, 8);
  CM_f_sin_2_1L(c, x, 23, 8);
  CM_f_multiply_constant_2_1L(c, sin(parameters[0]), 23, 8);
  CM_f_add_2_1L(c, kp, 23, 8);
  CM_f_divinto_constant_3_1L(kp, c, 1.0, 23, 8);
  CM_f_acos_1_1L(c, 23, 8);
  
/*check for points that cannot be projected*/
  CM_store_context(saved_context);
  CM_f_ge_constant_1L(c, PI/2, 23, 8);
  CM_logand_context_with_test();
  CM_f_move_constant_1L(x, RMAX, 23, 8);
  CM_f_move_zero_1L(y, 23, 8);
  CM_u_move_constant_1L(success, 0, 1);
  
/*Perform computation for other cases*/
  CM_invert_context();
  CM_logand_context(saved_context);
/*Use theta as a temp field here*/
  CM_f_move_1L(lat, x, 23, 8);
  CM_f_move_1L(lon, y, 23, 8);
  CM_f_subtract_constant_3_1L(theta, lon, parameters[1], 23, 8);
  CM_f_sin_1_1L(theta, 23, 8);
  CM_f_cos_1_1L(x, 23, 8);
  CM_f_multiply_2_1L(x, theta, 23, 8);
  CM_f_multiply_2_1L(x, kp, 23, 8);
  
  CM_f_subtract_constant_3_1L(theta, lon, parameters[1], 23, 8);
  CM_f_cos_1_1L(theta, 23, 8);
  CM_f_cos_2_1L(y, lat, 23, 8);
  CM_f_multiply_2_1L(theta, y, 23, 8);
  CM_f_multiply_constant_2_1L(theta, sin(parameters[0]), 23, 8);
  CM_f_sin_2_1L(y, lat, 23, 8);
  CM_f_multiply_constant_2_1L(y, cos(parameters[0]), 23, 8);
  CM_f_subtract_2_1L(y, theta, 23, 8);
  CM_f_multiply_2_1L(y, kp, 23, 8);
  
/*Check for points projected beyond the limits*/
  CM_load_context(saved_context);
  CM_f_gt_constant_1L(c, CMAX, 23, 8);
  CM_logand_context_with_test();
  CM_f_atan2_3_1L(theta, y, x, 23, 8);
  CM_f_cos_2_1L(x, theta ,23, 8);
  CM_f_multiply_constant_2_1L(x, RMAX, 23, 8);
  CM_f_sin_2_1L(y, theta, 23, 8);
  CM_f_multiply_constant_2_1L(y, RMAX, 23, 8);
  CM_u_move_constant_1L(success, 0, 1);
  CM_load_context(saved_context);
  
/*Scale data for output*/
  CM_f_divide_constant_2_1L(x, RMAX, 23, 8);
  CM_f_divide_constant_2_1L(y, RMAX, 23, 8);
}



/*  Function Header  ----------------------------
    Description:  Performs inverse gnomonic projection on connection
                  machine.
    Parameters:
                data: CM_field_id_t pointing to array containing 
		      2-D rectangular coordinates.
		parameters: Array of float containing projection parameters.
		            The parameters are the center point latitude phi0,
			    and the center point longitude  lambda0.
    Preconditions:  Data is in rectangular (x, y) form, and data 
                    field is in current vp_set.
    Postconditions:  None.
    Return value:  Performs operation on memory 
                   elements so no value returned explicitly.
		   Result is in angular (lat,lon) form.
    Logic:
    End Header       ----------------------------  */

void GD_inverse_gnomonic(data, success, parameters)
     CM_field_id_t  data;
     CM_field_id_t  success;
     float  *parameters;
{
  CM_field_id_t lat = data;
  CM_field_id_t lon =  CM_add_offset_to_field_id(data, 32);
  CM_field_id_t c = CM_allocate_stack_field(32);
  CM_field_id_t theta = CM_allocate_stack_field(32);
  CM_field_id_t rho = CM_allocate_stack_field(32);
  CM_field_id_t x = CM_allocate_stack_field(32);
  CM_field_id_t y = CM_allocate_stack_field(32);
  CM_field_id_t saved_context = CM_allocate_stack_field(32);
  
  CM_u_move_constant_1L(success, 1, 1);
  CM_f_multiply_constant_2_1L(lat, RMAX, 23, 8);
  CM_f_multiply_constant_2_1L(lon, RMAX, 23, 8);
/*Use theta as a temporary field*/
  CM_f_multiply_3_1L(theta, lon, lon, 23, 8);
  CM_f_multiply_3_1L(rho, lat, lat, 23, 8);
  CM_f_add_2_1L(rho, theta, 23, 8);
  CM_f_sqrt_1_1L(rho, 23, 8);
  
/*Check for rho > RMAX*/
  CM_store_context(saved_context);
  CM_f_gt_constant_1L(rho, RMAX, 23, 8);
  CM_logand_context_with_test();
  CM_f_move_constant_1L(rho, RMAX, 23, 8);
  CM_f_atan2_3_1L(theta, lon, lat, 23, 8);
  CM_f_cos_2_1L(lat, theta, 23, 8);
  CM_f_multiply_2_1L(lat, rho, 23, 8);
  CM_f_sin_2_1L(lon, theta, 23, 8);
  CM_f_multiply_2_1L(lon, rho, 23, 8);
  CM_u_move_constant_1L(success, 0, 1);
  CM_load_context(saved_context);
  
/*Check for rho = 0 which is center point*/
  CM_f_eq_zero_1L(rho, 23, 8);
  CM_logand_context_with_test();
  CM_f_move_constant_1L(lat, parameters[0], 23, 8);
  CM_f_move_constant_1L(lon, parameters[1], 23, 8);
  
/*Perform computations for points where rho != 0*/
  CM_invert_context();
  CM_logand_context(saved_context);
  CM_f_atan_2_1L(c, rho, 23, 8);
  CM_f_move_1L(x, lat, 23, 8);
  CM_f_move_1L(y, lon, 23, 8);
/*Use theta as temporary field here*/
  CM_f_sin_2_1L(theta, c, 23, 8);
  CM_f_multiply_2_1L(theta, y, 23, 8);
  CM_f_multiply_constant_2_1L(theta, cos(parameters[0]), 23, 8);
  CM_f_divide_2_1L(theta, rho, 23, 8);
  CM_f_cos_2_1L(lat, c, 23, 8);
  CM_f_multiply_constant_2_1L(lat, sin(parameters[0]), 23, 8);
  CM_f_add_2_1L(lat, theta, 23, 8);
  CM_f_asin_1_1L(lat, 23, 8);
/*determine if phi0 = 90*/
  if (fabs(parameters[0]) == PI/2)
    {
      if (parameters[0] == PI/2) 
	CM_logxor_constant_2_1L(CM_add_offset_to_field_id(y, 31), 1, 1);
      CM_f_atan2_3_1L(lon, x, y, 23, 8);
    }
  else
    {
      CM_f_sin_2_1L(theta, c, 23, 8);
      CM_f_multiply_2_1L(theta, y, 23, 8);
      CM_f_multiply_constant_2_1L(theta, sin(parameters[0]), 23, 8);
      CM_f_cos_2_1L(lon, c, 23, 8);
      CM_f_multiply_2_1L(lon, rho, 23, 8);
      CM_f_multiply_constant_2_1L(lon, cos(parameters[0]), 23, 8);
      CM_f_subtract_2_1L(lon, theta, 23, 8);
      CM_f_sin_2_1L(theta, c, 23, 8);
      CM_f_multiply_2_1L(theta, x, 23, 8);
      CM_f_atan2_3_1L(lon, theta, lon, 23, 8);
      CM_f_add_constant_2_1L(lon, parameters[1], 23, 8);
    }
}

