
/*  File Header  --------------------------------
    File Name:   lambert.c
    Author and Date:  Gary Meeker  6/7/90
    Description:  Performs lambert conformal conic map projection. Contains
                  both serial and parallel code descriptions for both
		  the forward and reverse transformations.
    End Header   --------------------------------  */


#include <cm/paris.h>
#include <math.h>

#define PI 3.14159265358979323846


/*  Function Header  ----------------------------
    Description:  Performs lambert conformal conic projection on front end
                  machine. Output is scaled to range from -1 to 1 in both 
		  the x and y dimensions.
    Parameters:
                data: Array of float containing (lat, lon)
		lconst: Array of float containing the projection constants
		        (n, F, rho0).
		parameters: Array of float containing projection parameters.
		            The parameters are the two standard parallels,
			    phi1, phi2; the center point latitude, phi0;
			    the center point longitude, lambda0; and the
			    scaling factor. The 6th element is a flag to
			    disable limit checks. A value of 1000 disables
			    the checks.
    Preconditions:  Data is in latitude, longitude form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly.
    Logic:   
    End Header       ----------------------------  */

void GD_fe_lambert_f(data,lconst, success, parameters)
     float *data, *lconst;
     int *success;
     float *parameters;
{
  float rho, theta;


  *success = 1;
  rho = lconst[1]/pow(tan(PI/4 + 0.5*data[0]), lconst[0]);
  theta = lconst[0]*(data[1]-parameters[3]);
  
  data[0] = parameters[4]*rho*sin(theta);
  data[1] = parameters[4]*(lconst[2]-rho*cos(theta));
  if (parameters[5] != 1000.0) 
    {
      if (data[0] > 1.0) 
	{
	  data[0] = 1.0;
	  *success = 0;
	} 
      if (data[0] < -1.0) 
	{
	  data[0] = -1.0;
	  *success = 0;
	}

      if (data[1] > 1.0)
	{
	  data[1] = 1.0;
	  *success = 0;
	}
      if (data[1] < -1.0)
	{
	  data[1] = -1.0;
	  *success = 0;
	}
    }
}


/*  Function Header  ----------------------------
    Description: Computes constants used in lambert conformal conic
                 projection, (n, F, rho0). It determines the center
		 latitude and longitude from the latitude and longitude
		 limits. Finally, it computes a scale factor to ensure 
		 that all output coordinates are within a square with 
		 center (0, 0) and sides of length 2.  This code runs on
		 the front end machine, but is used in both the front end
		 and connection machine versions of the map projection.
    Parameters:
                lconst: Array of float where computed constants will be stored.
		        The order of the stored constants is (n, F, rho).
		parameters: Array of float containing projection parameters.
		            The parameters are the two standard parallels,
			    phi1, phi2; the north and south latitude limits
			    nphi, sphi; the east and west longitude limits
			    elambda, wlambda.

    Preconditions: None
    Postconditions:  None.
    Return value: Returns constant values in lconst array, and modifies the 
                  values in the parameters array so that the third through 
		  fifth elements become the center latitude phi0, the 
		  center longitude lambda0, and the scale factor 
		  respectively. 
    Logic:
    End Header       ----------------------------  */

void GD_fe_lambert_scale(lconst, success, parameters)
     float *lconst;
     int *success;
     float *parameters;
{
  float c0, c1, c2, cs1, cs2, data[2], nphi, sphi,
        elambda, wlambda, xscale, yscale;
  
  nphi = parameters[2];
  sphi = parameters[3];
  elambda = parameters[4];
  wlambda = parameters[5];
/*set parameters to new values for use by other procedures*/
  parameters[2] = (nphi+sphi)/2;
  parameters[3] = (elambda+wlambda)/2;
  parameters[4] = 1.0; /*set scale to 1 for determining x,y limits*/
  parameters[5] = 1000.0; /*flag to disable limit check in fwd transformation*/
  c0 = PI/4 + 0.5*parameters[2];
  c1 = PI/4 + 0.5*parameters[0];
  c2 = PI/4 + 0.5*parameters[1];
  cs1 = cos(parameters[0]);
  cs2 = cos(parameters[1]);

  lconst[0] = log(cs1/cs2)/log(tan(c2)/tan(c1));
  lconst[1] = cs1*pow(tan(c2), lconst[0])/lconst[0];
  lconst[2] = lconst[1]/pow(tan(c0), lconst[0]);
/*Now check two extreme points on the box to determine scaling factor*/
  data[0] = sphi;
  data[1] = wlambda;
  GD_fe_lambert_f(data, lconst, success, parameters);
  xscale = fabs(1.0/data[0]);  
  data[0] = sphi;
  data[1] = parameters[3];
  GD_fe_lambert_f(data, lconst, success, parameters);
  yscale = fabs(1.0/data[1]);
  data[0] = nphi;
  data[1] = parameters[3];
  GD_fe_lambert_f(data, lconst, success, parameters);
  yscale = (yscale < fabs(1.0/data[1])) ? yscale : fabs(1.0/data[1]);
  parameters[4] = (xscale < yscale) ? xscale : yscale;
  parameters[5] = 0.0; /*reenable limit check*/
}



/*  Function Header  ----------------------------
    Description:  Performs inverse lambert conformal conic projection on 
                  front end machine.
    Parameters:
                data: Array of float containing (x, y)
		lconst: Array containing the projection parameters
		        (n, F, rho0).
		parameters: Array of float containing projection parameters.
		            The parameters are the two standard parallels,
			    phi1, phi2; the center point latitude, phi0;
			    the center point longitude, lambda0; and
			    the scale factor.
    Preconditions:  Data is in 2-D rectangular form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in (lat,lon) form.
    Logic:
    End Header       ----------------------------  */

void GD_fe_lambert_r(data,lconst, success, parameters)
     float *data, *lconst;
     int *success;
     float *parameters;
{
  float rho, theta, l;
  int neg;

  *success = 1;
  data[0] = data[0]/parameters[4];
  data[1] = data[1]/parameters[4];
  neg = (lconst[0]<0.0);
  l = lconst[2] - data[1];
  rho = sqrt(data[0]*data[0] + l*l);
  if (neg) rho = -rho;
  theta = atan2(data[0], l);

  data[0] = 2*atan(pow(lconst[1]/rho, 1/lconst[0])) - PI/2;
  data[1] = theta/lconst[0] + parameters[3];
}


/*  Function Header  ----------------------------
    Description:  Performs lambert conformal conic projection on 
                  front end machine.
    Parameters:
                data: Array of float containing (x, y)
		parameters: Array of float containing projection parameters.
		            The parameters are the two standard parallels,
			    phi1, phi2; the north and south latitude limits
			    nphi, sphi; the east and west longitude limits
			    elambda, wlambda.
    Preconditions:  Data is in 2-D rectangular form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in (lat,lon) form.
    Logic:
    End Header       ----------------------------  */

void GD_fe_lambert(data, success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  float lconst[3], params[6];
  int dummy;

  params[0] = parameters[0];
  params[1] = parameters[1];
  params[2] = parameters[2];
  params[3] = parameters[3];
  params[4] = parameters[4];
  params[5] = parameters[5];
  GD_fe_lambert_scale(lconst, &dummy, params);
  GD_fe_lambert_f(data, lconst, success, params);
}



/*  Function Header  ----------------------------
    Description:  Performs inverse lambert conformal conic projection on 
                  front end machine.
    Parameters:
                data: Array of float containing (lat, lon)
		parameters: Array of float containing projection parameters.
		            The parameters are the two standard parallels,
			    phi1, phi2; the north and south latitude limits
			    nphi, sphi; the east and west longitude limits
			    elambda, wlambda.
    Preconditions:  Data is in polar (lat, lon) form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in 2-D cartesian form
    Logic:
    End Header       ----------------------------  */

void GD_fe_inverse_lambert(data, success, parameters)
     float *data;
     int *success;
     float *parameters;
{
  float lconst[3], params[6];
  int dummy;
 
  params[0] = parameters[0];
  params[1] = parameters[1];
  params[2] = parameters[2];
  params[3] = parameters[3];
  params[4] = parameters[4];
  params[5] = parameters[5];
  GD_fe_lambert_scale(lconst, &dummy, params);
  GD_fe_lambert_r(data, lconst, success, params);
}
  
  
/*  Function Header  ----------------------------
    Description:  Performs  lambert conformal conic projection
                  on connection machine.
    Parameters:
                data: CM_field_id_t pointing to array containing (lat, lon)
		lconst: Array containing the projection parameters (n, F, rho0)
		parameters: Array of float containing projection parameters.
		            The parameters are the two standard parallels,
			    phi1, phi2; the center point latitude, phi0;
			    the center point longitude, lambda0; and the
			    scale factor. The 6th element is a flag to 
			    disable limit checking. A value of 1000 disables
			    the checks.
    Preconditions:  Data is in angular (lat, lon) form, and data field is in
                    current vp_set.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in rectangular form.
    Logic:
    End Header       ----------------------------  */


void GD_lambert_f (data, lconst, success, parameters)
     CM_field_id_t  data;
     float  *lconst;
     CM_field_id_t success;
     float *parameters;
{
  CM_field_id_t  x = data;
  CM_field_id_t  y = CM_add_offset_to_field_id (data,32);
  CM_field_id_t  rho = CM_allocate_stack_field(32);
  CM_field_id_t  theta = CM_allocate_stack_field(32);
  CM_field_id_t  saved_context = CM_allocate_stack_field(1);
  
  CM_u_move_constant_1L(success, 1, 1);
/* Use theta field as a temporary computation field*/
  CM_f_move_1L(theta, x, 23, 8);
  CM_f_mult_const_add_const_1L(theta, theta, 0.5, PI/4, 23, 8);
  CM_f_tan_1_1L(theta, 23, 8);
  CM_f_f_power_constant_2_1L(theta, lconst[0], 23, 8);
  CM_f_move_constant_1L(rho, lconst[1], 23, 8);
  CM_f_divide_2_1L(rho, theta, 23, 8);
  CM_f_move_1L(theta, y, 23, 8);
  CM_f_subtract_constant_2_1L(theta, parameters[3], 23, 8);
  CM_f_multiply_constant_2_1L(theta, lconst[0], 23, 8);
  CM_f_sin_2_1L(x, theta, 23, 8);
  CM_f_multiply_2_1L(x, rho, 23, 8);
  CM_f_cos_1_1L(theta, 23, 8);
  CM_f_multiply_2_1L(theta, rho, 23, 8);
  CM_f_move_constant_1L(y, lconst[2], 23, 8);
  CM_f_subtract_2_1L(y, theta, 23, 8);
  CM_f_multiply_constant_2_1L(x, parameters[4], 23, 8);
  CM_f_multiply_constant_2_1L(y, parameters[4], 23, 8);

  if (parameters[5] != 1000.0)
    {
      CM_store_context(saved_context);
      CM_f_gt_constant_1L(x, 1.0, 23, 8);
      CM_logand_context_with_test();
      CM_f_move_constant_1L(x, 1.0, 23, 8);
      CM_u_move_constant_1L(success, 0, 1);
      CM_load_context(saved_context);
      CM_f_lt_constant_1L(x, -1.0, 23, 8);
      CM_logand_context_with_test();
      CM_f_move_constant_1L(x, -1.0, 23, 8);
      CM_u_move_constant_1L(success, 0, 1);
      CM_load_context(saved_context);

      CM_f_gt_constant_1L(y, 1.0, 23, 8);
      CM_logand_context_with_test();
      CM_f_move_constant_1L(y, 1.0, 23, 8);
      CM_u_move_constant_1L(success, 0, 1);
      CM_load_context(saved_context);
      CM_f_lt_constant_1L(y, -1.0, 23, 8);
      CM_logand_context_with_test();
      CM_f_move_constant_1L(y, -1.0, 23, 8);
      CM_u_move_constant_1L(success, 0, 1);
      CM_load_context(saved_context);
    }
  CM_deallocate_stack_through(rho);
}

/*  Function Header  ----------------------------
    Description:  Performs inverse lambert conformal conic projection
                  on connection machine.
    Parameters:
                data: CM_field_id_t pointing to array containing 
		      2-D rectangular coordinates.
		lconst: Array containing the projection parameters 
		        (n, F, rho0)
		parameters: Array of float containing projection parameters.
		            The parameters are the two standard parallels,
			    phi1, phi2; the center point latitude, phi0;
			    the center point longitude, lambda0; and the 
			    scale factor.
    Preconditions:  Data is in rectangular (x, y) form, and data 
                    field is in current vp_set.
    Postconditions:  None.
    Return value:  Performs operation on memory 
                   elements so no value returned explicitly.
		   Result is in angular (lat,lon) form.
    Logic:
    End Header       ----------------------------  */

void GD_lambert_r(data, lconst, success, parameters)
     CM_field_id_t  data;
     float  *lconst;
     CM_field_id_t success;
     float *parameters;
{
  CM_field_id_t  lat = data;
  CM_field_id_t  lon = CM_add_offset_to_field_id(data, 32);
  CM_field_id_t  rho = CM_allocate_stack_field(32);
  CM_field_id_t  theta = CM_allocate_stack_field(32);  
  CM_field_id_t  l = CM_allocate_stack_field(32);
  CM_field_id_t  neg = CM_allocate_stack_field(1);

  CM_u_move_constant_1L(success, 1, 1);
  CM_f_multiply_constant_2_1L(lat, 1/parameters[4], 23,8);
  CM_f_multiply_constant_2_1L(lon, 1/parameters[4], 23, 8);
  if (lconst[0] < 0) CM_u_move_constant_1L(neg, 1, 1);
  else CM_u_move_constant_1L(neg, 0, 1);
  CM_f_subfrom_constant_3_1L(l, lon, lconst[2], 23, 8);
  CM_f_atan2_3_1L(theta, lat, l, 23, 8);
  CM_f_u_power_constant_2_1L(l, 2, 23, 8);
  CM_f_u_power_constant_3_1L(rho, lat, 2, 23, 8);
  CM_f_add_2_1L(rho, l, 23, 8);
  CM_f_sqrt_1_1L(rho, 23, 8);
  CM_logxor_2_1L(CM_add_offset_to_field_id(rho, 31), neg, 1);
  lconst[0] = 1.0/lconst[0];
  CM_f_divinto_constant_3_1L(lat, rho, lconst[1], 23, 8);
  CM_f_f_power_constant_2_1L(lat, lconst[0], 23, 8);
  CM_f_atan_1_1L(lat, 23, 8);
  CM_f_multiply_constant_2_1L(lat, 2.0, 23, 8);
  CM_f_subtract_constant_2_1L(lat, PI/2, 23, 8);
  CM_f_mult_const_add_const_1L(lon, theta, lconst[0], parameters[3], 23, 8);
  CM_deallocate_stack_through(rho);
}



/*  Function Header  ----------------------------
    Description:  Performs lambert conformal conic projection on 
                  connection machine.
    Parameters:
                data: CM_field_id_t pointing to an array of float
		      containing (x, y)
		parameters: Array of float containing projection parameters.
		            The parameters are the two standard parallels,
			    phi1, phi2; the north and south latitude limits
			    nphi, sphi; the east and west longitude limits
			    elambda, wlambda.
    Preconditions:  Data is in 2-D rectangular form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in (lat,lon) form.
    Logic:
    End Header       ----------------------------  */

void GD_lambert(data, success, parameters)
     CM_field_id_t data;
     CM_field_id_t success;
     float *parameters;
{
  float lconst[3], params[6];
  int dummy;

  params[0] = parameters[0];
  params[1] = parameters[1];
  params[2] = parameters[2];
  params[3] = parameters[3];
  params[4] = parameters[4];
  params[5] = parameters[5];
  GD_fe_lambert_scale(lconst, &dummy, params);
  GD_lambert_f(data, lconst, success, params);
}



/*  Function Header  ----------------------------
    Description:  Performs inverse lambert conformal conic projection on 
                  connection machine.
    Parameters:
		data: CM_field_id_t pointing to an array of float
		      containing (x, y)
		parameters: Array of float containing projection parameters.
		            The parameters are the two standard parallels,
			    phi1, phi2; the north and south latitude limits
			    nphi, sphi; the east and west longitude limits
			    elambda, wlambda.
    Preconditions:  Data is in polar (lat, lon) form.
    Postconditions:  None.
    Return value:  Performs operation on memory elements so no value
                   returned explicitly. Result is in 2-D cartesian form
    Logic:
    End Header       ----------------------------  */

void GD_inverse_lambert(data, success, parameters)
     CM_field_id_t data;
     CM_field_id_t success;
     float *parameters;
{
  float lconst[3], params[6];
  int dummy;

  params[0] = parameters[0];
  params[1] = parameters[1];
  params[2] = parameters[2];
  params[3] = parameters[3];
  params[4] = parameters[4];
  params[5] = parameters[5];
  GD_fe_lambert_scale(lconst, &dummy, params);
  GD_lambert_r(data, lconst, success, params);
}

/*void GD_lambert_border(dest,parameters)*/
/*     CM_field_id_t dest;*/
/*     float *parameters;*/
/*{*/
/*  CM_my_news_coordinate_1L (dest+96,0,16);*/
/*  CM_f_u_float_2_2L (dest,dest+96,14,23,8);*/
/*  CM_f_mult_const_add_const_1L (dest,dest,2.0/16384.0,-1.0,23,8);*/
/*  CM_f_move_constant_1L (dest+32,1.0,23,8);*/
/*  CM_u_move_1L (dest+32+31,dest+96+15,1);*/
/*  CM_load_context(dest+96+14);*/
/*  CM_swap_2_1L (dest,dest+32,32);*/
/*  CM_set_context();*/
/*  CM_f_multiply_constant_2_1L (dest+32,FACTOR,23,8);*/
/*  CM_get_from_news_always_1L (dest+128,dest,0,CM_upward,64);*/
/*  GD_lambert(data, success, parameters)*/
/*  CM_f_move_constant_1L (dest+64,0.0,23,8);*/
/*  CM_f_move_constant_1L (dest+96,0.0,23,8);*/
/*  CM_f_move_constant_1L (dest+64+128,0.0,23,8);*/
/*  CM_f_move_constant_1L (dest+96+128,0.0,23,8);*/
/*}*/
