/*   File Header   ---------------------------------

     File Name:       mem.c

     Author and Date: George Lamb--7/19/90

     Description: this file contains code to manage and allocate cm
     fields for transfer and to keep track of the amount of memory used by
     the dm.  When a resolution is first opened, a memory descriptor is
     created which tracks the resolution's use.  If the resolution is
     gridded, no cacheing is done.  Otherwise, the values are kept online
     in the resolution 'values' field until the DM takes up
     'GD_free_percent' of useable memory.  At this point, the least
     important values are trashed, to be reread in from files when
     requested again.

     Future Updates: Swapping should be done for client data, and
     cacheing should be done for gridded data, if an efficient algorithm
     can be found.

     variables defined:
       int   GD_bits_std_used
       int   GD_bits_arb_used
       int   GD_sets_std_loaded
       int   GD_sets_arb_loaded
       float GD_current_time
       float GD_free_percent

     functions defined:
       GD_mem_desc  *GD_new_mem_desc(res)
       GD_mem_desc  *GD_allocate_mem_desc(res)
       void          GD_update_mem_desc(md)
       int           GD_print_mem_list()
       void          GD_deallocate_mem_desc(md)
       float         GD_mem_eval(md)
       int           GD_free_memory()
       int           GD_out_of_memory()
       void          GD_mem_stats()
       int           GD_compute_bits(desc)
       CM_field      GD_cmalloc_std(type, number)
       CM_field      GD_cmalloc_arb(type, cs, number, bits, total_bits)
       void          GD_cmfree(field, bits)
     
End Header   ---------------------------------- */

#include <geodata.h>
#include <geodata_private.h>
#include <stdio.h>
#include <cm/paris.h>

int    GD_bits_std_used = 0, GD_bits_arb_used = 0,   /* per processor */
       GD_sets_std_loaded = 0, GD_sets_arb_loaded = 0;
float  GD_current_time=0, GD_free_percent=.25;




/*   Function Header   -----------------------------
     NAME:        GD_mem_stats
     DESCRIPTION: print out the memory usage
     PARAMETERS:  none
     PRECOND:     none
     POSTCOND:    a message is printed onto std out
     RETURN:      void
     LOGIC:       obvious
     End Header ---------------------------------- */

/* changed by eah to by readable by 'cc' */
void GD_mem_stats()
{
  char message[256];
  sprintf(message, "Geodata memory usage:\n%-5d source data sets,\n %-6d source bits/physical processor,\n %-5d cache  data sets, %-6d cache bits/physical processor,\n %-6.2f %% of total memory.\n",
	  GD_sets_std_loaded,
	  GD_bits_std_used, 
	  GD_sets_arb_loaded,
	  GD_bits_arb_used,(float) (GD_bits_std_used + GD_bits_arb_used) /
	  (float) (CM_physical_memory_limit) * 100.0);
  GB_message(GB_NPRINT, "GD_mem_stats", message);
}   /* end of GD_mem_stats */



/*   Function Header   -----------------------------
     NAME:        GD_out_of_memory
     DESCRIPTION: compare cached and src bits used to total memory
     PARAMETERS:  none
     PRECOND:     none
     POSTCOND:    none
     RETURN:      boolean
     LOGIC:       obvious
     End Header ---------------------------------- */

int GD_out_of_memory()
{ if (((float) (GD_bits_std_used + GD_bits_arb_used) / /*  Should return 1 here...  */
       (float) CM_physical_memory_limit) > GD_free_percent) { GD_mem_stats(); return(1);}
  else return(0); }



/*   Function Header   -----------------------------
     NAME:        GD_compute_bits
     DESCRIPTION: compute bits for a given desc.  If desc=NULL, 
     compute for a std form of the given type.
     PARAMETERS:  desc--a data desc
                  type--{POINT, MESH, VECTOR}
     PRECOND:     none
     POSTCOND:    none
     RETURN:      #bits
     LOGIC:       if desc==NULL, switch on type for std types
                  otherwise switch on desc->type & coord syst
     End Header ---------------------------------- */

int GD_compute_bits(desc, type)
     GD_data_desc *desc;
     int type;
{
  int size;

  if (desc == NULL) return((type == GD_VECTOR) ? 8*32 : 4*32);
  size = (desc->type == GD_VECTOR) ? 2 : 1;
  switch (desc->coord_syst){
  case GD_CART_2D: return ((desc->rank == 2) ? size*3*32 : size*4*32);
  case GD_MAP: return(size*3*32);
  default: return(size*4*32);
  }    
}      /* end of GD_compute_bits */



/*   Function Header   -----------------------------
     NAME:        GD_cmalloc_std
     DESCRIPTION: allocate a number of std sized heap field
     PARAMETERS:  type--{point, vector, mesh}
                  number--of fields
     PRECOND:     none
     POSTCOND:    number std sized heap field are allocated
     RETURN:      a ptr to the first field
     LOGIC:       
     1.  check the size of the cache and free space if necessary
     2.  compute the size of the std fields
     3.  update the cache bit counter and #bits loaded
     4.  CM allocate and return the new fields
     End Header ---------------------------------- */

CM_field_id_t GD_cmalloc_std (type, number)
     unsigned type, number;
{
  int size;
/*  if (GD_out_of_memory()) GD_free_memory(); */
  size = (type == GD_VECTOR) ? 8*32*number : 4*32*number;
/*  printf("Allocating %d bits in cmalloc_std\n",
         size * CM_geometry_total_vp_ratio(CM_vp_set_geometry(CM_current_vp_set))); */
/*  GD_mem_stats(); */
  GD_bits_std_used += size * 
    CM_geometry_total_vp_ratio(CM_vp_set_geometry(CM_current_vp_set));
  GD_sets_std_loaded++;
/*  GD_mem_stats(); */
  return(CM_allocate_heap_field (size));
}    /* end of GD_cmalloc_std */



/*   Function Header   -----------------------------
     NAME:        GD_cmalloc_arb
     DESCRIPTION: allocate a number of bit sized heap fields
     PARAMETERS:  number--of fields
                  bits--per field
     PRECOND:     none
     POSTCOND:    number*bits heap field are allocated
     RETURN:      a ptr to the first field
     LOGIC:       
     1.  check the size of the cache and free space if necessary
     2.  update the cache bit counter and #bits loaded
     4.  CM allocate and return the new fields
     End Header ---------------------------------- */

CM_field_id_t GD_cmalloc_arb (number, bits)
     unsigned number, bits;
{
  int tb = number * bits;
/*  if (GD_out_of_memory()) GD_free_memory();  */
/*  printf("Allocating %d bits in cmalloc_arb\n",
         tb * CM_geometry_total_vp_ratio(CM_vp_set_geometry(CM_current_vp_set)));*/
  GD_bits_arb_used += tb * 
    CM_geometry_total_vp_ratio(CM_vp_set_geometry(CM_current_vp_set));
  GD_sets_arb_loaded++;
  return(CM_allocate_heap_field (tb));
}    /* end of GD_cmalloc_arb */


/*   Function Header   -----------------------------
     NAME:        GD_cmfree
     DESCRIPTION: free the given field and update accounting by bits
     PARAMETERS:  field--ptr to a cm field
                  bits--total for the field
     PRECOND:     field is allocated on the cm
     POSTCOND:    field is deallocated, and bits loaded is updated
     RETURN:      void
     LOGIC:       
     1.  obvious, except field must be set to 0 after deallocating
     to insure it's not accessed again.
     End Header ---------------------------------- */

int GD_cmfree(field, bits)
     CM_field_id_t *field;
     int bits;
{
  GD_sets_std_loaded--;
/*  printf("Deallocating %d bits\n",bits *
    CM_geometry_total_vp_ratio(CM_vp_set_geometry(CM_field_vp_set(*field))));*/
/*  GD_mem_stats(); */
  GD_bits_std_used -= bits *
    CM_geometry_total_vp_ratio(CM_vp_set_geometry(CM_field_vp_set(*field)));
  if (CM_is_field_in_stack(*field)) CM_deallocate_stack_through(*field);
  else CM_deallocate_heap_field(*field);
/*   GD_mem_stats(); */
  *field = 0;
}     /* end of GD_cmfree */

