/*   File Header   ---------------------------------
     File Name: miller.c
     Author and Date: 
     Description: performs miller projection
     End Header   ---------------------------------- */


#include <geodata.h>
#include <geodata_private.h>
#include <cm/paris.h>
#include <math.h>


#define LAT_SCALE   1.8427300347011		/* ln(tan(0.45*PI)) */
#define LON_SCALE   GS_PI		        /* PI */
#define FACTOR 0.75

/*   Function Header   -----------------------------
     Description: FE miller projection
     Parameters: data   *float pointing to contiguous
                        array containing (lat,lon,alt)                        
                 parameters  *float to projection parameters

     Preconditions: in the vp_set of data, 
     				data in 3D angular  
     Postconditions: none
     Return Value: does operation in place
     Logic:
     End Header   ---------------------------------- */

void GD_fe_miller (data, success, parameters)	
	float *data;
        int *success;
        float *parameters;
{
	int neg = 0;
	float lat;
	
	*success = 1;
	lat = data[0];
	
 	data[0] = data[1] / LON_SCALE;
	
	if (lat<0) {
		neg = 1;
		lat = -lat;
	}
	data[1] = log(tan(0.4*lat + GS_PI/4.0)) / LAT_SCALE;
	if (neg) 
		data[1] = -data[1];
	data[1]*=FACTOR;
}	/* GD_fe_miller */


/*   Function Header   -----------------------------
     Description: FE inverse miller projection
     Parameters: data   *float pointing to contiguous
                        array containing (x,y,alt)                        
                 parameters  *float to projection parameters
     Preconditions: in the vp_set of data, 
     				data in 3D rectangular
     Postconditions: none
     Return Value: does operation in place
     Logic:
     End Header   ---------------------------------- */
     
void GD_fe_inverse_miller (data,success,parameters)	
	float *data;
        int *success;
        float *parameters;
{
	int neg = 0;
	float y;
	
	*success = 1;
	y = data[1];
	
 	data[1] = data[0] * LON_SCALE;
 		
 	if (y<0) {
		neg = 1;
		y = -y;
	}
 	data[0] = 2.5 * (atan(exp(y*LAT_SCALE)) - GS_PI/4.0);
	if (neg) 
		data[0] = -data[0];
	data[1]/=FACTOR;
}	/* GD_fe_inverse_miller */


/*   Function Header   -----------------------------
     Description: CM miller projection
     Parameters: data   CM_field_id_t pointing to contiguous
                        array containing (lat,lon,alt)
                 parameters  *float to projection parameters
     Preconditions: in the vp_set of data, 
     				data in 3D angular
     Postconditions: none
     Return Value: does operation in place
     Logic:
     End Header   ---------------------------------- */


void GD_miller (data,success,parameters)
	CM_field_id_t data;
        CM_field_id_t success;
	float *parameters;
{
	CM_field_id_t x = data;
	CM_field_id_t y = CM_add_offset_to_field_id (data,32);
  	CM_field_id_t sign = CM_allocate_stack_field (1);
  	
	CM_u_move_constant_1L(success, 1, 1);
  	CM_swap_2_1L (x,y,32);  	

        /*  Map longitudes into range -pi..pi  */
  	CM_store_context(sign);
        CM_f_lt_constant_1L(x,-GS_PI,23,8);
        CM_logand_context_with_test();
        while(CM_global_count_context()){
          CM_f_add_constant_2_1L(x,2*GS_PI,23,8);
          CM_f_lt_constant_1L(x,-GS_PI,23,8);
          CM_logand_context_with_test();
        }
        CM_load_context(sign);
        CM_f_gt_constant_1L(x,GS_PI,23,8);
        CM_logand_context_with_test();
        while(CM_global_count_context()){
          CM_f_subtract_constant_2_1L(x,2*GS_PI,23,8);
          CM_f_gt_constant_1L(x,GS_PI,23,8);
          CM_logand_context_with_test();
        }
        CM_load_context(sign);
  	
  	CM_f_multiply_constant_2_1L (x,1.0/LON_SCALE,23,8);

  	CM_f_lt_zero_1L (y,23,8);
  	CM_store_test(sign);
  	CM_f_abs_1_1L (y,23,8);
  	CM_f_mult_const_add_const_1L (y,y,0.4,GS_PI/4.0,23,8);
  	CM_f_tan_1_1L (y,23,8);
  	CM_f_ln_1_1L (y,23,8);
	CM_f_multiply_constant_2_1L (y,1.0/LAT_SCALE,23,8);
  	CM_logxor_2_1L(CM_add_offset_to_field_id(y,31),sign,1);
  	CM_deallocate_stack_through(sign);
	CM_f_multiply_constant_2_1L (data+32,FACTOR,23,8);

}	/* GD_miller */


/*   Function Header   -----------------------------
     Description: CM inverse miller projection
     Parameters: data   CM_field_id_t pointing to contiguous
                        array containing (x,y,alt)
                 parameters  *float to projection parameters
     Preconditions: in the vp_set of data, 
     				data in 3D rectangular
     Postconditions: none
     Return Value: does operation in place
     Logic:
     End Header   ---------------------------------- */


void GD_inverse_miller (data,success,parameters)	
	CM_field_id_t data;
        CM_field_id_t success;
  	float *parameters;
{
	CM_field_id_t lat = data;
	CM_field_id_t lon = CM_add_offset_to_field_id (data,32);
  	CM_field_id_t sign = CM_allocate_stack_field (1);
  	
	CM_u_move_constant_1L(success, 1, 1);
  	CM_swap_2_1L (lat,lon,32);  	
  	
  	CM_f_multiply_constant_2_1L (lon,LON_SCALE,23,8);

  	CM_f_lt_zero_1L (lat,23,8);
  	CM_store_test(sign);
  	CM_f_abs_1_1L (lat,23,8);
  	CM_f_multiply_constant_2_1L (lat,LAT_SCALE,23,8);
  	CM_f_exp_1_1L (lat,23,8);
  	CM_f_atan_1_1L (lat,23,8);
  	CM_f_sub_const_mult_const_1L (lat,lat,GS_PI/4.0,2.5,23,8);
  	CM_logxor_2_1L(CM_add_offset_to_field_id(lat,31),sign,1);
  	CM_deallocate_stack_through(sign);
	CM_f_divide_constant_2_1L (data+32,FACTOR,23,8);
  	
}	/* GD_inverse_miller */



/* use context instead of all 64k*/
void GD_miller_border(dest,parameters)
     CM_field_id_t dest;
     float *parameters;
{
  CM_my_news_coordinate_1L (dest+96,0,16);
  CM_f_u_float_2_2L (dest,dest+96,14,23,8);
  CM_f_mult_const_add_const_1L (dest,dest,2.0/16384.0,-1.0,23,8);
  CM_f_move_constant_1L (dest+32,1.0,23,8);
  CM_u_move_1L (dest+32+31,dest+96+15,1);
  CM_load_context(dest+96+14);
  CM_swap_2_1L (dest,dest+32,32);
  CM_set_context();
  CM_f_multiply_constant_2_1L (dest+32,FACTOR,23,8);
  CM_get_from_news_always_1L (dest+128,dest,0,CM_upward,64);
  CM_f_move_constant_1L (dest+64,0.0,23,8);
  CM_f_move_constant_1L (dest+96,0.0,23,8);
  CM_f_move_constant_1L (dest+64+128,0.0,23,8);
  CM_f_move_constant_1L (dest+96+128,0.0,23,8);
}



