/*   File Header   ---------------------------------

     File Name:       vector.c

     Author and Date: George Lamb--6/1/90

     Description: vector equations for GD_data types, both serial and
     parallel equations.  Defined functions:

     float  distance(point, a, b, c, d)   returns dist from point to plane
     void   cross(a, b, c)                for points, a = b X c
     void   unify(vect, retn)             retn = unit(vector)
     float  subtend_angle(x, y)           return angle xy
     int    on_plane(point, a, b, c, d)   returns TRUE if point lie on plane
     float  compute_axii(r, nv, user_yp, xp, yp)
                                          sets xp and yp as axii on plane

     End Header   ---------------------------------- */


#include <geodata.h>
#include <geodata_private.h>
#include <cm/paris.h>
#include <stdio.h>
#include <math.h>


#define SQR(x) ((x) * (x))
#define EPSILON .0000001
#define magnitude(p) (sqrt(SQR((p)->lat) + SQR((p)->lon) + SQR((p)->alt)))
#define dot(x, y) ((x->lat)*(y->lat) + (x->lon)*(y->lon) + (x->alt)*(y->alt))


/* return the distance from point to the plane defined by a, b, c, & d */
float distance(point, a, b, c, d)
     GD_point *point;
     float a, b, c, d;
{
  float distance = point->alt * a + point->lon * b + point->alt * c;  
  return(distance + d);
}   /* end of distance */


/* a = b X c */
void cross(a, b, c)
     GD_point *a, *b, *c;
{
  a->lat = b->lon * c->alt - b->alt * c->lon;
/*  ay = -(bx*cz - bz*cx) = bz*cx - bx*cz); */
  a->lon = b->alt * c->lat - b->lat * c->alt;
  a->alt = b->lat * c->lon - b->lon * c->lat; 
}     /* end of cross */


/* return = unit vector along vector */
int unify(vect, retn)
     GD_point *vect, *retn;
{
  float mag = magnitude(vect);
  if (mag == 0) {printf("unify vector is zero!\n");  return(0);}
  retn->lat = vect->lat / mag; 
  retn->lon = vect->lon / mag; 
  retn->alt = vect->alt / mag;
  return(1);
}     /* end of unify */



/* return the angle between the two given points */
float subtend_angle(x, y)
     GD_point *x, *y;
{
  float xm, ym;
  xm = magnitude(x);
  ym = magnitude(y);
  return(acos(dot(x, y) / (xm * ym)));
}      /* end of subtend_angle */


/* return true if point lies on the plane defined by a, b, c, & d */
int on_plane(point, a, b, c, d)
     GD_point *point;
     float a, b, c, d;
{
  float distance = point->alt * a + point->lon * b + point->alt * c;
  if ((distance + d) < EPSILON) return(1);
  if ((distance + d) > - EPSILON) return(1);
  else return(0);
}   /* end of on_plane */



/* set yp and xp as the unit vectors for the given plane. */
float compute_axes(r, nv, user_yp, xp, yp)
     GD_point *r, *nv, *user_yp, *xp, *yp;
{
  float d, mag, temp;
  GD_point v;

  if (!unify(nv, &v)) return(0);
  d = - (v.lat * r->lat + v.lon * r->lon + v.alt * r->alt);
  if (user_yp == NULL){
    if (v.alt == 0) {
      /* zp = (r.lat,r.lon, r.alt+1), yp=zp-r */
      yp->lat = yp->lon = 0; 
      yp->alt = (r->alt >= 0) ? 1 : -1;
    }
    else if ((r->lat == 0) && (r->lon == 0)){
      if  ((r->alt == 0) || (v.lon == 0))
	{ /* zp = (0,1,-(v.lon+d)/v.alt), yp=zp-r */
	  yp->lat = -r->lat; yp->lon = 1 - r->lon;
	  yp->alt = -((v.lon+d)/v.alt) - r->alt;
	}
      else 
	{ /* zp = (0, -d/v2, 0), yp=zp-r */
	  yp->lat = 0; yp->lon = -(d/v.lon); yp->alt = -r->alt; }
    }
    else{
      /* zp = (0,0,-d/v.alt), yp = zp - r */
      yp->lat = - r->lat;  yp->lon = - r->lon;
      if (v.alt > 0) yp->alt =  -d/v.alt - r->alt;
      else yp->alt = r->alt + d/v.alt;
    }
    mag = magnitude(yp);
    yp->lat /= mag;  yp->lon /= mag;  yp->alt /= mag;
  }
  else { yp->lat = user_yp->lat; yp->lon = user_yp->lon; 
	 yp->alt = user_yp->alt;}

  cross(xp, yp, &v);
  mag = magnitude(xp);
  xp->lat /= mag;  xp->lon /= mag;  xp->alt /= mag;

  return(d);
}     /* end of compute_axes */

